@@ -3,7 +3,6 @@
 // special feature is that its mostly *modeless*: you can use this menu while playing, without turning menus on or off
 // implementationwise, it is *stateless*: it keeps no internal gui structure, hit tests are instant, usage & implementation is greatly simplified
 
-#include "pch.h"
 #include "engine.h"
 
 #include "textedit.h"
@@ -14,7 +13,9 @@ static struct gui *windowhit = NULL;
 
 static float firstx, firsty;
 
-static enum {FIELDCOMMIT, FIELDABORT, FIELDEDIT, FIELDSHOW} fieldmode = FIELDSHOW; 
+enum {FIELDCOMMIT, FIELDABORT, FIELDEDIT, FIELDSHOW, FIELDKEY};
+
+static int fieldmode = FIELDSHOW; 
 static bool fieldsactive = false;
 
 static bool hascursor;
@@ -107,10 +108,10 @@ struct gui : g3d_gui
         if(!name) 
         {
             static string title;
-            s_sprintf(title)("%d", tpos);
+            formatstring(title)("%d", tpos);
             name = title;
         }
-        int w = text_width(name) - 2*INSERT;
+        int w = max(text_width(name) - 2*INSERT, 0);
         if(layoutpass) 
         {  
             ty = max(ty, ysize); 
@@ -122,7 +123,7 @@ struct gui : g3d_gui
             int h = FONTH-2*INSERT,
                 x1 = curx + tx,
                 x2 = x1 + w + ((skinx[3]-skinx[2]) + (skinx[5]-skinx[4]))*SKIN_SCALE,
-                y1 = cury - ((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE-h,
+                y1 = cury - ((skiny[6]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE-h,
                 y2 = cury;
             bool hit = tcurrent && windowhit==this && hitx>=x1 && hity>=y1 && hitx<x2 && hity<y2;
             if(hit && (!guiclicktab || mousebuttons&G3D_DOWN)) 
@@ -131,8 +132,8 @@ struct gui : g3d_gui
                 color = 0xFF0000;
             }
             
-            skin_(x1-skinx[visible()?2:6]*SKIN_SCALE, y1-skiny[1]*SKIN_SCALE, w, h, visible()?10:19, 9);
-            text_(name, x1 + (skinx[3]-skinx[2])*SKIN_SCALE - INSERT, y1 + (skiny[2]-skiny[1])*SKIN_SCALE - INSERT, tcolor, visible());
+            drawskin(x1-skinx[visible()?2:6]*SKIN_SCALE, y1-skiny[1]*SKIN_SCALE, w, h, visible()?10:19, 9, gui2d ? 1 : 2, light, alpha);
+            text_(name, x1 + (skinx[3]-skinx[2])*SKIN_SCALE - (w ? INSERT : INSERT/2), y1 + (skiny[2]-skiny[1])*SKIN_SCALE - INSERT, tcolor, visible());
         }
         tx += w + ((skinx[5]-skinx[4]) + (skinx[3]-skinx[2]))*SKIN_SCALE; 
     }
@@ -239,12 +240,12 @@ struct gui : g3d_gui
         return layout(size+SHADOW, size+SHADOW);
     }
     
-    int texture(Texture *t, float scale, int rotate, int xoff, int yoff, Texture *glowtex, const vec &glowcolor)
+    int texture(Texture *t, float scale, int rotate, int xoff, int yoff, Texture *glowtex, const vec &glowcolor, Texture *layertex)
     {
         autotab();
         if(scale==0) scale = 1;
         int size = (int)(scale*2*FONTH)-SHADOW;
-        if(t!=notexture && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW), rotate, xoff, yoff, glowtex, glowcolor);
+        if(t!=notexture && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW), rotate, xoff, yoff, glowtex, glowcolor, layertex);
         return layout(size+SHADOW, size+SHADOW);
     }
 
@@ -259,7 +260,7 @@ struct gui : g3d_gui
             if(!label)
             {
                 static string s;
-                s_sprintf(s)("%d", val);
+                formatstring(s)("%d", val);
                 label = s;
             }
             int w = text_width(label);
@@ -294,6 +295,16 @@ struct gui : g3d_gui
     }
 
     char *field(const char *name, int color, int length, int height, const char *initval, int initmode)
+    {
+        return field_(name, color, length, height, initval, initmode, FIELDEDIT);
+    }
+
+    char *keyfield(const char *name, int color, int length, int height, const char *initval, int initmode)
+    {
+        return field_(name, color, length, height, initval, initmode, FIELDKEY);
+    }
+
+    char *field_(const char *name, int color, int length, int height, const char *initval, int initmode, int fieldtype = FIELDEDIT)
     {	
         editor *e = useeditor(name, initmode, false, initval); // generate a new editor if necessary
         if(layoutpass)
@@ -328,13 +339,14 @@ struct gui : g3d_gui
             {
                 if(mousebuttons&G3D_DOWN) //mouse request focus
                 {   
+                    if(fieldtype==FIELDKEY) e->clear();
                     useeditor(name, initmode, true); 
                     e->mark(false);
-                    fieldmode = FIELDEDIT;
+                    fieldmode = fieldtype;
                 } 
             }
             bool editing = (fieldmode != FIELDSHOW) && (e==currentfocus());
-            if(hit && editing && (mousebuttons&G3D_PRESSED)!=0) e->hit(int(floor(hitx-(curx+FONTW/2))), int(floor(hity-cury)), (mousebuttons&G3D_DRAGGED)!=0); //mouse request position
+            if(hit && editing && (mousebuttons&G3D_PRESSED)!=0 && fieldtype==FIELDEDIT) e->hit(int(floor(hitx-(curx+FONTW/2))), int(floor(hity-cury)), (mousebuttons&G3D_DRAGGED)!=0); //mouse request position
             if(editing && ((fieldmode==FIELDCOMMIT) || (fieldmode==FIELDABORT) || !hit)) // commit field if user pressed enter or wandered out of focus 
             {
                 if(fieldmode==FIELDCOMMIT || (fieldmode!=FIELDABORT && !hit)) result = e->currentline().text;
@@ -426,7 +438,7 @@ struct gui : g3d_gui
         defaultshader->set();
     }
 
-    void icon_(Texture *t, bool overlaid, bool tiled, int x, int y, int size, bool hit, int rotate = 0, int xoff = 0, int yoff = 0, Texture *glowtex = NULL, const vec &glowcolor = vec(1, 1, 1)) 
+    void icon_(Texture *t, bool overlaid, bool tiled, int x, int y, int size, bool hit, int rotate = 0, int xoff = 0, int yoff = 0, Texture *glowtex = NULL, const vec &glowcolor = vec(1, 1, 1), Texture *layertex = NULL) 
     {
         float xs, ys, xt, yt;
         if(tiled)
@@ -472,8 +484,8 @@ struct gui : g3d_gui
         }
         loopk(4) { tc[k][0] = tc[k][0]/xt - float(xoff)/t->xs; tc[k][1] = tc[k][1]/yt - float(yoff)/t->ys; }
         vec color = hit ? vec(1, 0.5f, 0.5f) : (overlaid ? vec(1, 1, 1) : light);
-        glColor3fv(color.v);
         glBindTexture(GL_TEXTURE_2D, t->id);
+        glColor3fv(color.v);
         glBegin(GL_QUADS);
         glTexCoord2fv(tc[0]); glVertex2f(x,    y);
         glTexCoord2fv(tc[1]); glVertex2f(x+xs, y);
@@ -482,10 +494,10 @@ struct gui : g3d_gui
         glEnd();
         if(glowtex)
         {
-            if(hit || overlaid) { loopk(3) color[k] *= glowcolor[k]; glColor3fv(color.v); }
-            else glColor3fv(glowcolor.v);
             glBlendFunc(GL_SRC_ALPHA, GL_ONE);
             glBindTexture(GL_TEXTURE_2D, glowtex->id);
+            if(hit || overlaid) { loopk(3) color[k] *= glowcolor[k]; glColor3fv(color.v); }
+            else glColor3fv(glowcolor.v);
             glBegin(GL_QUADS);
             glTexCoord2fv(tc[0]); glVertex2f(x,    y);
             glTexCoord2fv(tc[1]); glVertex2f(x+xs, y);
@@ -494,12 +506,24 @@ struct gui : g3d_gui
             glEnd();
             glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
         }
+        if(layertex)
+        {
+            glBindTexture(GL_TEXTURE_2D, layertex->id);
+            glColor3fv(color.v);
+            glBegin(GL_QUADS);
+            glTexCoord2fv(tc[0]); glVertex2f(x+xs/2, y+ys/2);
+            glTexCoord2fv(tc[1]); glVertex2f(x+xs,   y+ys/2);
+            glTexCoord2fv(tc[2]); glVertex2f(x+xs,   y+ys);
+            glTexCoord2fv(tc[3]); glVertex2f(x+xs/2, y+ys);
+            glEnd();
+        }
+            
         if(tiled) defaultshader->set();
         if(overlaid) 
         {
-            if(!overlaytex) overlaytex = textureload("data/guioverlay.png");
-            glColor3fv(light.v);
+            if(!overlaytex) overlaytex = textureload("data/guioverlay.png", 3);
             glBindTexture(GL_TEXTURE_2D, overlaytex->id);
+            glColor3fv(light.v);
             glBegin(GL_QUADS);
             rect_(x, y, xs, ys, 0);
             glEnd();
@@ -510,7 +534,7 @@ struct gui : g3d_gui
     {		
         if(visible())
         {
-            if(!slidertex) slidertex = textureload("data/guislider.png");
+            if(!slidertex) slidertex = textureload("data/guislider.png", 3);
             glEnable(GL_TEXTURE_2D);
             glBindTexture(GL_TEXTURE_2D, slidertex->id);
             glBegin(GL_QUADS);
@@ -532,6 +556,17 @@ struct gui : g3d_gui
         layout(ishorizontal() ? FONTH : 0, ishorizontal() ? 0 : FONTH);
     }
 
+    void textbox(const char *text, int width, int height, int color) 
+    {
+        width *= FONTW;
+        height *= FONTH;
+        int w, h;
+        text_bounds(text, w, h, width);
+        if(h > height) height = h;
+        if(visible()) draw_text(text, curx, cury, color>>16, (color>>8)&0xFF, color&0xFF, 0xFF, -1, width);
+        layout(width, height);
+    }
+
     int button_(const char *text, int color, const char *icon, bool clickable, bool center)
     {
         const int padding = 10;
@@ -549,8 +584,8 @@ struct gui : g3d_gui
         
             if(icon)
             {
-                s_sprintfd(tname)("packages/icons/%s.jpg", icon);
-                icon_(textureload(tname), false, false, x, cury, ICON_SIZE, clickable && hit);
+                defformatstring(tname)("packages/icons/%s.jpg", icon);
+                icon_(textureload(tname, 3), false, false, x, cury, ICON_SIZE, clickable && hit);
                 x += ICON_SIZE;
             }
             if(icon && text) x += padding;
@@ -563,18 +598,18 @@ struct gui : g3d_gui
     static const int skinx[], skiny[];
     static const struct patch { ushort left, right, top, bottom; uchar flags; } patches[];
 
-    void skin_(int x, int y, int gapw, int gaph, int start, int n)//int vleft, int vright, int vtop, int vbottom, int start, int n) 
+    static void drawskin(int x, int y, int gapw, int gaph, int start, int n, int passes = 1, const vec &light = vec(1, 1, 1), float alpha = 0.80f)//int vleft, int vright, int vtop, int vbottom, int start, int n) 
     {
-        if(!skintex) skintex = textureload("data/guiskin.png");
+        if(!skintex) skintex = textureload("data/guiskin.png", 3);
         glBindTexture(GL_TEXTURE_2D, skintex->id);
         int gapx1 = INT_MAX, gapy1 = INT_MAX, gapx2 = INT_MAX, gapy2 = INT_MAX;
         float wscale = 1.0f/(SKIN_W*SKIN_SCALE), hscale = 1.0f/(SKIN_H*SKIN_SCALE);
         
-        loopj(gui2d ? 1 : 2)
+        loopj(passes)
         {	
             bool quads = false;
-            if(!gui2d) glDepthFunc(j ? GL_LEQUAL : GL_GREATER);
-            glColor4f(j ? light.x : 1.0f, j ? light.y : 1.0f, j ? light.z : 1.0f, gui2d || j ? 0.80f : 0.35f); //ghost when its behind something in depth
+            if(passes>1) glDepthFunc(j ? GL_LEQUAL : GL_GREATER);
+            glColor4f(j ? light.x : 1.0f, j ? light.y : 1.0f, j ? light.z : 1.0f, passes<=1 || j ? alpha : alpha/2); //ghost when its behind something in depth
             loopi(n)
             {
                 const patch &p = patches[start+i];
@@ -640,7 +675,7 @@ struct gui : g3d_gui
             if(quads) glEnd();
             else break; //if it didn't happen on the first pass, it won't happen on the second..
         }
-        if(!gui2d) glDepthFunc(GL_ALWAYS);
+        if(passes>1) glDepthFunc(GL_ALWAYS);
     } 
 
     vec origin, scale, *savedorigin;
@@ -650,18 +685,19 @@ struct gui : g3d_gui
 
     static float basescale, maxscale;
     static bool passthrough;
+    static float alpha;
     static vec light;
 
     void adjustscale()
     {
-        int w = xsize + (skinx[2]-skinx[1])*SKIN_SCALE + (skinx[10]-skinx[9])*SKIN_SCALE, h = ysize + (skiny[8]-skiny[6])*SKIN_SCALE;
+        int w = xsize + (skinx[2]-skinx[1])*SKIN_SCALE + (skinx[10]-skinx[9])*SKIN_SCALE, h = ysize + (skiny[9]-skiny[7])*SKIN_SCALE;
         if(tcurrent) h += ((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE + FONTH-2*INSERT;
-        else h += (skiny[5]-skiny[3])*SKIN_SCALE;
+        else h += (skiny[6]-skiny[3])*SKIN_SCALE;
 
         float aspect = float(screen->h)/float(screen->w), fit = 1.0f;
         if(w*aspect*basescale>1.0f) fit = 1.0f/(w*aspect*basescale);
         if(h*basescale*fit>maxscale) fit *= maxscale/(h*basescale*fit);
-        origin = vec(0.5f-((w-xsize)/2 - (skinx[2]-skinx[1])*SKIN_SCALE)*aspect*scale.x*fit, 0.5f + (0.5f*h-(skiny[8]-skiny[6])*SKIN_SCALE)*scale.y*fit, 0);
+        origin = vec(0.5f-((w-xsize)/2 - (skinx[2]-skinx[1])*SKIN_SCALE)*aspect*scale.x*fit, 0.5f + (0.5f*h-(skiny[9]-skiny[7])*SKIN_SCALE)*scale.y*fit, 0);
         scale = vec(aspect*scale.x*fit, scale.y*fit, 1);
     }
 
@@ -673,7 +709,8 @@ struct gui : g3d_gui
             if(allowinput) hascursor = true;
         }
         basescale = initscale;
-        if(layoutpass) scale.x = scale.y = scale.z = basescale*min((totalmillis-starttime)/300.0f, 1.0f);
+        if(layoutpass) scale.x = scale.y = scale.z = min(basescale*(totalmillis-starttime)/300.0f, basescale);
+        alpha = allowinput ? 0.80f : 0.60f;
         passthrough = scale.x<basescale || !allowinput;
         curdepth = -1;
         curlist = -1;
@@ -711,8 +748,8 @@ struct gui : g3d_gui
                 light.mul(1.0f + max(intensity, 0.0f));
             }
 
-            skin_(curx-skinx[2]*SKIN_SCALE, cury-skiny[5]*SKIN_SCALE, xsize, ysize, 0, 9);
-            if(!tcurrent) skin_(curx-skinx[5]*SKIN_SCALE, cury-skiny[5]*SKIN_SCALE, xsize, 0, 9, 1);
+            drawskin(curx-skinx[2]*SKIN_SCALE, cury-skiny[6]*SKIN_SCALE, xsize, ysize, 0, 9, gui2d ? 1 : 2, light, alpha);
+            if(!tcurrent) drawskin(curx-skinx[5]*SKIN_SCALE, cury-skiny[6]*SKIN_SCALE, xsize, 0, 9, 1, gui2d ? 1 : 2, light, alpha);
         }
     }
 
@@ -722,12 +759,12 @@ struct gui : g3d_gui
         {	
             xsize = max(tx, xsize);
             ysize = max(ty, ysize);
-            ysize = max(ysize, (skiny[6]-skiny[5])*SKIN_SCALE);
+            ysize = max(ysize, (skiny[7]-skiny[6])*SKIN_SCALE);
             if(tcurrent) *tcurrent = max(1, min(*tcurrent, tpos));
             if(gui2d) adjustscale();
             if(!windowhit && !passthrough)
             {
-                int intersects = INTERSECT_MIDDLE;
+                float dist = 0;
                 if(gui2d)
                 {
                     hitx = (cursorx - origin.x)/scale.x;
@@ -735,23 +772,27 @@ struct gui : g3d_gui
                 }
                 else
                 {
-                    vec planenormal = vec(origin).sub(camera1->o).set(2, 0).normalize(), intersectionpoint;
-                    intersects = intersect_plane_line(camera1->o, worldpos, origin, planenormal, intersectionpoint);
-                    vec intersectionvec = vec(intersectionpoint).sub(origin), xaxis(-planenormal.y, planenormal.x, 0);
-                    hitx = xaxis.dot(intersectionvec)/scale.x;
-                    hity = -intersectionvec.z/scale.y;
+                    plane p;
+                    p.toplane(vec(origin).sub(camera1->o).set(2, 0).normalize(), origin);
+                    if(p.rayintersect(camera1->o, camdir, dist) && dist>=0)
+                    {
+                        vec hitpos(camdir);
+                        hitpos.mul(dist).add(camera1->o).sub(origin);
+                        hitx = vec(-p.y, p.x, 0).dot(hitpos)/scale.x;
+                        hity = -hitpos.z/scale.y;
+                    }
                 }
                 if((mousebuttons & G3D_PRESSED) && (fabs(hitx-firstx) > 2 || fabs(hity - firsty) > 2)) mousebuttons |= G3D_DRAGGED;
-                if(intersects>=INTERSECT_MIDDLE && hitx>=-xsize/2 && hitx<=xsize/2 && hity<=0)
+                if(dist>=0 && hitx>=-xsize/2 && hitx<=xsize/2 && hity<=0)
                 {
-                    if(hity>=-ysize || (tcurrent && hity>=-ysize-(FONTH-2*INSERT)-((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE && hitx<=tx-xsize/2))
+                    if(hity>=-ysize || (tcurrent && hity>=-ysize-(FONTH-2*INSERT)-((skiny[6]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE && hitx<=tx-xsize/2))
                         windowhit = this;
                 }
             }
         }
         else
         {
-            if(tcurrent && tx<xsize) skin_(curx+tx-skinx[5]*SKIN_SCALE, -ysize-skiny[5]*SKIN_SCALE, xsize-tx, FONTH, 9, 1);
+            if(tcurrent && tx<xsize) drawskin(curx+tx-skinx[5]*SKIN_SCALE, -ysize-skiny[6]*SKIN_SCALE, xsize-tx, FONTH, 9, 1, gui2d ? 1 : 2, light, alpha);
             glPopMatrix();
         }
         poplist();
@@ -761,25 +802,25 @@ struct gui : g3d_gui
 Texture *gui::skintex = NULL, *gui::overlaytex = NULL, *gui::slidertex = NULL;
 
 //chop skin into a grid
-const int gui::skiny[] = {0, 7, 21, 34, 48, 56, 104, 111, 116, 128},
-          gui::skinx[] = {0, 11, 23, 37, 105, 119, 137, 151, 215, 229, 245, 256};
+const int gui::skiny[] = {0, 7, 21, 34, 43, 48, 56, 104, 111, 117, 128},
+          gui::skinx[] = {0, 11, 23, 37, 105, 119, 137, 151, 215, 229, 246, 256};
 //Note: skinx[3]-skinx[2] = skinx[7]-skinx[6]
 //      skinx[5]-skinx[4] = skinx[9]-skinx[8]		 
 const gui::patch gui::patches[] = 
 { //arguably this data can be compressed - it depends on what else needs to be skinned in the future
-    {1,2,3,5,  0},    // body
-    {2,9,4,5,  0x01},
-    {9,10,3,5, 0},
+    {1,2,3,6,  0},    // body
+    {2,9,5,6,  0x01},
+    {9,10,3,6, 0},
 
-    {1,2,5,6,  0x10},
-    {2,9,5,6,  0x11},
-    {9,10,5,6, 0x10},
+    {1,2,6,7,  0x10},
+    {2,9,6,7,  0x11},
+    {9,10,6,7, 0x10},
 
-    {1,2,6,8,  0},
-    {2,9,6,8,  0x01},
-    {9,10,6,8, 0},
+    {1,2,7,9,  0},
+    {2,9,7,9,  0x01},
+    {9,10,7,9, 0},
 
-    {5,6,3,4, 0x01}, // top
+    {5,6,3,5, 0x01}, // top
 
     {2,3,1,2, 0},    // selected tab
     {3,4,1,2, 0x01},
@@ -787,9 +828,9 @@ const gui::patch gui::patches[] =
     {2,3,2,3, 0x10},
     {3,4,2,3, 0x11},
     {4,5,2,3, 0x10},
-    {2,3,3,4, 0},
-    {3,4,3,4, 0x01},
-    {4,5,3,4, 0},
+    {2,3,3,5, 0},
+    {3,4,3,5, 0x01},
+    {4,5,3,5, 0},
 
     {6,7,1,2, 0},    // deselected tab
     {7,8,1,2, 0x01},
@@ -797,13 +838,13 @@ const gui::patch gui::patches[] =
     {6,7,2,3, 0x10},
     {7,8,2,3, 0x11},
     {8,9,2,3, 0x10},
-    {6,7,3,4, 0},
-    {7,8,3,4, 0x01},
-    {8,9,3,4, 0},
+    {6,7,3,5, 0},
+    {7,8,3,5, 0x01},
+    {8,9,3,5, 0},
 };
 
 vector<gui::list> gui::lists;
-float gui::basescale, gui::maxscale = 1, gui::hitx, gui::hity;
+float gui::basescale, gui::maxscale = 1, gui::hitx, gui::hity, gui::alpha;
 bool gui::passthrough, gui::shouldmergehits = false, gui::shouldautotab = true;
 vec gui::light;
 int gui::curdepth, gui::curlist, gui::xsize, gui::ysize, gui::curx, gui::cury;
@@ -814,11 +855,28 @@ VARP(guipushdist, 1, 4, 64);
 
 bool menukey(int code, bool isdown, int cooked)
 {
+    editor *e = currentfocus();
+    if(fieldmode == FIELDKEY)
+    {
+        switch(code)
+        {
+            case SDLK_ESCAPE:
+                if(isdown) fieldmode = FIELDCOMMIT;
+                return true;
+        }
+        const char *keyname = getkeyname(code);
+        if(keyname && isdown)
+        {
+            if(e->lines.length()!=1 || !e->lines[0].empty()) e->insert(" ");
+            e->insert(keyname);
+        }
+        return true;
+    }
+
     if(code==-1 && g3d_windowhit(isdown, true)) return true;
     else if(code==-3 && g3d_windowhit(isdown, false)) return true;
 
-    editor *e = currentfocus();
-    if((fieldmode == FIELDSHOW) || !e)
+    if(fieldmode == FIELDSHOW || !e)
     {
         if(windowhit) switch(code)
         {
@@ -882,6 +940,8 @@ bool menukey(int code, bool isdown, int cooked)
         case SDLK_DOWN:
         case SDLK_LEFT:
         case SDLK_RIGHT:
+        case SDLK_LSHIFT:
+        case SDLK_RSHIFT:
         case -4:
         case -5:
             break;
@@ -918,7 +978,7 @@ VARNP(gui2d, usegui2d, 0, 1, 1);
 
 void g3d_addgui(g3d_callback *cb, vec &origin, int flags)
 {
-    bool gui2d = flags&GUI_FORCE_2D || (flags&GUI_2D && usegui2d);
+    bool gui2d = flags&GUI_FORCE_2D || (flags&GUI_2D && usegui2d) || mainmenu;
     if(!gui2d && flags&GUI_FOLLOW && useguifollow) origin.z = player->o.z-(player->eyeheight-1);
     gui &g = (gui2d ? guis2d : guis3d).add();
     g.cb = cb;
@@ -940,8 +1000,11 @@ bool g3d_windowhit(bool on, bool act)
     extern int cleargui(int n);
     if(act) 
     {
-        if(on) { firstx = gui::hitx; firsty = gui::hity; }
-        mousebuttons |= (actionon=on) ? G3D_DOWN : G3D_UP;
+        if(actionon || windowhit)
+        {
+            if(on) { firstx = gui::hitx; firsty = gui::hity; }
+            mousebuttons |= (actionon=on) ? G3D_DOWN : G3D_UP;
+        }
     } else if(!on && windowhit) cleargui(1);
     return (guis2d.length() && hascursor) || (windowhit && !windowhit->gui2d);
 }
@@ -958,9 +1021,9 @@ void g3d_render()
     // call all places in the engine that may want to render a gui from here, they call g3d_addgui()
     extern void g3d_texturemenu();
     
-    g3d_texturemenu();
+    if(!mainmenu) g3d_texturemenu();
     g3d_mainmenu();
-    cl->g3d_gamemenus();
+    if(!mainmenu) game::g3d_gamemenus();
 
     guis2d.sort(g3d_sort);
     guis3d.sort(g3d_sort);
@@ -984,6 +1047,7 @@ void g3d_render()
 
     if(guis3d.length())
     {
+        glEnable(GL_DEPTH_TEST);
         glDepthFunc(GL_ALWAYS);
         glDepthMask(GL_FALSE);
 
@@ -991,6 +1055,7 @@ void g3d_render()
 
         glDepthFunc(GL_LESS);
         glDepthMask(GL_TRUE);
+        glDisable(GL_DEPTH_TEST);
     }
 
     if(guis2d.length())
@@ -1004,12 +1069,8 @@ void g3d_render()
         glPushMatrix();
         glLoadIdentity();
 
-        glDisable(GL_DEPTH_TEST);
-
         loopvrev(guis2d) guis2d[i].cb->gui(guis2d[i], false);
 
-        glEnable(GL_DEPTH_TEST);
-
         glMatrixMode(GL_PROJECTION);
         glPopMatrix();
         glMatrixMode(GL_MODELVIEW);
@@ -1032,3 +1093,19 @@ void g3d_render()
     mousebuttons = 0;
 }
 
+void consolebox(int x1, int y1, int x2, int y2)
+{
+    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+    glPushMatrix();
+    glTranslatef(x1, y1, 0);
+    float bw = x2 - x1, bh = y2 - y1, aspect = bw/bh, sh = bh, sw = sh*aspect;
+    bw *= float(4*FONTH)/(SKIN_H*SKIN_SCALE);
+    bh *= float(4*FONTH)/(SKIN_H*SKIN_SCALE);
+    sw /= bw + (gui::skinx[2]-gui::skinx[1] + gui::skinx[10]-gui::skinx[9])*SKIN_SCALE;
+    sh /= bh + (gui::skiny[9]-gui::skiny[7] + gui::skiny[6]-gui::skiny[4])*SKIN_SCALE;
+    glScalef(sw, sh, 1);
+    gui::drawskin(-gui::skinx[1]*SKIN_SCALE, -gui::skiny[4]*SKIN_SCALE, int(bw), int(bh), 0, 9, 1, vec(1, 1, 1), 0.60f);
+    gui::drawskin((-gui::skinx[1] + gui::skinx[2] - gui::skinx[5])*SKIN_SCALE, -gui::skiny[4]*SKIN_SCALE, int(bw), 0, 9, 1, 1, vec(1, 1, 1), 0.60f);
+    glPopMatrix();
+}
+
