@@ -1,7 +1,6 @@
 // command.cpp: implements the parsing and execution of a tiny script language which
 // is largely backwards compatible with the quake console language.
 
-#include "pch.h"
 #include "engine.h"
 
 char *exchangestr(char *o, const char *n) { delete[] o; return newstring(n); }
@@ -31,32 +30,38 @@ void clear_command()
     if(idents) idents->clear();
 }
 
+void clearoverride(ident &i)
+{
+    if(i.override==NO_OVERRIDE) return;
+    switch(i.type)
+    {
+        case ID_ALIAS:
+            if(i.action[0])
+            {
+                if(i.action != i.isexecuting) delete[] i.action;
+                i.action = newstring("");
+            }
+            break;
+        case ID_VAR:
+            *i.storage.i = i.overrideval.i;
+            i.changed();
+            break;
+        case ID_FVAR:
+            *i.storage.f = i.overrideval.f;
+            i.changed();
+            break;
+        case ID_SVAR:
+            delete[] *i.storage.s;
+            *i.storage.s = i.overrideval.s;
+            i.changed();
+            break;
+    }
+    i.override = NO_OVERRIDE;
+}
+
 void clearoverrides()
 {
-    enumerate(*idents, ident, i,
-        if(i.override!=NO_OVERRIDE)
-        {
-            switch(i.type)
-            {
-                case ID_ALIAS: 
-                    if(i.action[0]) i.action = exchangestr(i.action, ""); 
-                    break;
-                case ID_VAR: 
-                    *i.storage.i = i.overrideval.i;
-                    i.changed();
-                    break;
-                case ID_FVAR:
-                    *i.storage.f = i.overrideval.f;
-                    i.changed();
-                    break;
-                case ID_SVAR:
-                    delete[] *i.storage.s;
-                    *i.storage.s = i.overrideval.s;
-                    i.changed();
-                    break;
-            }
-            i.override = NO_OVERRIDE;
-        });
+    enumerate(*idents, ident, i, clearoverride(i));
 }
 
 void pushident(ident &id, char *val)
@@ -85,7 +90,7 @@ ident *newident(const char *name)
     if(!id)
     {
         ident init(ID_ALIAS, newstring(name), newstring(""), persistidents ? IDF_PERSIST : 0);
-        id = idents->access(init.name, &init);
+        id = &idents->access(init.name, init);
     }
     return id;
 }
@@ -106,8 +111,17 @@ void pop(char *name)
     if(id) popident(*id);
 }
 
+void resetvar(char *name)
+{
+    ident *id = idents->access(name);
+    if(!id) return;
+    if(id->flags&IDF_READONLY) conoutf(CON_ERROR, "variable %s is read-only", id->name);
+    else clearoverride(*id);
+}
+
 COMMAND(push, "ss");
 COMMAND(pop, "s");
+COMMAND(resetvar, "s");
 
 void aliasa(const char *name, char *action)
 {
@@ -116,7 +130,7 @@ void aliasa(const char *name, char *action)
     {
         ident b(ID_ALIAS, newstring(name), action, persistidents ? IDF_PERSIST : 0);
         if(overrideidents) b.override = OVERRIDDEN;
-        idents->access(b.name, &b);
+        idents->access(b.name, b);
     }
     else if(b->type != ID_ALIAS)
     {
@@ -150,15 +164,15 @@ int variable(const char *name, int min, int cur, int max, int *storage, void (*f
 {
     if(!idents) idents = new identtable;
     ident v(ID_VAR, name, min, cur, max, storage, (void *)fun, flags);
-    idents->access(name, &v);
+    idents->access(name, v);
     return cur;
 }
 
-float fvariable(const char *name, float cur, float *storage, void (*fun)(), int flags)
+float fvariable(const char *name, float min, float cur, float max, float *storage, void (*fun)(), int flags)
 {
     if(!idents) idents = new identtable;
-    ident v(ID_FVAR, name, cur, storage, (void *)fun, flags);
-    idents->access(name, &v);
+    ident v(ID_FVAR, name, min, cur, max, storage, (void *)fun, flags);
+    idents->access(name, v);
     return cur;
 }
 
@@ -166,7 +180,7 @@ char *svariable(const char *name, const char *cur, char **storage, void (*fun)()
 {
     if(!idents) idents = new identtable;
     ident v(ID_SVAR, name, newstring(cur), storage, (void *)fun, flags);
-    idents->access(name, &v);
+    idents->access(name, v);
     return v.val.s;
 }
 
@@ -174,22 +188,44 @@ char *svariable(const char *name, const char *cur, char **storage, void (*fun)()
     ident *id = idents->access(name); \
     if(!id || id->type!=vartype) return retval;
 #define GETVAR(id, name, retval) _GETVAR(id, ID_VAR, name, retval)
-void setvar(const char *name, int i, bool dofunc) 
-{ 
+#define OVERRIDEVAR(errorval, saveval, resetval, clearval) \
+    if(overrideidents || id->flags&IDF_OVERRIDE) \
+    { \
+        if(id->flags&IDF_PERSIST) \
+        { \
+            conoutf(CON_ERROR, "cannot override persistent variable %s", id->name); \
+            errorval; \
+        } \
+        if(id->override==NO_OVERRIDE) { saveval; id->override = OVERRIDDEN; } \
+        else { clearval; } \
+    } \
+    else \
+    { \
+        if(id->override!=NO_OVERRIDE) { resetval; id->override = NO_OVERRIDE; } \
+        clearval; \
+    }
+
+void setvar(const char *name, int i, bool dofunc, bool doclamp)
+{
     GETVAR(id, name, );
-    *id->storage.i = clamp(i, id->minval, id->maxval); 
+    OVERRIDEVAR(return, id->overrideval.i = *id->storage.i, , )
+    if(doclamp) *id->storage.i = clamp(i, id->minval, id->maxval);
+    else *id->storage.i = i;
     if(dofunc) id->changed();
-} 
-void setfvar(const char *name, float f, bool dofunc)
+}
+void setfvar(const char *name, float f, bool dofunc, bool doclamp)
 {
     _GETVAR(id, ID_FVAR, name, );
-    *id->storage.f = f;
+    OVERRIDEVAR(return, id->overrideval.f = *id->storage.f, , );
+    if(doclamp) *id->storage.f = clamp(f, id->minvalf, id->maxvalf);
+    else *id->storage.f = f;
     if(dofunc) id->changed();
 }
 void setsvar(const char *name, const char *str, bool dofunc)
 {
     _GETVAR(id, ID_SVAR, name, );
-    *id->storage.s = exchangestr(*id->storage.s, str);
+    OVERRIDEVAR(return, id->overrideval.s = *id->storage.s, delete[] id->overrideval.s, delete[] *id->storage.s);
+    *id->storage.s = newstring(str);
     if(dofunc) id->changed();
 }
 int getvar(const char *name) 
@@ -233,14 +269,14 @@ bool addcommand(const char *name, void (*fun)(), const char *narg)
 {
     if(!idents) idents = new identtable;
     ident c(ID_COMMAND, name, narg, (void *)fun);
-    idents->access(name, &c);
+    idents->access(name, c);
     return false;
 }
 
 void addident(const char *name, ident *id)
 {
     if(!idents) idents = new identtable;
-    idents->access(name, id);
+    idents->access(name, *id);
 }
 
 static vector<vector<char> *> wordbufs;
@@ -275,6 +311,44 @@ void parsemacro(const char *&p, int level, vector<char> &wordbuf)
     while(*alias) wordbuf.add(*alias++);
 }
 
+const char *parsestring(const char *p)
+{
+    for(; *p; p++) switch(*p) 
+    {
+        case '\r':
+        case '\n':
+        case '\"':
+            return p;
+        case '^':
+            if(*++p) break;
+            return p;
+    }
+    return p;
+}
+
+int escapestring(char *dst, const char *src, const char *end)
+{
+    char *start = dst;
+    while(src < end)
+    {
+        int c = *src++;
+        if(c == '^')
+        {
+            if(src >= end) break;
+            int e = *src++;
+            switch(e)
+            {
+                case 'n': *dst++ = '\n'; break;
+                case 't': *dst++ = '\t'; break;
+                case 'f': *dst++ = '\f'; break;
+                default: *dst++ = e; break;
+            }            
+        }
+        else *dst++ = c;
+    }
+    return dst - start;
+}
+
 char *parseexp(const char *&p, int right)          // parse any nested set of () or []
 {
     if(bufnest++>=wordbufs.length()) wordbufs.add(new vector<char>);
@@ -283,36 +357,37 @@ char *parseexp(const char *&p, int right)          // parse any nested set of ()
     for(int brak = 1; brak; )
     {
         int c = *p++;
-        if(c=='\r') continue;               // hack
-        if(left=='[' && c=='@')
+        switch(c)
         {
-            parsemacro(p, brak, wordbuf);
-            continue;
-        }
-        if(c=='\"')
-        {
-            wordbuf.add(c);
-            const char *end = p+strcspn(p, "\"\r\n\0");
-            while(p < end) wordbuf.add(*p++);
-            if(*p=='\"') wordbuf.add(*p++);
-            continue;
-        }
-        if(c=='/' && *p=='/')
-        {
-            p += strcspn(p, "\n\0");
-            continue;
-        }
-            
+            case '\r': continue;
+            case '@': 
+                if(left == '[') { parsemacro(p, brak, wordbuf); continue; }
+                break;
+            case '\"':
+            {
+                wordbuf.add(c);
+                const char *end = parsestring(p);
+                wordbuf.put(p, end - p);
+                p = end;
+                if(*p=='\"') wordbuf.add(*p++);
+                continue;
+            }
+            case '/':
+                if(*p=='/')
+                {
+                    p += strcspn(p, "\n\0");
+                    continue;
+                }
+                break;
+            case '\0':
+                p--;
+                conoutf(CON_ERROR, "missing \"%c\"", right);
+                wordbuf.setsize(0); 
+                bufnest--;
+                return NULL; 
+        } 
         if(c==left) brak++;
         else if(c==right) brak--;
-        else if(!c) 
-        { 
-            p--;
-            conoutf(CON_ERROR, "missing \"%c\"", right);
-            wordbuf.setsize(0); 
-            bufnest--;
-            return NULL; 
-        }
         wordbuf.add(c);
     }
     wordbuf.pop();
@@ -338,8 +413,8 @@ char *lookup(char *n)                           // find value of ident reference
     ident *id = idents->access(n+1);
     if(id) switch(id->type)
     {
-        case ID_VAR: { s_sprintfd(t)("%d", *id->storage.i); return exchangestr(n, t); }
-        case ID_FVAR: { s_sprintfd(t)("%f", *id->storage.f); return exchangestr(n, t); }
+        case ID_VAR: { defformatstring(t)("%d", *id->storage.i); return exchangestr(n, t); }
+        case ID_FVAR: return exchangestr(n, floatstr(*id->storage.f));
         case ID_SVAR: return exchangestr(n, *id->storage.s);
         case ID_ALIAS: return exchangestr(n, id->action);
     }
@@ -347,7 +422,7 @@ char *lookup(char *n)                           // find value of ident reference
     return n;
 }
 
-char *parseword(const char *&p)                       // parse single argument, including expressions
+char *parseword(const char *&p, int arg, int &infix)                       // parse single argument, including expressions
 {
     for(;;)
     {
@@ -358,9 +433,10 @@ char *parseword(const char *&p)                       // parse single argument,
     if(*p=='\"')
     {
         p++;
-        const char *word = p;
-        p += strcspn(p, "\"\r\n\0");
-        char *s = newstring(word, p-word);
+        const char *end = parsestring(p);
+        char *s = newstring(end - p);
+        s[escapestring(s, p, end)] = '\0';
+        p = end;
         if(*p=='\"') p++;
         return s;
     }
@@ -375,6 +451,10 @@ char *parseword(const char *&p)                       // parse single argument,
         p += 2;
     }
     if(p-word==0) return NULL;
+    if(arg==1 && p-word==1) switch(*word)
+    {
+        case '=': infix = *word; break;
+    }
     char *s = newstring(word, p-word);
     if(*s=='$') return lookup(s);               // substitute variables
     return s;
@@ -396,12 +476,15 @@ char *conc(char **w, int n, bool space)
 
 VARN(numargs, _numargs, 0, 0, 25);
 
+static inline bool isinteger(char *c)
+{
+    return isdigit(c[0]) || ((c[0]=='+' || c[0]=='-' || c[0]=='.') && isdigit(c[1]));
+}
+
 #define parseint(s) strtol((s), NULL, 0)
 
 char *commandret = NULL;
 
-extern const char *addreleaseaction(const char *s);
-
 char *executeret(const char *p)               // all evaluation happens here, recursively
 {
     const int MAXWORDS = 25;                    // limit, remove
@@ -410,12 +493,12 @@ char *executeret(const char *p)               // all evaluation happens here, re
     #define setretval(v) { char *rv = v; if(rv) retval = rv; }
     for(bool cont = true; cont;)                // for each ; seperated statement
     {
-        int numargs = MAXWORDS;
+        int numargs = MAXWORDS, infix = 0;
         loopi(MAXWORDS)                         // collect all argument values
         {
             w[i] = (char *)"";
             if(i>numargs) continue;
-            char *s = parseword(p);             // parse and evaluate exps
+            char *s = parseword(p, i, infix);   // parse and evaluate exps
             if(s) w[i] = s;
             else numargs = i;
         }
@@ -427,17 +510,22 @@ char *executeret(const char *p)               // all evaluation happens here, re
         
         DELETEA(retval);
 
-        if(w[1][0]=='=' && !w[1][1])
+        if(infix)
         {
-            aliasa(c, numargs>2 ? w[2] : newstring(""));
-            w[2] = NULL;
+            switch(infix)
+            {
+                case '=':    
+                    aliasa(c, numargs>2 ? w[2] : newstring(""));
+                    w[2] = NULL;
+                    break;
+            }
         }
         else
         {     
             ident *id = idents->access(c);
             if(!id)
             {
-                if(!isdigit(*c) && ((*c!='+' && *c!='-' && *c!='.') || !isdigit(c[1]))) 
+                if(!isinteger(c))
                     conoutf(CON_ERROR, "unknown command: %s", c);
                 setretval(newstring(c));
             }
@@ -460,7 +548,9 @@ char *executeret(const char *p)               // all evaluation happens here, re
                         case 's':                                 v[n] = w[++wn];     n++; break;
                         case 'i': nstor[n].i = parseint(w[++wn]); v[n] = &nstor[n].i; n++; break;
                         case 'f': nstor[n].f = atof(w[++wn]);     v[n] = &nstor[n].f; n++; break;
+#ifndef STANDALONE
                         case 'D': nstor[n].i = addreleaseaction(id->name) ? 1 : 0; v[n] = &nstor[n].i; n++; break;
+#endif
                         case 'V': v[n++] = w+1; nstor[n].i = numargs-1; v[n] = &nstor[n].i; n++; break;
                         case 'C': if(!cargs) cargs = conc(w+1, numargs-1, true); v[n++] = cargs; break;
                         default: fatal("builtin declared with illegal type");
@@ -485,50 +575,84 @@ char *executeret(const char *p)               // all evaluation happens here, re
                 }
 
                 case ID_VAR:                        // game defined variables 
-                    if(numargs <= 1) conoutf("%s = %d", c, *id->storage.i);      // var with no value just prints its current value
-                    else if(id->minval>id->maxval) conoutf(CON_ERROR, "variable %s is read-only", id->name);
-                    else
+                    if(numargs <= 1) 
                     {
-                        #define OVERRIDEVAR(saveval, resetval) \
-                            if(overrideidents || id->flags&IDF_OVERRIDE) \
-                            { \
-                                if(id->flags&IDF_PERSIST) \
-                                { \
-                                    conoutf(CON_ERROR, "cannot override persistent variable %s", id->name); \
-                                    break; \
-                                } \
-                                if(id->override==NO_OVERRIDE) { saveval; id->override = OVERRIDDEN; } \
-                            } \
-                            else if(id->override!=NO_OVERRIDE) { resetval; id->override = NO_OVERRIDE; }
-                        OVERRIDEVAR(id->overrideval.i = *id->storage.i, )
+                        if(id->flags&IDF_HEX && id->maxval==0xFFFFFF) 
+                            conoutf("%s = 0x%.6X (%d, %d, %d)", c, *id->storage.i, (*id->storage.i>>16)&0xFF, (*id->storage.i>>8)&0xFF, *id->storage.i&0xFF);
+                        else
+                            conoutf(id->flags&IDF_HEX ? "%s = 0x%X" : "%s = %d", c, *id->storage.i);      // var with no value just prints its current value
+                    }
+                    else if(id->flags&IDF_READONLY) conoutf(CON_ERROR, "variable %s is read-only", id->name);
+#ifndef STANDALONE
+                    else if(!(id->flags&IDF_OVERRIDE) || overrideidents || game::allowedittoggle())
+#else
+                    else
+#endif
+                    {
+                        OVERRIDEVAR(break, id->overrideval.i = *id->storage.i, , )
                         int i1 = parseint(w[1]);
+                        if(id->flags&IDF_HEX && numargs > 2)
+                        {
+                            i1 <<= 16;
+                            i1 |= parseint(w[2])<<8;    
+                            i1 |= parseint(w[3]);
+                        }
                         if(i1<id->minval || i1>id->maxval)
                         {
                             i1 = i1<id->minval ? id->minval : id->maxval;                // clamp to valid range
-                            conoutf(CON_ERROR, "valid range for %s is %d..%d", id->name, id->minval, id->maxval);
+                            conoutf(CON_ERROR,
+                                id->flags&IDF_HEX ?
+                                    (id->minval <= 255 ? "valid range for %s is %d..0x%X" : "valid range for %s is 0x%X..0x%X") :
+                                    "valid range for %s is %d..%d", 
+                                id->name, id->minval, id->maxval);
                         }
                         *id->storage.i = i1;
                         id->changed();                                             // call trigger function if available
+#ifndef STANDALONE
+                        if(id->flags&IDF_OVERRIDE && !overrideidents) game::vartrigger(id);
+#endif
                     }
                     break;
                   
                 case ID_FVAR:
-                    if(numargs <= 1) conoutf("%s = %f", c, *id->storage.f);
+                    if(numargs <= 1) conoutf("%s = %s", c, floatstr(*id->storage.f));
+                    else if(id->flags&IDF_READONLY) conoutf(CON_ERROR, "variable %s is read-only", id->name);
+#ifndef STANDALONE
+                    else if(!(id->flags&IDF_OVERRIDE) || overrideidents || game::allowedittoggle())
+#else
                     else
+#endif
                     {
-                        OVERRIDEVAR(id->overrideval.f = *id->storage.f, );
-                        *id->storage.f = atof(w[1]);
+                        OVERRIDEVAR(break, id->overrideval.f = *id->storage.f, , );
+                        float f1 = atof(w[1]);
+                        if(f1<id->minvalf || f1>id->maxvalf)
+                        {
+                            f1 = f1<id->minvalf ? id->minvalf : id->maxvalf;                // clamp to valid range
+                            conoutf(CON_ERROR, "valid range for %s is %s..%s", id->name, floatstr(id->minvalf), floatstr(id->maxvalf));
+                        }
+                        *id->storage.f = f1;
                         id->changed();
+#ifndef STANDALONE
+                        if(id->flags&IDF_OVERRIDE && !overrideidents) game::vartrigger(id);
+#endif
                     }
                     break;
  
                 case ID_SVAR:
                     if(numargs <= 1) conoutf(strchr(*id->storage.s, '"') ? "%s = [%s]" : "%s = \"%s\"", c, *id->storage.s);
+                    else if(id->flags&IDF_READONLY) conoutf(CON_ERROR, "variable %s is read-only", id->name);
+#ifndef STANDALONE
+                    else if(!(id->flags&IDF_OVERRIDE) || overrideidents || game::allowedittoggle())
+#else
                     else
+#endif
                     {
-                        OVERRIDEVAR(id->overrideval.s = *id->storage.s, delete[] id->overrideval.s);
+                        OVERRIDEVAR(break, id->overrideval.s = *id->storage.s, delete[] id->overrideval.s, delete[] *id->storage.s);
                         *id->storage.s = newstring(w[1]);
                         id->changed();
+#ifndef STANDALONE
+                        if(id->flags&IDF_OVERRIDE && !overrideidents) game::vartrigger(id);
+#endif
                     }
                     break;
                         
@@ -539,7 +663,7 @@ char *executeret(const char *p)               // all evaluation happens here, re
                     {
                         if(i > argids.length())
                         {
-                            s_sprintfd(argname)("arg%d", i);
+                            defformatstring(argname)("arg%d", i);
                             argids.add(newident(argname));
                         }
                         pushident(*argids[i-1], w[i]); // set any arguments as (global) arg values so functions can access them
@@ -571,60 +695,104 @@ int execute(const char *p)
     return i;
 }
 
-bool execfile(const char *cfgfile)
+bool execfile(const char *cfgfile, bool msg)
 {
     string s;
-    s_strcpy(s, cfgfile);
+    copystring(s, cfgfile);
     char *buf = loadfile(path(s), NULL);
-    if(!buf) return false;
+    if(!buf) 
+    {
+        if(msg) conoutf(CON_ERROR, "could not read \"%s\"", cfgfile);
+        return false;
+    }
     execute(buf);
     delete[] buf;
     return true;
 }
 
-void exec(const char *cfgfile)
+#ifndef STANDALONE
+static int sortidents(ident **x, ident **y)
 {
-    if(!execfile(cfgfile)) conoutf(CON_ERROR, "could not read \"%s\"", cfgfile);
+    return strcmp((*x)->name, (*y)->name);
+}
+
+void writeescapedstring(stream *f, const char *s)
+{
+    f->putchar('"');
+    for(; *s; s++) switch(*s)
+    {
+        case '\n': f->write("^n", 2); break;
+        case '\t': f->write("^t", 2); break;
+        case '\f': f->write("^f", 2); break;
+        case '"': f->write("^\"", 2); break;
+        default: f->putchar(*s); break;
+    }
+    f->putchar('"');
 }
 
 void writecfg()
 {
-    FILE *f = openfile(path(cl->savedconfig(), true), "w");
+    stream *f = openfile(path(game::savedconfig(), true), "w");
     if(!f) return;
-    fprintf(f, "// automatically written on exit, DO NOT MODIFY\n// delete this file to have %s overwrite these settings\n// modify settings in game, or put settings in %s to override anything\n\n", cl->defaultconfig(), cl->autoexec());
-    cc->writeclientinfo(f);
-    fprintf(f, "\n");
+    f->printf("// automatically written on exit, DO NOT MODIFY\n// delete this file to have %s overwrite these settings\n// modify settings in game, or put settings in %s to override anything\n\n", game::defaultconfig(), game::autoexec());
+    game::writeclientinfo(f);
+    f->printf("\n");
     writecrosshairs(f);
-    enumerate(*idents, ident, id,
+    vector<ident *> ids;
+    enumerate(*idents, ident, id, ids.add(&id));
+    ids.sort(sortidents);
+    loopv(ids)
+    {
+        ident &id = *ids[i];
         if(id.flags&IDF_PERSIST) switch(id.type)
         {
-            case ID_VAR: fprintf(f, "%s %d\n", id.name, *id.storage.i); break;
-            case ID_FVAR: fprintf(f, "%s %f\n", id.name, *id.storage.f); break;
-            case ID_SVAR: fprintf(f, "%s [%s]\n", id.name, *id.storage.s); break;
+            case ID_VAR: f->printf("%s %d\n", id.name, *id.storage.i); break;
+            case ID_FVAR: f->printf("%s %s\n", id.name, floatstr(*id.storage.f)); break;
+            case ID_SVAR: f->printf("%s ", id.name); writeescapedstring(f, *id.storage.s); f->putchar('\n'); break;
         }
-    );
-    fprintf(f, "\n");
+    }
+    f->printf("\n");
     writebinds(f);
-    fprintf(f, "\n");
-    enumerate(*idents, ident, id,
+    f->printf("\n");
+    loopv(ids)
+    {
+        ident &id = *ids[i];
         if(id.type==ID_ALIAS && id.flags&IDF_PERSIST && id.override==NO_OVERRIDE && !strstr(id.name, "nextmap_") && id.action[0])
         {
-            fprintf(f, "\"%s\" = [%s]\n", id.name, id.action);
+            f->printf("\"%s\" = [%s]\n", id.name, id.action);
         }
-    );
-    fprintf(f, "\n");
+    }
+    f->printf("\n");
     writecompletions(f);
-    fclose(f);
+    delete f;
 }
 
 COMMAND(writecfg, "");
+#endif
 
 // below the commands that implement a small imperative language. thanks to the semantics of
 // () and [] expressions, any control construct can be defined trivially.
 
-void intret(int v) { s_sprintfd(b)("%d", v); commandret = newstring(b); }
+void intret(int v) { defformatstring(b)("%d", v); commandret = newstring(b); }
 
-ICOMMAND(if, "sss", (char *cond, char *t, char *f), commandret = executeret(cond[0]!='0' ? t : f));
+const char *floatstr(float v)
+{
+    static int n = 0;
+    static string t[3];
+    n = (n + 1)%3;
+    formatstring(t[n])(v==int(v) ? "%.1f" : "%.7g", v);
+    return t[n];
+}
+
+void floatret(float v)
+{
+    commandret = newstring(floatstr(v));
+}
+
+#undef ICOMMANDNAME
+#define ICOMMANDNAME(name) _stdcmd
+
+ICOMMAND(if, "sss", (char *cond, char *t, char *f), commandret = executeret(cond[0] && (!isinteger(cond) || parseint(cond)) ? t : f));
 ICOMMAND(loop, "sis", (char *var, int *n, char *body), 
 {
     if(*n<=0) return;
@@ -725,7 +893,7 @@ void getalias_(char *s)
     result(getalias(s));
 }
 
-COMMAND(exec, "s");
+ICOMMAND(exec, "s", (char *file), execfile(file));
 COMMAND(concat, "C");
 COMMAND(result, "s");
 COMMAND(concatword, "V");
@@ -735,35 +903,93 @@ COMMAND(substr, "sii");
 COMMAND(listlen, "s");
 COMMANDN(getalias, getalias_, "s");
 
-void add  (int *a, int *b) { intret(*a + *b); }          COMMANDN(+, add, "ii");
-void mul  (int *a, int *b) { intret(*a * *b); }          COMMANDN(*, mul, "ii");
-void sub  (int *a, int *b) { intret(*a - *b); }          COMMANDN(-, sub, "ii");
-void divi (int *a, int *b) { intret(*b ? *a / *b : 0); } COMMANDN(div, divi, "ii");
-void mod  (int *a, int *b) { intret(*b ? *a % *b : 0); } COMMAND(mod, "ii");
-void equal(int *a, int *b) { intret((int)(*a == *b)); }  COMMANDN(=, equal, "ii");
-void nequal(int *a, int *b) { intret((int)(*a != *b)); } COMMANDN(!=, nequal, "ii");
-void lt   (int *a, int *b) { intret((int)(*a < *b)); }   COMMANDN(<, lt, "ii");
-void gt   (int *a, int *b) { intret((int)(*a > *b)); }   COMMANDN(>, gt, "ii");
-void lte  (int *a, int *b) { intret((int)(*a <= *b)); } COMMANDN(<=, lte, "ii");
-void gte  (int *a, int *b) { intret((int)(*a >= *b)); } COMMANDN(>=, gte, "ii");
-void xora (int *a, int *b) { intret(*a ^ *b); }          COMMANDN(^, xora, "ii");
-void nota (int *a)         { intret(*a == 0); }          COMMANDN(!, nota, "i");
-void mina (int *a, int *b) { intret(min(*a, *b)); }      COMMANDN(min, mina, "ii");
-void maxa (int *a, int *b) { intret(max(*a, *b)); }      COMMANDN(max, maxa, "ii");
+void looplist(const char *var, const char *list, const char *body, bool search)
+{
+    ident *id = newident(var);
+    if(id->type!=ID_ALIAS) { if(search) intret(-1); return; }
+    int n = 0;
+    for(const char *s = list;;)
+    {
+        whitespaceskip;
+        if(!*s) { if(search) intret(-1); break; }
+        const char *start = s;
+        elementskip;
+        const char *end = s;
+        if(*start=='"') { start++; if(end[-1]=='"') --end; }
+        char *val = newstring(start, end-start);
+        if(n++) aliasa(id->name, val);
+        else pushident(*id, val);
+        if(execute(body) && search) { intret(n-1); break; }
+    }
+    if(n) popident(*id);
+}
 
-void anda (char *a, char *b) { intret(execute(a)!=0 && execute(b)!=0); }
-void ora  (char *a, char *b) { intret(execute(a)!=0 || execute(b)!=0); }
+ICOMMAND(listfind, "sss", (char *var, char *list, char *body), looplist(var, list, body, true));
+ICOMMAND(looplist, "sss", (char *var, char *list, char *body), looplist(var, list, body, false));
+ICOMMAND(loopfiles, "ssss", (char *var, char *dir, char *ext, char *body),
+{
+    ident *id = newident(var);
+    if(id->type!=ID_ALIAS) return;
+    vector<char *> files;
+    listfiles(dir, ext[0] ? ext : NULL, files);
+    loopv(files)
+    {
+        char *file = files[i];
+        bool redundant = false;
+        loopj(i) if(!strcmp(files[j], file)) { redundant = true; break; }
+        if(redundant) { delete[] file; continue; }
+        if(i) aliasa(id->name, file);
+        else pushident(*id, file);
+        execute(body);
+    }
+    if(files.length()) popident(*id);
+});
 
-COMMANDN(&&, anda, "ss");
-COMMANDN(||, ora, "ss");
+ICOMMAND(+, "ii", (int *a, int *b), intret(*a + *b));
+ICOMMAND(*, "ii", (int *a, int *b), intret(*a * *b));
+ICOMMAND(-, "ii", (int *a, int *b), intret(*a - *b));
+ICOMMAND(+f, "ff", (float *a, float *b), floatret(*a + *b));
+ICOMMAND(*f, "ff", (float *a, float *b), floatret(*a * *b));
+ICOMMAND(-f, "ff", (float *a, float *b), floatret(*a - *b));
+ICOMMAND(=, "ii", (int *a, int *b), intret((int)(*a == *b)));
+ICOMMAND(!=, "ii", (int *a, int *b), intret((int)(*a != *b)));
+ICOMMAND(<, "ii", (int *a, int *b), intret((int)(*a < *b)));
+ICOMMAND(>, "ii", (int *a, int *b), intret((int)(*a > *b)));
+ICOMMAND(<=, "ii", (int *a, int *b), intret((int)(*a <= *b)));
+ICOMMAND(>=, "ii", (int *a, int *b), intret((int)(*a >= *b)));
+ICOMMAND(=f, "ff", (float *a, float *b), intret((int)(*a == *b)));
+ICOMMAND(!=f, "ff", (float *a, float *b), intret((int)(*a != *b)));
+ICOMMAND(<f, "ff", (float *a, float *b), intret((int)(*a < *b)));
+ICOMMAND(>f, "ff", (float *a, float *b), intret((int)(*a > *b)));
+ICOMMAND(<=f, "ff", (float *a, float *b), intret((int)(*a <= *b)));
+ICOMMAND(>=f, "ff", (float *a, float *b), intret((int)(*a >= *b)));
+ICOMMAND(^, "ii", (int *a, int *b), intret(*a ^ *b));
+ICOMMAND(!, "i", (int *a), intret(*a == 0));
+ICOMMAND(&, "ii", (int *a, int *b), intret(*a & *b));
+ICOMMAND(|, "ii", (int *a, int *b), intret(*a | *b));
+ICOMMAND(~, "i", (int *a), intret(~*a));
+ICOMMAND(^~, "ii", (int *a, int *b), intret(*a ^ ~*b));
+ICOMMAND(&~, "ii", (int *a, int *b), intret(*a & ~*b));
+ICOMMAND(|~, "ii", (int *a, int *b), intret(*a | ~*b));
+ICOMMAND(<<, "ii", (int *a, int *b), intret(*a << *b));
+ICOMMAND(>>, "ii", (int *a, int *b), intret(*a >> *b));
+ICOMMAND(&&, "ss", (char *a, char *b), intret(execute(a)!=0 && execute(b)!=0));
+ICOMMAND(||, "ss", (char *a, char *b), intret(execute(a)!=0 || execute(b)!=0));
 
-void rndn(int *a)          { intret(*a>0 ? rnd(*a) : 0); }  COMMANDN(rnd, rndn, "i");
+ICOMMAND(div, "ii", (int *a, int *b), intret(*b ? *a / *b : 0));
+ICOMMAND(mod, "ii", (int *a, int *b), intret(*b ? *a % *b : 0));
+ICOMMAND(divf, "ff", (float *a, float *b), floatret(*b ? *a / *b : 0));
+ICOMMAND(modf, "ff", (float *a, float *b), floatret(*b ? fmod(*a, *b) : 0));
+ICOMMAND(min, "ii", (int *a, int *b), intret(min(*a, *b)));
+ICOMMAND(max, "ii", (int *a, int *b), intret(max(*a, *b)));
+ICOMMAND(minf, "ff", (float *a, float *b), floatret(min(*a, *b)));
+ICOMMAND(maxf, "ff", (float *a, float *b), floatret(max(*a, *b)));
 
-void strcmpa(char *a, char *b) { intret(strcmp(a,b)==0); }  COMMANDN(strcmp, strcmpa, "ss");
-
-ICOMMAND(echo, "C", (char *s), conoutf(CON_ECHO, "\f1%s", s));
-
-void strstra(char *a, char *b) { char *s = strstr(a, b); intret(s ? s-a : -1); } COMMANDN(strstr, strstra, "ss");
+ICOMMAND(rnd, "ii", (int *a, int *b), intret(*a - *b > 0 ? rnd(*a - *b) + *b : *b));
+ICOMMAND(strcmp, "ss", (char *a, char *b), intret(strcmp(a,b)==0));
+ICOMMAND(echo, "C", (char *s), conoutf("\f1%s", s));
+ICOMMAND(strstr, "ss", (char *a, char *b), { char *s = strstr(a, b); intret(s ? s-a : -1); });
+ICOMMAND(strlen, "s", (char *s), intret(strlen(s)));
 
 char *strreplace(const char *s, const char *oldval, const char *newval)
 {
@@ -788,13 +1014,14 @@ char *strreplace(const char *s, const char *oldval, const char *newval)
     }
 }
 
-void strreplacea(char *s, char *o, char *n) { commandret = strreplace(s, o, n); } COMMANDN(strreplace, strreplacea, "sss");
+ICOMMAND(strreplace, "sss", (char *s, char *o, char *n), commandret = strreplace(s, o, n));
 
+#ifndef STANDALONE
 struct sleepcmd
 {
     int millis;
     char *command;
-    bool override;
+    bool override, persist;
 };
 vector<sleepcmd> sleepcmds;
 
@@ -804,21 +1031,29 @@ void addsleep(int *msec, char *cmd)
     s.millis = *msec+lastmillis;
     s.command = newstring(cmd);
     s.override = overrideidents;
+    s.persist = persistidents;
 }
 
 COMMANDN(sleep, addsleep, "is");
 
 void checksleep(int millis)
 {
+    if(sleepcmds.empty()) return;
     loopv(sleepcmds)
     {
         sleepcmd &s = sleepcmds[i];
         if(s.millis && millis>s.millis)
         {
             char *cmd = s.command; // execute might create more sleep commands
+            s.command = NULL;
+            bool waspersisting = persistidents, wasoverriding = overrideidents;
+            persistidents = s.persist;
+            overrideidents = s.override;
             execute(cmd);
+            persistidents = waspersisting;
+            overrideidents = wasoverriding;
             delete[] cmd;
-            sleepcmds.remove(i--);
+            if(sleepcmds.inrange(i) && !sleepcmds[i].command) sleepcmds.remove(i--);
         }
     }
 }
@@ -826,7 +1061,7 @@ void checksleep(int millis)
 void clearsleep(bool clearoverrides)
 {
     int len = 0;
-    loopv(sleepcmds) 
+    loopv(sleepcmds) if(sleepcmds[i].command) 
     {
         if(clearoverrides && !sleepcmds[i].override) sleepcmds[len++] = sleepcmds[i];
         else delete[] sleepcmds[i].command;
@@ -840,4 +1075,5 @@ void clearsleep_(int *clearoverrides)
 }
 
 COMMANDN(clearsleep, clearsleep_, "i");
+#endif
 
