@@ -1,37 +1,36 @@
 // console.cpp: the console buffer, its display, and command line control
 
-#include "pch.h"
 #include "engine.h"
 
 struct cline { char *line; int type, outtime; };
 vector<cline> conlines;
 
-bool saycommandon = false;
+int commandmillis = -1;
 string commandbuf;
 char *commandaction = NULL, *commandprompt = NULL;
 int commandpos = -1;
 
 VARFP(maxcon, 10, 200, 1000, { while(conlines.length() > maxcon) delete[] conlines.pop().line; });
 
+#define CONSTRLEN 512
+
 void conline(int type, const char *sf)        // add a line to the console buffer
 {
     cline cl;
-    cl.line = conlines.length()>maxcon ? conlines.pop().line : newstringbuf("");   // constrain the buffer size
+    cl.line = conlines.length()>maxcon ? conlines.pop().line : newstring("", CONSTRLEN-1);   // constrain the buffer size
     cl.type = type;
     cl.outtime = totalmillis;                       // for how long to keep line on screen
     conlines.insert(0, cl);
-    s_strcpy(cl.line, sf);
+    copystring(cl.line, sf, CONSTRLEN);
 }
 
-#define CONSPAD (FONTH/3)
-
 void conoutfv(int type, const char *fmt, va_list args)
 {
-    string sf, sp;
-    formatstring(sf, fmt, args);
-    filtertext(sp, sf);
-    puts(sp);
-    conline(type, sf);
+    static char buf[CONSTRLEN];
+    vformatstring(buf, fmt, args, sizeof(buf));
+    conline(type, buf);
+    filtertext(buf, buf);
+    puts(buf);
 }
 
 void conoutf(const char *fmt, ...)
@@ -56,17 +55,19 @@ COMMAND(toggleconsole, "");
 
 int rendercommand(int x, int y, int w)
 {
-    if(!saycommandon) return 0;
+    if(commandmillis < 0) return 0;
 
-    s_sprintfd(s)("%s %s", commandprompt ? commandprompt : ">", commandbuf);
+    defformatstring(s)("%s %s", commandprompt ? commandprompt : ">", commandbuf);
     int width, height;
     text_bounds(s, width, height, w);
-    y-= height-FONTH;
+    y -= height;
     draw_text(s, x, y, 0xFF, 0xFF, 0xFF, 0xFF, (commandpos>=0) ? (commandpos+1+(commandprompt?strlen(commandprompt):1)) : strlen(s), w);
     return height;
 }
 
 void blendbox(int x1, int y1, int x2, int y2, bool border)
+;
+#if 0
 {
     notextureshader->set();
 
@@ -98,56 +99,54 @@ void blendbox(int x1, int y1, int x2, int y2, bool border)
 
     defaultshader->set();
 }
+#endif
 
 VARP(consize, 0, 5, 100);
+VARP(miniconsize, 0, 5, 100);
+VARP(miniconwidth, 0, 40, 100);
 VARP(confade, 0, 30, 60);
+VARP(miniconfade, 0, 30, 60);
 VARP(fullconsize, 0, 75, 100);
-VARP(confilter, 0, 0xFFFFFF, 0xFFFFFF);
-VARP(fullconfilter, 0, 0xFFFFFF, 0xFFFFFF);
+HVARP(confilter, 0, 0x7FFFFFF, 0x7FFFFFF);
+HVARP(fullconfilter, 0, 0x7FFFFFF, 0x7FFFFFF);
+HVARP(miniconfilter, 0, 0, 0x7FFFFFF);
 
-int conskip = 0;
+int conskip = 0, miniconskip = 0;
 
-void setconskip(int *n)
+void setconskip(int &skip, int filter, int n)
 {
-    int filter = fullconsole ? fullconfilter : confilter,
-        skipped = abs(*n),
-        dir = *n < 0 ? -1 : 1;
-    conskip = clamp(conskip, 0, conlines.length()-1);
-    while(skipped)
+    int offset = abs(n), dir = n < 0 ? -1 : 1;
+    skip = clamp(skip, 0, conlines.length()-1);
+    while(offset)
     {
-        conskip += dir;
-        if(!conlines.inrange(conskip))
+        skip += dir;
+        if(!conlines.inrange(skip))
         {
-            conskip = clamp(conskip, 0, conlines.length()-1);
+            skip = clamp(skip, 0, conlines.length()-1);
             return;
         }
-        if(conlines[conskip].type&filter) --skipped;
+        if(conlines[skip].type&filter) --offset;
     }
 }
 
-COMMANDN(conskip, setconskip, "i");
+ICOMMAND(conskip, "i", (int *n), setconskip(conskip, fullconsole ? fullconfilter : confilter, *n));
+ICOMMAND(miniconskip, "i", (int *n), setconskip(miniconskip, miniconfilter, *n));
 
-int renderconsole(int w, int h)                   // render buffer taking into account time & scrolling
+int drawconlines(int conskip, int confade, int conwidth, int conheight, int conoff, int filter, int y = 0, int dir = 1)
 {
-    int conheight = min(fullconsole ? ((h*3*fullconsize/100)/FONTH)*FONTH : (FONTH*consize), h*3 - 2*CONSPAD - 2*FONTH/3),
-        conwidth = w*3 - 2*CONSPAD - 2*FONTH/3,
-        filter = fullconsole ? fullconfilter : confilter;
-    
-    if(fullconsole) blendbox(CONSPAD, CONSPAD, conwidth+CONSPAD+2*FONTH/3, conheight+CONSPAD+2*FONTH/3, true);
-    
     int numl = conlines.length(), offset = min(conskip, numl);
-    
-    if(!fullconsole && confade)
+
+    if(confade)
     {
-        if(!conskip) 
+        if(!conskip)
         {
             numl = 0;
             loopvrev(conlines) if(totalmillis-conlines[i].outtime < confade*1000) { numl = i+1; break; }
-        } 
+        }
         else offset--;
     }
-   
-    int y = 0;
+
+    int totalheight = 0;
     loopi(numl) //determine visible height
     {
         // shuffle backwards to fill if necessary
@@ -156,21 +155,38 @@ int renderconsole(int w, int h)                   // render buffer taking into a
         char *line = conlines[idx].line;
         int width, height;
         text_bounds(line, width, height, conwidth);
-        y += height;
-        if(y > conheight) { numl = i; if(offset == idx) ++offset; break; }
+        if(totalheight + height > conheight) { numl = i; if(offset == idx) ++offset; break; }
+        totalheight += height;
     }
-    y = CONSPAD+FONTH/3;
+    if(dir > 0) y = conoff;
     loopi(numl)
     {
-        int idx = offset + numl-i-1;
+        int idx = offset + (dir > 0 ? numl-i-1 : i);
         if(!(conlines[idx].type&filter)) continue;
         char *line = conlines[idx].line;
-        draw_text(line, CONSPAD+FONTH/3, y, 0xFF, 0xFF, 0xFF, 0xFF, -1, conwidth);
         int width, height;
         text_bounds(line, width, height, conwidth);
-        y += height;
+        if(dir <= 0) y -= height; 
+        draw_text(line, conoff, y, 0xFF, 0xFF, 0xFF, 0xFF, -1, conwidth);
+        if(dir > 0) y += height;
     }
-    return fullconsole ? (2*CONSPAD+conheight+2*FONTH/3) : (y+CONSPAD+FONTH/3);
+    return y+conoff;
+}
+
+int renderconsole(int w, int h, int abovehud)                   // render buffer taking into account time & scrolling
+{
+    int conpad = fullconsole ? 0 : FONTH/4,
+        conoff = fullconsole ? FONTH : FONTH/3,
+        conheight = min(fullconsole ? ((h*fullconsize/100)/FONTH)*FONTH : FONTH*consize, h - 2*(conpad + conoff)),
+        conwidth = w - 2*(conpad + conoff) - (fullconsole ? 0 : game::clipconsole(w, h));
+    
+    extern void consolebox(int x1, int y1, int x2, int y2);
+    if(fullconsole) consolebox(conpad, conpad, conwidth+conpad+2*conoff, conheight+conpad+2*conoff);
+    
+    int y = drawconlines(conskip, fullconsole ? 0 : confade, conwidth, conheight, conpad+conoff, fullconsole ? fullconfilter : confilter);
+    if(!fullconsole && (miniconsize && miniconwidth))
+        drawconlines(miniconskip, miniconfade, (miniconwidth*(w - 2*(conpad + conoff)))/100, min(FONTH*miniconsize, abovehud - y), conpad+conoff, miniconfilter, abovehud, -1);
+    return fullconsole ? conheight + 2*(conpad + conoff) : y;
 }
 
 // keymap is defined externally in keymap.cfg
@@ -190,50 +206,62 @@ struct keym
     char *actions[NUMACTIONS];
     bool pressed;
 
-    keym() : code(-1), name(NULL), pressed(false) { memset(actions, 0, sizeof(actions)); }
+    keym() : code(-1), name(NULL), pressed(false) { loopi(NUMACTIONS) actions[i] = newstring(""); }
     ~keym() { DELETEA(name); loopi(NUMACTIONS) DELETEA(actions[i]); }
 };
 
-vector<keym> keyms;                                 
+hashtable<int, keym> keyms(128);
 
-void keymap(char *code, char *key)
+void keymap(int *code, char *key)
 {
     if(overrideidents) { conoutf(CON_ERROR, "cannot override keymap %s", code); return; }
-    keym &km = keyms.add();
-    km.code = atoi(code);
+    keym &km = keyms[*code];
+    km.code = *code;
+    DELETEA(km.name);
     km.name = newstring(key);
-    loopi(keym::NUMACTIONS) km.actions[i] = newstring("");
 }
     
-COMMAND(keymap, "ss");
+COMMAND(keymap, "is");
 
 keym *keypressed = NULL;
 char *keyaction = NULL;
 
+const char *getkeyname(int code)
+{
+    keym *km = keyms.access(code);
+    return km ? km->name : NULL;
+}
+
+void searchbinds(char *action, int type)
+{
+    vector<char> names;
+    enumerate(keyms, keym, km,
+    {
+        if(!strcmp(km.actions[type], action))
+        {
+            if(names.length()) names.add(' ');
+            names.put(km.name, strlen(km.name));
+        }
+    });
+    names.add('\0');
+    result(names.getbuf());
+}
+
 keym *findbind(char *key)
 {
-    loopv(keyms) if(!strcasecmp(keyms[i].name, key)) return &keyms[i];
+    enumerate(keyms, keym, km,
+    {
+        if(!strcasecmp(km.name, key)) return &km;
+    });
     return NULL;
 }   
     
-void getbind(char *key)
+void getbind(char *key, int type)
 {
     keym *km = findbind(key);
-    result(km ? km->actions[keym::ACTION_DEFAULT] : "");
+    result(km ? km->actions[type] : "");
 }   
 
-void getspecbind(char *key)
-{
-    keym *km = findbind(key);
-    result(km ? km->actions[keym::ACTION_SPECTATOR] : "");
-}
-
-void geteditbind(char *key)
-{
-    keym *km = findbind(key);
-    result(km ? km->actions[keym::ACTION_EDITING] : "");
-}  
-
 void bindkey(char *key, char *action, int state, const char *cmd)
 {
     if(overrideidents) { conoutf(CON_ERROR, "cannot override %s \"%s\"", cmd, key); return; }
@@ -241,42 +269,38 @@ void bindkey(char *key, char *action, int state, const char *cmd)
     if(!km) { conoutf(CON_ERROR, "unknown key \"%s\"", key); return; }
     char *&binding = km->actions[state];
     if(!keypressed || keyaction!=binding) delete[] binding;
-    binding = newstring(action);
+    // trim white-space to make searchbinds more reliable
+    while(isspace(*action)) action++;
+    int len = strlen(action);
+    while(len>0 && isspace(action[len-1])) len--;
+    binding = newstring(action, len);
 }
 
-void bindnorm(char *key, char *action) { bindkey(key, action, keym::ACTION_DEFAULT, "bind"); }
-void bindspec(char *key, char *action) { bindkey(key, action, keym::ACTION_SPECTATOR, "specbind"); }
-void bindedit(char *key, char *action) { bindkey(key, action, keym::ACTION_EDITING, "editbind"); }
+ICOMMAND(bind,     "ss", (char *key, char *action), bindkey(key, action, keym::ACTION_DEFAULT, "bind"));
+ICOMMAND(specbind, "ss", (char *key, char *action), bindkey(key, action, keym::ACTION_SPECTATOR, "specbind"));
+ICOMMAND(editbind, "ss", (char *key, char *action), bindkey(key, action, keym::ACTION_EDITING, "editbind"));
+ICOMMAND(getbind,     "s", (char *key), getbind(key, keym::ACTION_DEFAULT));
+ICOMMAND(getspecbind, "s", (char *key), getbind(key, keym::ACTION_SPECTATOR));
+ICOMMAND(geteditbind, "s", (char *key), getbind(key, keym::ACTION_EDITING));
+ICOMMAND(searchbinds,     "s", (char *action), searchbinds(action, keym::ACTION_DEFAULT));
+ICOMMAND(searchspecbinds, "s", (char *action), searchbinds(action, keym::ACTION_SPECTATOR));
+ICOMMAND(searcheditbinds, "s", (char *action), searchbinds(action, keym::ACTION_EDITING));
 
-COMMANDN(bind,     bindnorm, "ss");
-COMMANDN(specbind, bindspec, "ss");
-COMMANDN(editbind, bindedit, "ss");
-COMMAND(getbind, "s");
-COMMAND(getspecbind, "s");
-COMMAND(geteditbind, "s");
-
-void saycommand(char *init)                         // turns input to the command line on or off
+void inputcommand(char *init, char *action = NULL, char *prompt = NULL) // turns input to the command line on or off
 {
-    SDL_EnableUNICODE(saycommandon = (init!=NULL));
-    if(!editmode) keyrepeat(saycommandon);
-    s_strcpy(commandbuf, init ? init : "");
+    commandmillis = init ? totalmillis : -1;
+    SDL_EnableUNICODE(commandmillis >= 0 ? 1 : 0);
+    if(!editmode) keyrepeat(commandmillis >= 0);
+    copystring(commandbuf, init ? init : "");
     DELETEA(commandaction);
     DELETEA(commandprompt);
     commandpos = -1;
+    if(action && action[0]) commandaction = newstring(action);
+    if(prompt && prompt[0]) commandprompt = newstring(prompt);
 }
 
-void inputcommand(char *init, char *action, char *prompt)
-{
-    saycommand(init);
-    if(action[0]) commandaction = newstring(action);
-    if(prompt[0]) commandprompt = newstring(prompt);
-}
-
-void mapmsg(char *s) { s_strncpy(hdr.maptitle, s, 128); }
-
-COMMAND(saycommand, "C");
+ICOMMAND(saycommand, "C", (char *init), inputcommand(init));
 COMMAND(inputcommand, "sss");
-COMMAND(mapmsg, "s");
 
 #if !defined(WIN32) && !defined(__APPLE__)
 #include <X11/Xlib.h>
@@ -289,7 +313,7 @@ void pasteconsole()
     if(!IsClipboardFormatAvailable(CF_TEXT)) return; 
     if(!OpenClipboard(NULL)) return;
     char *cb = (char *)GlobalLock(GetClipboardData(CF_TEXT));
-    s_strcat(commandbuf, cb);
+    concatstring(commandbuf, cb);
     GlobalUnlock(cb);
     CloseClipboard();
     #elif defined(__APPLE__)
@@ -334,7 +358,7 @@ struct hline
 
     void restore()
     {
-        s_strcpy(commandbuf, buf);
+        copystring(commandbuf, buf);
         if(commandpos >= (int)strlen(commandbuf)) commandpos = -1;
         DELETEA(commandaction);
         DELETEA(commandprompt);
@@ -364,7 +388,7 @@ struct hline
             execute(action);
         }
         else if(buf[0]=='/') execute(buf+1);
-        else cc->toserver(buf);
+        else game::toserver(buf);
     }
 };
 vector<hline *> history;
@@ -372,7 +396,7 @@ int histpos = 0;
 
 void history_(int *n)
 {
-    static bool inhistory = true;
+    static bool inhistory = false;
     if(!inhistory && history.inrange(*n))
     {
         inhistory = true;
@@ -420,7 +444,12 @@ void execbind(keym &k, bool isdown)
     }
     if(isdown)
     {
-        int state = editmode ? keym::ACTION_EDITING : (player->state==CS_SPECTATOR ? keym::ACTION_SPECTATOR : keym::ACTION_DEFAULT);
+        int state = keym::ACTION_DEFAULT;
+        if(!mainmenu)
+        {
+            if(editmode) state = keym::ACTION_EDITING;
+            else if(player->state==CS_SPECTATOR) state = keym::ACTION_SPECTATOR;
+        }
         char *&action = k.actions[state][0] ? k.actions[state] : k.actions[keym::ACTION_DEFAULT];
         keyaction = action;
         keypressed = &k;
@@ -536,13 +565,13 @@ void consolekey(int code, bool isdown, int cooked)
                 else h = history.last();
             }
             histpos = history.length();
-            saycommand(NULL);
+            inputcommand(NULL);
             if(h) h->run();
         }
         else if(code==SDLK_ESCAPE)
         {
             histpos = history.length();
-            saycommand(NULL);
+            inputcommand(NULL);
         }
     }
 }
@@ -551,27 +580,38 @@ extern bool menukey(int code, bool isdown, int cooked);
 
 void keypress(int code, bool isdown, int cooked)
 {
-    keym *haskey = NULL;
-    loopv(keyms) if(keyms[i].code==code) { haskey = &keyms[i]; break; }        
+    keym *haskey = keyms.access(code);
     if(haskey && haskey->pressed) execbind(*haskey, isdown); // allow pressed keys to release
     else if(!menukey(code, isdown, cooked)) // 3D GUI mouse button intercept   
     {
-        if(saycommandon) consolekey(code, isdown, cooked);
+        if(commandmillis >= 0) consolekey(code, isdown, cooked);
         else if(haskey) execbind(*haskey, isdown);
     }
 }
 
 void clear_console()
 {
-    keyms.setsize(0);
+    keyms.clear();
 }
 
-void writebinds(FILE *f)
+static int sortbinds(keym **x, keym **y)
 {
-    loopv(keyms) loopj(3)
+    return strcmp((*x)->name, (*y)->name);
+}
+
+void writebinds(stream *f)
+{
+    static const char *cmds[3] = { "bind", "specbind", "editbind" };
+    vector<keym *> binds;
+    enumerate(keyms, keym, km, binds.add(&km));
+    binds.sort(sortbinds);
+    loopj(3)
     {
-        static const char *cmds[3] = { "bind", "specbind", "editbind" };
-        if(*keyms[i].actions[j]) fprintf(f, "%s \"%s\" [%s]\n", cmds[j], keyms[i].name, keyms[i].actions[j]);
+        loopv(binds)
+        {
+            keym &km = *binds[i];
+            if(*km.actions[j]) f->printf("%s \"%s\" [%s]\n", cmds[j], km.name, km.actions[j]);
+        }
     }
 }
 
@@ -593,9 +633,22 @@ struct filesval
     int type;
     char *dir, *ext;
     vector<char *> files;
+    int millis;
     
-    filesval(int type, const char *dir, const char *ext) : type(type), dir(newstring(dir)), ext(ext && ext[0] ? newstring(ext) : NULL) {}
-    ~filesval() { DELETEA(dir); DELETEA(ext); loopv(files) DELETEA(files[i]); files.setsize(0); }
+    filesval(int type, const char *dir, const char *ext) : type(type), dir(newstring(dir)), ext(ext && ext[0] ? newstring(ext) : NULL), millis(-1) {}
+    ~filesval() { DELETEA(dir); DELETEA(ext); files.deletecontentsa(); }
+
+    static int comparefiles(char **x, char **y) { return strcmp(*x, *y); }
+
+    void update()
+    {
+        if(type!=FILES_DIR || millis >= commandmillis) return;
+        files.deletecontentsa();        
+        listfiles(dir, ext, files);
+        files.sort(comparefiles); 
+        loopv(files) if(i && !strcmp(files[i], files[i-1])) delete[] files.remove(i--);
+        millis = totalmillis;
+    }
 };
 
 static inline bool htcmp(const fileskey &x, const fileskey &y)
@@ -672,9 +725,9 @@ void complete(char *s)
     if(*s!='/')
     {
         string t;
-        s_strcpy(t, s);
-        s_strcpy(s, "/");
-        s_strcat(s, t);
+        copystring(t, s);
+        copystring(s, "/");
+        concatstring(s, t);
     }
     if(!s[1]) return;
     if(!completesize) { completesize = (int)strlen(s)-1; lastcomplete[0] = '\0'; }
@@ -686,7 +739,7 @@ void complete(char *s)
         if(end)
         {
             string command;
-            s_strncpy(command, s+1, min(size_t(end-s), sizeof(command)));
+            copystring(command, s+1, min(size_t(end-s), sizeof(command)));
             filesval **hasfiles = completions.access(command);
             if(hasfiles) f = *hasfiles;
         }
@@ -694,13 +747,13 @@ void complete(char *s)
 
     const char *nextcomplete = NULL;
     string prefix;
-    s_strcpy(prefix, "/");
+    copystring(prefix, "/");
     if(f) // complete using filenames
     {
         int commandsize = strchr(s, ' ')+1-s;
-        s_strncpy(prefix, s, min(size_t(commandsize+1), sizeof(prefix)));
-        if(f->type==FILES_DIR && f->files.empty()) listfiles(f->dir, f->ext, f->files);
-        loopi(f->files.length())
+        copystring(prefix, s, min(size_t(commandsize+1), sizeof(prefix)));
+        f->update();
+        loopv(f->files)
         {
             if(strncmp(f->files[i], s+commandsize, completesize+1-commandsize)==0 &&
                strcmp(f->files[i], lastcomplete) > 0 && (!nextcomplete || strcmp(f->files[i], nextcomplete) < 0))
@@ -709,7 +762,6 @@ void complete(char *s)
     }
     else // complete using command names
     {
-        extern hashtable<const char *, ident> *idents;
         enumerate(*idents, ident, id,
             if(strncmp(id.name, s+1, completesize)==0 &&
                strcmp(id.name, lastcomplete) > 0 && (!nextcomplete || strcmp(id.name, nextcomplete) < 0))
@@ -718,19 +770,29 @@ void complete(char *s)
     }
     if(nextcomplete)
     {
-        s_strcpy(s, prefix);
-        s_strcat(s, nextcomplete);
-        s_strcpy(lastcomplete, nextcomplete);
+        copystring(s, prefix);
+        concatstring(s, nextcomplete);
+        copystring(lastcomplete, nextcomplete);
     }
     else lastcomplete[0] = '\0';
 }
 
-void writecompletions(FILE *f)
+static int sortcompletions(char **x, char **y)
 {
-    enumeratekt(completions, char *, k, filesval *, v,
-        if(!v) continue;
-        if(v->type==FILES_LIST) fprintf(f, "listcomplete \"%s\" [%s]\n", k, v->dir);
-        else fprintf(f, "complete \"%s\" \"%s\" \"%s\"\n", k, v->dir, v->ext ? v->ext : "*");
-    );
+    return strcmp(*x, *y);
+}
+
+void writecompletions(stream *f)
+{
+    vector<char *> cmds;
+    enumeratekt(completions, char *, k, filesval *, v, { if(v) cmds.add(k); });
+    cmds.sort(sortcompletions);
+    loopv(cmds)
+    {
+        char *k = cmds[i];
+        filesval *v = completions[k];
+        if(v->type==FILES_LIST) f->printf("listcomplete \"%s\" [%s]\n", k, v->dir);
+        else f->printf("complete \"%s\" \"%s\" \"%s\"\n", k, v->dir, v->ext ? v->ext : "*");
+    }
 }
 
