@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 
 struct decalvert
@@ -22,7 +21,8 @@ enum
     DF_ROTATE     = 1<<1,
     DF_INVMOD     = 1<<2,
     DF_OVERBRIGHT = 1<<3,
-    DF_ADD        = 1<<4
+    DF_ADD        = 1<<4,
+    DF_SATURATE   = 1<<5
 };
 
 VARFP(maxdecaltris, 1, 1024, 16384, initdecals());
@@ -63,7 +63,7 @@ struct decalrenderer
         }
         decals = new decalinfo[tris];
         maxdecals = tris;
-        tex = textureload(texname);
+        tex = textureload(texname, 3);
         maxverts = tris*3 + 3;
         availverts = maxverts - 3; 
         verts = new decalvert[maxverts];
@@ -243,9 +243,7 @@ struct decalrenderer
             if(renderpath!=R_FIXEDFUNCTION)
             {
                 glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR); 
-                static Shader *overbrightdecalshader = NULL;
-                if(!overbrightdecalshader) overbrightdecalshader = lookupshaderbyname("overbrightdecal");
-                overbrightdecalshader->set();
+                SETSHADER(overbrightdecal);
             }
             else if(hasTE)
             {
@@ -259,7 +257,13 @@ struct decalrenderer
             if(flags&DF_INVMOD) glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
             else if(flags&DF_ADD) glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
             else glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-            foggedshader->set();
+
+            if(flags&DF_SATURATE)
+            {
+                if(renderpath!=R_FIXEDFUNCTION) SETSHADER(saturatedecal);
+                else if(hasTE) setuptmu(0, "C * T x 2");
+            }
+            else foggedshader->set();
         }
 
         glBindTexture(GL_TEXTURE_2D, tex->id);
@@ -278,7 +282,7 @@ struct decalrenderer
         xtravertsva += count;
 
         if(flags&(DF_ADD|DF_INVMOD|DF_OVERBRIGHT)) glFogfv(GL_FOG_COLOR, oldfogc);
-        if(flags&DF_OVERBRIGHT && hasTE) resettmu(0);
+        if(flags&(DF_OVERBRIGHT|DF_SATURATE) && hasTE) resettmu(0);
     }
 
     decalinfo &newdecal()
@@ -342,7 +346,7 @@ struct decalrenderer
         }
 
         ushort dstart = endvert;
-        gendecaltris(worldroot, ivec(0, 0, 0), hdr.worldsize>>1);
+        gendecaltris(worldroot, ivec(0, 0, 0), worldsize>>1);
         if(dbgdec)
         {
             int nverts = endvert < dstart ? endvert + maxverts - dstart : endvert - dstart;
@@ -506,13 +510,13 @@ void cleardecals()
 
 VARNP(decals, showdecals, 0, 1, 1);
 
-void renderdecals(int time)
+void renderdecals(bool mainpass)
 {
     bool rendered = false;
     loopi(sizeof(decals)/sizeof(decals[0]))
     {
         decalrenderer &d = decals[i];
-        if(time)
+        if(mainpass)
         {
             d.clearfadeddecals();
             d.fadeindecals();
