@@ -1,11 +1,9 @@
-#include "cube.h"
-#include "iengine.h"
-#include "igame.h"
+#ifndef __ENGINE_H__
+#define __ENGINE_H__
 
-extern igameclient     *cl;
-extern igameserver     *sv;
-extern iclientcom      *cc;
-extern icliententities *et;
+#include "cube.h"
+
+#ifndef STANDALONE
 
 #include "world.h"
 #include "octa.h"
@@ -53,12 +51,25 @@ extern PFNGLFRAMEBUFFERTEXTURE2DEXTPROC    glFramebufferTexture2D_;
 extern PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbuffer_;
 extern PFNGLGENERATEMIPMAPEXTPROC          glGenerateMipmap_;
 
+// GL_EXT_framebuffer_blit
+#ifndef GL_EXT_framebuffer_blit
+#define GL_READ_FRAMEBUFFER_EXT           0x8CA8
+#define GL_DRAW_FRAMEBUFFER_EXT           0x8CA9
+#define GL_DRAW_FRAMEBUFFER_BINDING_EXT   0x8CA6
+#define GL_READ_FRAMEBUFFER_BINDING_EXT   0x8CAA
+typedef void (APIENTRYP PFNGLBLITFRAMEBUFFEREXTPROC) (GLint srcX0, GLint srcY0, GLint srcX1, GLint srcY1, GLint dstX0, GLint dstY0, GLint dstX1, GLint dstY1, GLbitfield mask, GLenum filter);
+#endif
+extern PFNGLBLITFRAMEBUFFEREXTPROC         glBlitFramebuffer_;
+
 // GL_EXT_draw_range_elements
 extern PFNGLDRAWRANGEELEMENTSEXTPROC glDrawRangeElements_;
 
 // GL_EXT_blend_minmax
 extern PFNGLBLENDEQUATIONEXTPROC glBlendEquation_;
 
+// GL_EXT_blend_color
+extern PFNGLBLENDCOLOREXTPROC glBlendColor_;
+
 // GL_EXT_multi_draw_arrays
 extern PFNGLMULTIDRAWARRAYSEXTPROC   glMultiDrawArrays_;
 extern PFNGLMULTIDRAWELEMENTSEXTPROC glMultiDrawElements_;
@@ -71,11 +82,21 @@ extern PFNGLMULTIDRAWELEMENTSEXTPROC glMultiDrawElements_;
 #define GL_DEPTH24_STENCIL8_EXT 0x88F0
 #endif
 
+// GL_ARB_texture_compression
+extern PFNGLCOMPRESSEDTEXIMAGE3DARBPROC    glCompressedTexImage3D_;
+extern PFNGLCOMPRESSEDTEXIMAGE2DARBPROC    glCompressedTexImage2D_;
+extern PFNGLCOMPRESSEDTEXIMAGE1DARBPROC    glCompressedTexImage1D_;
+extern PFNGLCOMPRESSEDTEXSUBIMAGE3DARBPROC glCompressedTexSubImage3D_;
+extern PFNGLCOMPRESSEDTEXSUBIMAGE2DARBPROC glCompressedTexSubImage2D_;
+extern PFNGLCOMPRESSEDTEXSUBIMAGE1DARBPROC glCompressedTexSubImage1D_;
+extern PFNGLGETCOMPRESSEDTEXIMAGEARBPROC   glGetCompressedTexImage_;
+
 extern dynent *player;
 extern physent *camera1;                // special ent that acts as camera, same object as player1 in FPS mode
 
-extern header hdr;                      // current map header
-extern int worldscale;
+extern int worldscale, worldsize;
+extern int mapversion;
+extern char *maptitle;
 extern vector<ushort> texmru;
 extern int xtraverts, xtravertsva;
 extern int curtexnum;
@@ -83,11 +104,8 @@ extern const ivec cubecoords[8];
 extern const ushort fv[6][4];
 extern const uchar fvmasks[64];
 extern const uchar faceedgesidx[6][4];
-extern bool inbetweenframes;
+extern bool inbetweenframes, renderedframe;
 
-extern int curtime;                     // current frame time
-extern int lastmillis;                  // last time
-extern int totalmillis;                 // total elapsed time
 extern SDL_Surface *screen;
 extern int zpass, glowpass;
 
@@ -116,18 +134,20 @@ struct font
 extern font *curfont;
 
 // texture
-extern int hwtexsize, hwcubetexsize, hwmaxaniso;
+extern int hwtexsize, hwcubetexsize, hwmaxaniso, maxtexsize;
 
 extern Texture *textureload(const char *name, int clamp = 0, bool mipit = true, bool msg = true);
+extern int texalign(void *data, int w, int bpp);
+extern void cleanuptexture(Texture *t);
 extern void loadalphamask(Texture *t);
-extern GLuint cubemapfromsky(int size);
-extern Texture *cubemapload(const char *name, bool mipit = true, bool msg = true);
+extern void loadlayermasks();
+extern Texture *cubemapload(const char *name, bool mipit = true, bool msg = true, bool transient = false);
 extern void drawcubemap(int size, const vec &o, float yaw, float pitch, const cubemapside &side);
-extern Slot    &lookuptexture(int tex, bool load = true);
+extern Slot &lookupmaterialslot(int slot, bool load = true);
+extern Slot &lookuptexture(int slot, bool load = true);
 extern void loadshaders();
-extern Shader  *lookupshader(int slot);
-extern void createtexture(int tnum, int w, int h, void *pixels, int clamp, bool mipit, GLenum component = GL_RGB, GLenum target = GL_TEXTURE_2D, bool compress = false, bool filter = true);
-extern void renderfullscreenshader(int w, int h);
+extern void createtexture(int tnum, int w, int h, void *pixels, int clamp, int filter, GLenum component = GL_RGB, GLenum target = GL_TEXTURE_2D, int pw = 0, int ph = 0, int pitch = 0, bool resize = true);
+extern void renderpostfx();
 extern void initenvmaps();
 extern void genenvmaps();
 extern ushort closestenvmap(const vec &o);
@@ -136,10 +156,11 @@ extern GLuint lookupenvmap(ushort emid);
 extern GLuint lookupenvmap(Slot &slot);
 extern bool reloadtexture(Texture &tex);
 extern bool reloadtexture(const char *name);
+extern void setuptexcompress();
 
 // shadowmap
 
-extern int shadowmap;
+extern int shadowmap, shadowmapcasters;
 extern bool shadowmapping;
 
 extern bool isshadowmapcaster(const vec &o, float rad);
@@ -157,8 +178,8 @@ extern void clearpvs();
 extern bool pvsoccluded(const ivec &bborigin, const ivec &bbsize);
 extern bool waterpvsoccluded(int height);
 extern void setviewcell(const vec &p);
-extern void savepvs(gzFile f);
-extern void loadpvs(gzFile f);
+extern void savepvs(stream *f);
+extern void loadpvs(stream *f, int numpvs);
 extern int getnumviewcells();
 
 static inline bool pvsoccluded(const ivec &bborigin, int size)
@@ -167,23 +188,30 @@ static inline bool pvsoccluded(const ivec &bborigin, int size)
 }
 
 // rendergl
-extern bool hasVBO, hasDRE, hasOQ, hasTR, hasFBO, hasDS, hasTF, hasBE, hasCM, hasNP2, hasTC, hasTE, hasMT, hasD3, hasstencil, hasAF, hasVP2, hasVP3, hasPP, hasMDA, hasTE3, hasTE4, hasVP, hasFP, hasGLSL, hasGM, hasNVFB;
+extern bool hasVBO, hasDRE, hasOQ, hasTR, hasFBO, hasDS, hasTF, hasBE, hasBC, hasCM, hasNP2, hasTC, hasTE, hasMT, hasD3, hasAF, hasVP2, hasVP3, hasPP, hasMDA, hasTE3, hasTE4, hasVP, hasFP, hasGLSL, hasGM, hasNVFB, hasSGIDT, hasSGISH, hasDT, hasSH, hasNVPCF, hasRN, hasPBO, hasFBB;
+extern int hasstencil;
 
 extern bool envmapping, renderedgame;
-extern GLfloat mvmatrix[16], projmatrix[16], mvpmatrix[16], invmvmatrix[16];
+extern glmatrixf mvmatrix, projmatrix, mvpmatrix, invmvmatrix, invmvpmatrix;
+extern bvec fogcolor;
 
 extern void gl_checkextensions();
 extern void gl_init(int w, int h, int bpp, int depth, int fsaa);
 extern void cleangl();
-extern void rendergame();
+extern void rendergame(bool mainpass = false);
 extern void invalidatepostfx();
 extern void gl_drawframe(int w, int h);
+extern void gl_drawmainmenu(int w, int h);
 extern void enablepolygonoffset(GLenum type);
 extern void disablepolygonoffset(GLenum type);
+extern void calcspherescissor(const vec &center, float size, float &sx1, float &sy1, float &sx2, float &sy2);
+extern int pushscissor(float sx1, float sy1, float sx2, float sy2);
+extern void popscissor();
 extern void setfogplane(const plane &p, bool flush = false);
 extern void setfogplane(float scale = 0, float z = 0, bool flush = false, float fadescale = 0, float fadeoffset = 0);
+extern void recomputecamera();
 extern void findorientation();
-extern void writecrosshairs(FILE *f);
+extern void writecrosshairs(stream *f);
 
 // renderextras
 extern void render3dbox(vec &o, float tofloor, float toceil, float xradius, float yradius = 0);
@@ -201,7 +229,6 @@ extern void validatec(cube *c, int size);
 extern bool isvalidcube(cube &c);
 extern cube &lookupcube(int tx, int ty, int tz, int tsize = 0);
 extern cube &neighbourcube(int x, int y, int z, int size, int rsize, int orient);
-extern int lookupmaterial(const vec &o);
 extern void newclipplanes(cube &c);
 extern void freeclipplanes(cube &c);
 extern void forcemip(cube &c);
@@ -219,12 +246,14 @@ extern bool flataxisface(cube &c, int orient);
 extern int genclipplane(cube &c, int i, vec *v, plane *clip);
 extern void genclipplanes(cube &c, int x, int y, int z, int size, clipplanes &p);
 extern bool visibleface(cube &c, int orient, int x, int y, int z, int size, uchar mat = MAT_AIR, uchar nmat = MAT_AIR, uchar matmask = MATF_VOLUME);
+extern int visibletris(cube &c, int orient, int x, int y, int z, int size);
 extern int visibleorient(cube &c, int orient);
 extern bool threeplaneintersect(plane &pl1, plane &pl2, plane &pl3, vec &dest);
+extern void genvectorvert(const ivec &p, cube &c, ivec &v);
 extern void freemergeinfo(cube &c);
 extern void genmergedverts(cube &cu, int orient, const ivec &co, int size, const mergeinfo &m, vvec *vv, plane *p = NULL);
 extern int calcmergedsize(int orient, const ivec &co, int size, const mergeinfo &m, const vvec *vv);
-extern void invalidatemerges(cube &c);
+extern void invalidatemerges(cube &c, bool msg);
 extern void calcmerges();
 
 struct cubeface : mergeinfo
@@ -251,12 +280,12 @@ static inline uchar octantrectangleoverlap(const ivec &c, int size, const ivec &
 
 static inline bool insideworld(const vec &o)
 {
-    return o.x>=0 && o.x<hdr.worldsize && o.y>=0 && o.y<hdr.worldsize && o.z>=0 && o.z<hdr.worldsize;
+    return o.x>=0 && o.x<worldsize && o.y>=0 && o.y<worldsize && o.z>=0 && o.z<worldsize;
 }
 
 static inline bool insideworld(const ivec &o)
 {
-    return uint(o.x)<uint(hdr.worldsize) && uint(o.y)<uint(hdr.worldsize) && uint(o.z)<uint(hdr.worldsize);
+    return uint(o.x)<uint(worldsize) && uint(o.y)<uint(worldsize) && uint(o.z)<uint(worldsize);
 }
 
 // ents
@@ -267,13 +296,16 @@ extern void pasteundoents(undoblock *u);
 
 // octaedit
 extern void cancelsel();
-extern void render_texture_panel(int w, int h);
+extern void rendertexturepanel(int w, int h);
 extern void addundo(undoblock *u);
+extern void commitchanges(bool force = false);
+extern void rendereditcursor();
+extern void tryedit();
 
 // octarender
 extern void octarender();
 extern void allchanged(bool load = false);
-extern void vaclearc(cube *c);
+extern void clearvas(cube *c);
 extern vtxarray *newva(int x, int y, int z, int size);
 extern void destroyva(vtxarray *va, bool reparent = true);
 extern bool readva(vtxarray *va, ushort *&edata, uchar *&vdata);
@@ -316,6 +348,7 @@ extern void updatedynlights();
 extern int finddynlights();
 extern void calcdynlightmask(vtxarray *va);
 extern int setdynlights(vtxarray *va, const ivec &vaorigin);
+extern bool getdynlight(int n, vec &o, float &radius, vec &color);
 
 // material
 
@@ -336,15 +369,15 @@ extern int refracting;
 extern bool reflecting, fading, fogging;
 extern float reflectz;
 extern int reflectdist, vertwater, refractfog, waterrefract, waterreflect, waterfade, caustics, waterfallrefract, waterfog, lavafog;
+extern bvec watercolor, waterfallcolor, lavacolor;
 
 extern void cleanreflections();
 extern void queryreflections();
 extern void drawreflections();
 extern void renderwater();
 extern void renderlava(materialsurface &m, Texture *tex, float scale);
-extern void getwatercolour(uchar *wcol);
-extern void getwaterfallcolour(uchar *fcol);
-extern void getlavacolour(uchar *lcol);
+extern void loadcaustics(bool force = false);
+extern void preloadwatershaders(bool force = false);
 
 // glare
 extern bool glaring;
@@ -358,32 +391,32 @@ extern bool depthfxing;
 extern void drawdepthfxtex();
 
 // server
-extern vector<char *> gameargs;
+extern vector<const char *> gameargs;
 
-extern void initserver(bool dedicated);
+extern void initserver(bool listen, bool dedicated);
 extern void cleanupserver();
-extern void serverslice(uint timeout);
+extern void serverslice(bool dedicated, uint timeout);
 
-extern uchar *retrieveservers(uchar *buf, int buflen);
+extern ENetSocket connectmaster();
 extern void localclienttoserver(int chan, ENetPacket *);
 extern void localconnect();
 extern bool serveroption(char *opt);
 
 // serverbrowser
 extern bool resolverwait(const char *name, ENetAddress *address);
-extern int connectwithtimeout(ENetSocket sock, const char *hostname, ENetAddress &address);
-extern void addserver(char *servername);
-extern char *getservername(int n);
+extern int connectwithtimeout(ENetSocket sock, const char *hostname, const ENetAddress &address);
+extern void addserver(const char *name, int port = 0, const char *password = NULL, bool keep = false);
 extern void writeservercfg();
 
 // client
-extern void localdisconnect();
-extern void localservertoclient(int chan, uchar *buf, int len);
-extern void connects(char *servername);
+extern void localdisconnect(bool cleanup = true);
+extern void localservertoclient(int chan, ENetPacket *packet);
+extern void connectserv(const char *servername, int port, const char *serverpassword);
 extern void abortconnect();
 extern void clientkeepalive();
 
 // command
+extern hashtable<const char *, ident> *idents;
 extern bool overrideidents, persistidents;
 
 extern void explodelist(const char *s, vector<char *> &elems);
@@ -395,8 +428,17 @@ extern void checksleep(int millis);
 extern void clearsleep(bool clearoverrides = true);
 
 // console
-extern void writebinds(FILE *f);
-extern void writecompletions(FILE *f);
+extern void keypress(int code, bool isdown, int cooked);
+extern int rendercommand(int x, int y, int w);
+extern int renderconsole(int w, int h, int abovehud);
+extern void conoutf(const char *s, ...);
+extern void conoutf(int type, const char *s, ...);
+extern void resetcomplete();
+extern void complete(char *s);
+const char *getkeyname(int code);
+extern const char *addreleaseaction(const char *s);
+extern void writebinds(stream *f);
+extern void writecompletions(stream *f);
 
 // main
 enum
@@ -405,6 +447,8 @@ enum
     INIT_LOAD,
     INIT_RESET
 };
+extern int initing;
+
 enum
 {
     CHANGE_GFX   = 1<<0,
@@ -414,8 +458,13 @@ extern bool initwarning(const char *desc, int level = INIT_RESET, int type = CHA
 
 extern void pushevent(const SDL_Event &e);
 extern bool interceptkey(int sym);
-extern void computescreen(const char *text = NULL, Texture *t = NULL, const char *overlaytext = NULL);
-extern void show_out_of_renderloop_progress(float bar1, const char *text1, float bar2 = 0, const char *text2 = NULL, GLuint tex = 0);
+
+extern float loadprogress;
+extern void renderbackground(const char *caption = NULL, Texture *mapshot = NULL, const char *mapname = NULL, const char *mapinfo = NULL, bool restore = false, bool force = false);
+extern void renderprogress(float bar, const char *text, GLuint tex = 0, bool background = false);
+
+extern void getfps(int &fps, int &bestdiff, int &worstdiff);
+extern void swapbuffers();
 
 // menu
 extern void menuprocess();
@@ -439,7 +488,7 @@ enum
     TRIG_DISAPPEAR  = 1<<3,
     TRIG_AUTO_RESET = 1<<4,
     TRIG_RUMBLE     = 1<<5,
-    TRIG_LOCKED     = 1<<6,
+    TRIG_LOCKED     = 1<<6
 };
 
 #define NUMTRIGGERTYPES 16
@@ -448,6 +497,8 @@ extern int triggertypes[NUMTRIGGERTYPES];
 
 #define checktriggertype(type, flag) (triggertypes[(type) & (NUMTRIGGERTYPES-1)] & (flag))
 
+extern vector<int> outsideents;
+
 extern void entitiesinoctanodes();
 extern void attachentities();
 extern void freeoctaentities(cube &c);
@@ -464,16 +515,36 @@ extern void loadskin(const char *dir, const char *altdir, Texture *&skin, Textur
 extern mapmodelinfo &getmminfo(int i);
 extern void startmodelquery(occludequery *query);
 extern void endmodelquery();
+extern void preloadmodelshaders();
 
 // renderparticles
 extern void particleinit();
 extern void clearparticles();
-extern void entity_particles();
+extern void clearparticleemitters();
+extern void seedparticles();
+extern void updateparticles();
+extern void renderparticles(bool mainpass = false);
+extern bool printparticles(extentity &e, char *buf);
 
 // decal
 extern void initdecals();
 extern void cleardecals();
-extern void renderdecals(int time);
+extern void renderdecals(bool mainpass = false);
+
+// blob
+
+enum
+{
+    BLOB_STATIC = 0,
+    BLOB_DYNAMIC
+};
+
+extern int showblobs;
+
+extern void initblobs(int type = -1);
+extern void resetblobs();
+extern void renderblob(int type, const vec &o, float radius, float fade = 1);
+extern void flushblobs();
 
 // rendersky
 extern int explicitsky;
@@ -485,17 +556,52 @@ extern bool limitsky();
 // 3dgui
 extern void g3d_render();
 extern bool g3d_windowhit(bool on, bool act);
+
+// menus
+extern int mainmenu;
+
+extern void clearmainmenu();
 extern void g3d_mainmenu();
 
 // sound
-extern void checkmapsounds();
 extern void clearmapsounds();
-extern void updatevol();
+extern void checkmapsounds();
+extern void updatesounds();
 
 extern void initmumble();
 extern void closemumble();
 extern void updatemumble();
 
 // grass
+extern void generategrass();
 extern void rendergrass();
 
+// blendmap
+extern int blendpaintmode;
+
+extern bool setblendmaporigin(const ivec &o, int size);
+extern bool hasblendmap();
+extern uchar lookupblendmap(const vec &pos);
+extern void resetblendmap();
+extern void enlargeblendmap();
+extern void optimizeblendmap();
+extern void stoppaintblendmap();
+extern void trypaintblendmap();
+extern void renderblendbrush(GLuint tex, float x, float y, float w, float h);
+extern void renderblendbrush();
+extern bool loadblendmap(stream *f, int info);
+extern void saveblendmap(stream *f);
+extern uchar shouldsaveblendmap();
+
+// recorder
+
+namespace recorder
+{
+    void stop();
+    void capture();
+}
+
+#endif
+
+#endif
+
