@@ -1,184 +1,3 @@
-// eye space depth texture for soft particles, done at low res then blurred to prevent ugly jaggies
-VARP(depthfxfpscale, 1, 1<<12, 1<<16);
-VARP(depthfxscale, 1, 1<<6, 1<<8);
-VARP(depthfxblend, 1, 16, 64);
-VAR(depthfxmargin, 0, 16, 64);
-VAR(depthfxbias, 0, 1, 64);
-
-extern void cleanupdepthfx();
-VARFP(fpdepthfx, 0, 0, 1, cleanupdepthfx());
-VARFP(depthfxprecision, 0, 0, 1, cleanupdepthfx());
-VARP(depthfxemuprecision, 0, 1, 1);
-VARFP(depthfxsize, 6, 7, 12, cleanupdepthfx());
-VARP(depthfx, 0, 1, 1);
-VARFP(depthfxrect, 0, 0, 1, cleanupdepthfx());
-VARFP(depthfxfilter, 0, 1, 1, cleanupdepthfx());
-VARP(blurdepthfx, 0, 1, 7);
-VARP(blurdepthfxsigma, 1, 50, 200);
-VAR(depthfxscissor, 0, 2, 2);
-VAR(debugdepthfx, 0, 0, 1);
-
-#define MAXDFXRANGES 4
-
-void *depthfxowners[MAXDFXRANGES];
-float depthfxranges[MAXDFXRANGES];
-int numdepthfxranges = 0;
-vec depthfxmin(1e16f, 1e16f, 1e16f), depthfxmax(1e16f, 1e16f, 1e16f);
-
-static struct depthfxtexture : rendertarget
-{
-    const GLenum *colorformats() const
-    {
-        static const GLenum colorfmts[] = { GL_FLOAT_RG16_NV, GL_RGB16F_ARB, GL_RGB16, GL_RGBA, GL_RGBA8, GL_RGB, GL_RGB8, GL_FALSE };
-        return &colorfmts[hasTF && hasFBO ? (fpdepthfx ? (hasNVFB && texrect() && !filter() ? 0 : 1) : (depthfxprecision ? 2 : 3)) : 3];
-    }
-
-    float eyedepth(const vec &p) const
-    {
-        return max(-(p.x*mvmatrix[2] + p.y*mvmatrix[6] + p.z*mvmatrix[10] + mvmatrix[14]), 0.0f);
-    }
-
-    void addscissorvert(const vec &v, float &sx1, float &sy1, float &sx2, float &sy2)
-    {
-        float w = v.x*mvpmatrix[3] + v.y*mvpmatrix[7] + v.z*mvpmatrix[11] + mvpmatrix[15],
-              x = (v.x*mvpmatrix[0] + v.y*mvpmatrix[4] + v.z*mvpmatrix[8] + mvpmatrix[12]) / w,
-              y = (v.x*mvpmatrix[1] + v.y*mvpmatrix[5] + v.z*mvpmatrix[9] + mvpmatrix[13]) / w;
-        sx1 = min(sx1, x);
-        sy1 = min(sy1, y);
-        sx2 = max(sx2, x);
-        sy2 = max(sy2, y);
-    }
-
-    bool addscissorbox(const vec &center, float size)
-    {
-        extern float fovy, aspect;
-        vec e(center.x*mvmatrix[0] + center.y*mvmatrix[4] + center.z*mvmatrix[8] + mvmatrix[12],
-              center.x*mvmatrix[1] + center.y*mvmatrix[5] + center.z*mvmatrix[9] + mvmatrix[13],
-              center.x*mvmatrix[2] + center.y*mvmatrix[6] + center.z*mvmatrix[10] + mvmatrix[14]);
-        float zz = e.z*e.z, xx = e.x*e.x, yy = e.y*e.y, rr = size*size,
-              dx = zz*(xx + zz) - rr*zz, dy = zz*(yy + zz) - rr*zz,
-              focaldist = 1.0f/tan(fovy*0.5f*RAD),
-              left = -1, right = 1, bottom = -1, top = 1;
-        #define CHECKPLANE(c, dir, focaldist, low, high) \
-        do { \
-            float nc = (size*e.c dir drt)/(c##c + zz), \
-                  nz = (size - nc*e.c)/e.z, \
-                  pz = (c##c + zz - rr)/(e.z - nz/nc*e.c); \
-            if(pz < 0) \
-            { \
-                float c = nz*(focaldist)/nc, \
-                      pc = -pz*nz/nc; \
-                if(pc < e.c) low = c; \
-                else if(pc > e.c) high = c; \
-            } \
-        } while(0)
-        if(dx > 0)
-        {
-            float drt = sqrt(dx);
-            CHECKPLANE(x, -, focaldist/aspect, left, right);
-            CHECKPLANE(x, +, focaldist/aspect, left, right);
-        }
-        if(dy > 0)
-        {
-            float drt = sqrt(dy);
-            CHECKPLANE(y, -, focaldist, bottom, top);
-            CHECKPLANE(y, +, focaldist, bottom, top);
-        }
-        return addblurtiles(left, bottom, right, top);
-    }
-
-    bool addscissorbox(const vec &bbmin, const vec &bbmax)
-    {
-        float sx1 = 1, sy1 = 1, sx2 = -1, sy2 = -1;
-        loopi(8)
-        {
-            vec v(i&1 ? bbmax.x : bbmin.x, i&2 ? bbmax.y : bbmin.y, i&4 ? bbmax.z : bbmin.z);
-            addscissorvert(v, sx1, sy1, sx2, sy2);
-        }
-        return addblurtiles(sx1, sy1, sx2, sy2);
-    }
-
-    bool screenview() const { return depthfxrect!=0; }
-    bool texrect() const { return depthfxrect && hasTR; }
-    bool filter() const { return depthfxfilter!=0; }
-    bool highprecision() const { return colorfmt==GL_FLOAT_RG16_NV || colorfmt==GL_RGB16F_ARB || colorfmt==GL_RGB16; }
-    bool emulatehighprecision() const { return depthfxemuprecision && !depthfxfilter; }
-
-    bool shouldrender()
-    {
-        extern void finddepthfxranges();
-        finddepthfxranges();
-        return (numdepthfxranges && scissorx1 < scissorx2 && scissory1 < scissory2) || debugdepthfx;
-    }
-
-    bool dorender()
-    {
-        glClearColor(1, 1, 1, 1);
-        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
-
-        depthfxing = true;
-        refracting = -1;
-
-        extern void renderdepthobstacles(const vec &bbmin, const vec &bbmax, float scale, float *ranges, int numranges);
-        float scale = depthfxscale;
-        float *ranges = depthfxranges;
-        int numranges = numdepthfxranges;
-        if(highprecision())
-        {
-            scale = depthfxfpscale;
-            ranges = NULL;
-            numranges = 0;
-        }
-        else if(emulatehighprecision())
-        {
-            scale = depthfxfpscale;
-            ranges = NULL;
-            numranges = -3;
-        }
-        renderdepthobstacles(depthfxmin, depthfxmax, scale, ranges, numranges);
-
-        refracting = 0;
-        depthfxing = false;
-
-        return numdepthfxranges > 0;
-    }
-
-    void dodebug(int w, int h)
-    {
-        if(numdepthfxranges > 0)
-        {
-            glColor3f(0, 1, 0);
-            debugscissor(w, h, true);
-            glColor3f(0, 0, 1);
-            debugblurtiles(w, h, true);
-            glColor3f(1, 1, 1);
-        }
-    }
-} depthfxtex;
-
-void cleanupdepthfx()
-{
-    depthfxtex.cleanup(true);
-}
-
-void viewdepthfxtex()
-{
-    if(!depthfx) return;
-    depthfxtex.debug();
-}
-
-bool depthfxing = false;
-
-void drawdepthfxtex()
-{
-    if(!depthfx || renderpath==R_FIXEDFUNCTION) return;
-
-    // Apple/ATI bug - fixed-function fog state can force software fallback even when fragment program is enabled
-    glDisable(GL_FOG);
-    depthfxtex.render(1<<depthfxsize, 1<<depthfxsize, blurdepthfx, blurdepthfxsigma/100.0f);
-    glEnable(GL_FOG);
-}
-
 //cache our unit hemisphere
 static GLushort *hemiindices = NULL;
 static vec *hemiverts = NULL;
@@ -262,7 +81,7 @@ static GLuint createexpmodtex(int size, float minval)
     }
     GLuint tex = 0;
     glGenTextures(1, &tex);
-    createtexture(tex, size, size, data, 3, true, GL_ALPHA);
+    createtexture(tex, size, size, data, 3, 2, GL_ALPHA);
     delete[] data;
     return tex;
 }
@@ -400,8 +219,6 @@ static void setupexplosion()
                     if(explosion2d) SETSHADER(explosion2dsoft8rect); else SETSHADER(explosion3dsoft8rect);
                 }
                 else if(explosion2d) SETSHADER(explosion2dsoftrect); else SETSHADER(explosion3dsoftrect);
-
-                setlocalparamf("depthfxview", SHPARAM_VERTEX, 6, 0.5f*depthfxtex.vieww, 0.5f*depthfxtex.viewh);
             }
             else
             {
@@ -410,13 +227,7 @@ static void setupexplosion()
                     if(explosion2d) SETSHADER(explosion2dsoft8); else SETSHADER(explosion3dsoft8);
                 }
                 else if(explosion2d) SETSHADER(explosion2dsoft); else SETSHADER(explosion3dsoft);
-
-                setlocalparamf("depthfxview", SHPARAM_VERTEX, 6, 0.5f*float(depthfxtex.vieww)/depthfxtex.texw, 0.5f*float(depthfxtex.viewh)/depthfxtex.texh);
             }
-
-            glActiveTexture_(GL_TEXTURE2_ARB);
-            glBindTexture(depthfxtex.target, depthfxtex.rendertex);
-            glActiveTexture_(GL_TEXTURE0_ARB);
         }
         else if(explosion2d) SETSHADER(explosion2d); else SETSHADER(explosion3d);
     }
@@ -584,7 +395,7 @@ static const float WOBBLE = 1.25f;
 struct fireballrenderer : listrenderer
 {
     fireballrenderer(int type)
-        : listrenderer("packages/particles/explosion.jpg", type, 0, 0)
+        : listrenderer("packages/particles/explosion.jpg", type)
     {}
 
     void startrender()
@@ -656,6 +467,12 @@ struct fireballrenderer : listrenderer
         return numranges;
     }
 
+    void seedemitter(particleemitter &pe, const vec &o, const vec &d, int fade, float size, int gravity)
+    {
+        pe.maxfade = max(pe.maxfade, fade);
+        pe.extendbb(o, (size+1+pe.ent->attr2)*WOBBLE); 
+    }
+
     void renderpart(listparticle *p, const vec &o, const vec &d, int blend, int ts, uchar *color)
     {
         float pmax = p->val,
@@ -706,41 +523,7 @@ struct fireballrenderer : listrenderer
         {
             setlocalparamf("center", SHPARAM_VERTEX, 0, o.x, o.y, o.z);
             setlocalparamf("animstate", SHPARAM_VERTEX, 1, size, psize, pmax, float(lastmillis));
-            if(!glaring && !reflecting && !refracting && depthfx && depthfxtex.rendertex && numdepthfxranges>0)
-            {
-                float scale = 0, offset = -1, texscale = 0;
-                if(!depthfxtex.highprecision())
-                {
-                    float select[4] = { 0, 0, 0, 0 };
-                    if(!depthfxtex.emulatehighprecision())
-                    {
-                        loopi(numdepthfxranges) if(depthfxowners[i]==p)
-                        {
-                            select[i] = float(depthfxscale)/depthfxblend;
-                            scale = 1.0f/depthfxblend;
-                            offset = -float(depthfxranges[i] - depthfxbias)/depthfxblend;
-                            break;
-                        }
-                    }
-                    else if(2*(p->size + pmax)*WOBBLE >= depthfxblend)
-                    {
-                        select[0] = float(depthfxfpscale)/depthfxblend;
-                        select[1] = select[0]/256;
-                        select[2] = select[1]/256;
-                        scale = 1.0f/depthfxblend;
-                        offset = 0;
-                    }
-                    setlocalparamfv("depthfxselect", SHPARAM_PIXEL, 6, select);
-                }
-                else if(2*(p->size + pmax)*WOBBLE >= depthfxblend)
-                {
-                    scale = 1.0f/depthfxblend;
-                    offset = 0;
-                    texscale = float(depthfxfpscale)/depthfxblend;
-                }
-                setlocalparamf("depthfxparams", SHPARAM_VERTEX, 5, scale, offset, texscale, inside ? blend/(2*255.0f) : 0);
-                setlocalparamf("depthfxparams", SHPARAM_PIXEL, 5, scale, offset, texscale, inside ? blend/(2*255.0f) : 0);
-            }
+            binddepthfxparams(depthfxblend, inside ? blend/(2*255.0f) : 0, 2*(p->size + pmax)*WOBBLE >= depthfxblend, p);
         }
 
         glRotatef(lastmillis/7.0f, -rotdir.x, rotdir.y, -rotdir.z);
@@ -752,17 +535,3 @@ struct fireballrenderer : listrenderer
 };
 static fireballrenderer fireballs(PT_FIREBALL|PT_GLARE), noglarefireballs(PT_FIREBALL);
 
-void finddepthfxranges()
-{
-    depthfxmin = vec(1e16f, 1e16f, 1e16f);
-    depthfxmax = vec(0, 0, 0);
-    numdepthfxranges = fireballs.finddepthfxranges(depthfxowners, depthfxranges, MAXDFXRANGES, depthfxmin, depthfxmax);
-    loopk(3)
-    {
-        depthfxmin[k] -= depthfxmargin;
-        depthfxmax[k] += depthfxmargin;
-    }
-
-    if(depthfxscissor<2 && numdepthfxranges>0) depthfxtex.addscissorbox(depthfxmin, depthfxmax);
-}
-
