@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 #include "rendertarget.h"
 
@@ -39,14 +38,12 @@ void drawglaretex()
     glaretex.render(1<<glaresize, 1<<glaresize, blurglare, blurglaresigma/100.0f);
 }
 
-FVARP(glarescale, 1);
+FVARP(glarescale, 0, 1, 8);
 
 void addglare()
 {
     if(!glare || renderpath==R_FIXEDFUNCTION) return;
 
-    glDisable(GL_DEPTH_TEST);
-    
     glEnable(GL_BLEND);
     glBlendFunc(GL_ONE, GL_ONE);
 
@@ -66,7 +63,5 @@ void addglare()
     glEnd();
 
     glDisable(GL_BLEND);
-
-    glEnable(GL_DEPTH_TEST);
 }
      
