@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 
 struct QuadNode
@@ -134,7 +133,7 @@ struct material
     {"glass", MAT_GLASS},
     {"noclip", MAT_NOCLIP},
     {"lava", MAT_LAVA},
-    {"aiclip", MAT_AICLIP},
+    {"gameclip", MAT_GAMECLIP},
     {"death", MAT_DEATH}
 };
 
@@ -305,7 +304,7 @@ int optimizematsurfs(materialsurface *matbuf, int matsurfs)
          }
          else if(cur-start>=4)
          {
-            QuadNode vmats(0, 0, hdr.worldsize);
+            QuadNode vmats(0, 0, worldsize);
             loopi(cur-start) vmats.insert(start[i].o[C[dim]], start[i].o[R[dim]], start[i].csize);
             vmats.genmatsurfs(start->material, start->orient, start->o[dim], matbuf);
          }
@@ -418,11 +417,17 @@ void setupmaterials(int start, int len)
     }
     if(hasmat&(1<<MAT_WATER))
     {
-        extern void loadcaustics();
-        loadcaustics();
-        lookuptexture(-MAT_WATER);
+        loadcaustics(true);
+        preloadwatershaders(true);
+        lookupmaterialslot(MAT_WATER);
     }
-    if(hasmat&(1<<MAT_LAVA)) lookuptexture(-MAT_LAVA);
+    if(hasmat&(1<<MAT_LAVA)) 
+    {
+        useshaderbyname("lava");
+        useshaderbyname("lavaglare");
+        lookupmaterialslot(MAT_LAVA);
+    }
+    if(hasmat&(1<<MAT_GLASS)) useshaderbyname("glass");
 }
 
 VARP(showmat, 0, 1, 1);
@@ -481,7 +486,7 @@ void sortmaterials(vector<materialsurface *> &vismats)
             materialsurface &m = va->matbuf[i];
             if(!editmode || !showmat)
             {
-                if(m.material==MAT_WATER && (m.orient==O_TOP || (refracting<0 && reflectz>hdr.worldsize))) continue;
+                if(m.material==MAT_WATER && (m.orient==O_TOP || (refracting<0 && reflectz>worldsize))) continue;
                 if(m.material&MAT_EDIT) continue;
                 if(glaring && m.material!=MAT_LAVA) continue;
             }
@@ -507,13 +512,13 @@ void rendermatgrid(vector<materialsurface *> &vismats)
             lastmat = curmat;
             switch(curmat)
             {
-                case MAT_WATER:  glColor3ub( 0,  0, 85); break; // blue
-                case MAT_CLIP:   glColor3ub(85,  0,  0); break; // red
-                case MAT_GLASS:  glColor3ub( 0, 85, 85); break; // cyan
-                case MAT_NOCLIP: glColor3ub( 0, 85,  0); break; // green
-                case MAT_LAVA:   glColor3ub(85, 40,  0); break; // orange
-                case MAT_AICLIP: glColor3ub(85, 85,  0); break; // yellow
-                case MAT_DEATH:  glColor3ub(40, 40, 40); break; // black
+                case MAT_WATER:    glColor3ub( 0,  0, 85); break; // blue
+                case MAT_CLIP:     glColor3ub(85,  0,  0); break; // red
+                case MAT_GLASS:    glColor3ub( 0, 85, 85); break; // cyan
+                case MAT_NOCLIP:   glColor3ub( 0, 85,  0); break; // green
+                case MAT_LAVA:     glColor3ub(85, 40,  0); break; // orange
+                case MAT_GAMECLIP: glColor3ub(85, 85,  0); break; // yellow
+                case MAT_DEATH:    glColor3ub(40, 40, 40); break; // black
             }
         }
         drawmaterial(m.orient, m.o.x, m.o.y, m.o.z, m.csize, m.rsize, -0.1f);
@@ -546,7 +551,7 @@ void drawglass(int orient, int x, int y, int z, int csize, int rsize, float offs
     xtraverts += 4;
 }
 
-VARP(waterfallenv, 0, 1, 1);
+VARFP(waterfallenv, 0, 1, 1, preloadwatershaders());
 
 void rendermaterials()
 {
@@ -556,11 +561,10 @@ void rendermaterials()
 
     glDisable(GL_CULL_FACE);
 
-    Slot &wslot = lookuptexture(-MAT_WATER), &lslot = lookuptexture(-MAT_LAVA);
-    uchar wcol[4], wfcol[4];
-    getwatercolour(wcol);
-    getwaterfallcolour(wfcol);
-    wcol[3] = wfcol[3] = 192;
+    Slot &wslot = lookupmaterialslot(MAT_WATER), &lslot = lookupmaterialslot(MAT_LAVA);
+    uchar wcol[4] = { watercolor[0], watercolor[1], watercolor[2], 192 }, 
+          wfcol[4] = { waterfallcolor[0], waterfallcolor[1], waterfallcolor[2], 192 };
+    if(!wfcol[0] && !wfcol[1] && !wfcol[2]) memcpy(wfcol, wcol, 3);
     int lastorient = -1, lastmat = -1;
     GLenum textured = GL_TEXTURE_2D;
     bool begin = false, depth = true, blended = false, overbright = false, usedcamera = false, usedwaterfall = false;
@@ -791,13 +795,13 @@ void rendermaterials()
                     }
                     switch(curmat&~MAT_EDIT)
                     {
-                        case MAT_WATER:  glColor3ub(255, 128,   0); break; // blue
-                        case MAT_CLIP:   glColor3ub(  0, 255, 255); break; // red
-                        case MAT_GLASS:  glColor3ub(255,   0,   0); break; // cyan
-                        case MAT_NOCLIP: glColor3ub(255,   0, 255); break; // green
-                        case MAT_LAVA:   glColor3ub(  0, 128, 255); break; // orange
-                        case MAT_AICLIP: glColor3ub(  0,   0, 255); break; // yellow
-                        case MAT_DEATH:  glColor3ub(192, 192, 192); break; // black
+                        case MAT_WATER:    glColor3ub(255, 128,   0); break; // blue
+                        case MAT_CLIP:     glColor3ub(  0, 255, 255); break; // red
+                        case MAT_GLASS:    glColor3ub(255,   0,   0); break; // cyan
+                        case MAT_NOCLIP:   glColor3ub(255,   0, 255); break; // green
+                        case MAT_LAVA:     glColor3ub(  0, 128, 255); break; // orange
+                        case MAT_GAMECLIP: glColor3ub(  0,   0, 255); break; // yellow
+                        case MAT_DEATH:    glColor3ub(192, 192, 192); break; // black
                     }   
                     break;
                 }
