@@ -29,6 +29,16 @@ struct md5hierarchy
     int parent, flags, start;
 };
 
+struct md5adjustment
+{
+    float yaw, pitch, roll;
+    vec translate;
+
+    md5adjustment(float yaw, float pitch, float roll, const vec &translate) : yaw(yaw), pitch(pitch), roll(roll), translate(translate) {}
+};
+
+vector<md5adjustment> md5adjustments;
+
 struct md5 : skelmodel
 {
     md5(const char *name) : skelmodel(name) {}
@@ -85,34 +95,15 @@ struct md5 : skelmodel
             }
         }
 
-        void buildnorms(bool areaweight = true)
-        {
-            loopi(numverts) verts[i].norm = vec(0, 0, 0);
-            loopi(numtris)
-            {
-                tri &t = tris[i];
-                vert &v1 = verts[t.vert[0]], &v2 = verts[t.vert[1]], &v3 = verts[t.vert[2]];
-                vec norm;
-                norm.cross(vec(v2.pos).sub(v1.pos), vec(v3.pos).sub(v1.pos)); 
-                if(!areaweight) norm.normalize();
-                v1.norm.add(norm);
-                v2.norm.add(norm);
-                v3.norm.add(norm);
-            }
-            loopi(numverts) verts[i].norm.normalize();
-        }
-
-        void load(FILE *f, char *buf, size_t bufsize)
+        void load(stream *f, char *buf, size_t bufsize)
         {
             md5weight w;
             md5vert v;
             tri t;
             int index;
 
-            for(;;)
+            while(f->getline(buf, bufsize) && buf[0]!='}')
             {
-                fgets(buf, bufsize, f);
-                if(buf[0]=='}' || feof(f)) break;
                 if(strstr(buf, "// meshes:"))
                 {
                     char *start = strchr(buf, ':')+1;
@@ -176,42 +167,38 @@ struct md5 : skelmodel
         {
         }
 
-        bool loadmd5mesh(const char *filename)
+        bool loadmd5mesh(const char *filename, float smooth)
         {
-            FILE *f = openfile(filename, "r");
+            stream *f = openfile(filename, "r");
             if(!f) return false;
 
             char buf[512];
             vector<md5joint> basejoints;
-            for(;;)
+            while(f->getline(buf, sizeof(buf)))
             {
-                fgets(buf, sizeof(buf), f);
-                if(feof(f)) break;
                 int tmp;
                 if(sscanf(buf, " MD5Version %d", &tmp)==1)
                 {
-                    if(tmp!=10) { fclose(f); return false; }
+                    if(tmp!=10) { delete f; return false; }
                 }
                 else if(sscanf(buf, " numJoints %d", &tmp)==1)
                 {
-                    if(tmp<1) { fclose(f); return false; }
+                    if(tmp<1) { delete f; return false; }
                     if(skel->numbones>0) continue;
                     skel->numbones = tmp;
                     skel->bones = new boneinfo[skel->numbones];
                 }
                 else if(sscanf(buf, " numMeshes %d", &tmp)==1)
                 {
-                    if(tmp<1) { fclose(f); return false; }
+                    if(tmp<1) { delete f; return false; }
                 }
                 else if(strstr(buf, "joints {"))
                 {
                     string name;
                     int parent;
                     md5joint j;
-                    for(;;)
+                    while(f->getline(buf, sizeof(buf)) && buf[0]!='}')
                     {
-                        fgets(buf, sizeof(buf), f);
-                        if(buf[0]=='}' || feof(f)) break;
                         if(sscanf(buf, " %s %d ( %f %f %f ) ( %f %f %f )",
                             name, &parent, &j.pos.x, &j.pos.y, &j.pos.z,
                             &j.orient.x, &j.orient.y, &j.orient.z)==8)
@@ -232,7 +219,7 @@ struct md5 : skelmodel
                             basejoints.add(j);
                         }
                     }
-                    if(basejoints.length()!=skel->numbones) { fclose(f); return false; }
+                    if(basejoints.length()!=skel->numbones) { delete f; return false; }
                 }
                 else if(strstr(buf, "mesh {"))
                 {
@@ -259,13 +246,14 @@ struct md5 : skelmodel
             {
                 md5mesh &m = *(md5mesh *)meshes[i];
                 m.buildverts(basejoints);
-                m.buildnorms();
+                if(smooth <= 1) m.smoothnorms(smooth);
+                else m.buildnorms();
                 m.cleanup();
             }
             
             sortblendcombos();
 
-            fclose(f);
+            delete f;
             return true;
         }
 
@@ -274,7 +262,7 @@ struct md5 : skelmodel
             skelanimspec *sa = skel->findskelanim(filename);
             if(sa) return sa;
 
-            FILE *f = openfile(filename, "r");
+            stream *f = openfile(filename, "r");
             if(!f) return NULL;
 
             vector<md5hierarchy> hierarchy;
@@ -283,43 +271,34 @@ struct md5 : skelmodel
             float *animdata = NULL;
             dualquat *animbones = NULL;
             char buf[512];
-            for(;;)
+            while(f->getline(buf, sizeof(buf)))
             {
-                fgets(buf, sizeof(buf), f);
-                if(feof(f)) break;
                 int tmp;
                 if(sscanf(buf, " MD5Version %d", &tmp)==1)
                 {
-                    if(tmp!=10) { fclose(f); return NULL; }
+                    if(tmp!=10) { delete f; return NULL; }
                 }
                 else if(sscanf(buf, " numJoints %d", &tmp)==1)
                 {
-                    if(tmp!=skel->numbones) { fclose(f); return NULL; }
+                    if(tmp!=skel->numbones) { delete f; return NULL; }
                 }
                 else if(sscanf(buf, " numFrames %d", &animframes)==1)
                 {
-                    if(animframes<1) { fclose(f); return NULL; }
+                    if(animframes<1) { delete f; return NULL; }
                 }
                 else if(sscanf(buf, " frameRate %d", &tmp)==1);
                 else if(sscanf(buf, " numAnimatedComponents %d", &animdatalen)==1)
                 {
-                    if(animdatalen<1) { fclose(f); return NULL; }
-                    animdata = new float[animdatalen];
+                    if(animdatalen>0) animdata = new float[animdatalen];
                 }
                 else if(strstr(buf, "bounds {"))
                 {
-                    for(;;)
-                    {
-                        fgets(buf, sizeof(buf), f);
-                        if(buf[0]=='}' || feof(f)) break;
-                    }
+                    while(f->getline(buf, sizeof(buf)) && buf[0]!='}');
                 }
                 else if(strstr(buf, "hierarchy {"))
                 {
-                    for(;;)
+                    while(f->getline(buf, sizeof(buf)) && buf[0]!='}')
                     {
-                        fgets(buf, sizeof(buf), f);
-                        if(buf[0]=='}' || feof(f)) break;
                         md5hierarchy h;
                         if(sscanf(buf, " %s %d %d %d", h.name, &h.parent, &h.flags, &h.start)==4)
                             hierarchy.add(h);
@@ -327,10 +306,8 @@ struct md5 : skelmodel
                 }
                 else if(strstr(buf, "baseframe {"))
                 {
-                    for(;;)
+                    while(f->getline(buf, sizeof(buf)) && buf[0]!='}')
                     {
-                        fgets(buf, sizeof(buf), f);
-                        if(buf[0]=='}' || feof(f)) break;
                         md5joint j;
                         if(sscanf(buf, " ( %f %f %f ) ( %f %f %f )", &j.pos.x, &j.pos.y, &j.pos.z, &j.orient.x, &j.orient.y, &j.orient.z)==6)
                         {
@@ -341,7 +318,7 @@ struct md5 : skelmodel
                             basejoints.add(j);
                         }
                     }
-                    if(basejoints.length()!=skel->numbones) { fclose(f); return NULL; }
+                    if(basejoints.length()!=skel->numbones) { delete f; return NULL; }
                     animbones = new dualquat[(skel->numframes+animframes)*skel->numbones];
                     if(skel->bones)
                     {
@@ -359,15 +336,22 @@ struct md5 : skelmodel
                 }
                 else if(sscanf(buf, " frame %d", &tmp)==1)
                 {
-                    loopi(animdatalen) fscanf(f, "%f", &animdata[i]);
+                    for(int numdata = 0; f->getline(buf, sizeof(buf)) && buf[0]!='}';)
+                    {
+                        for(char *src = buf, *next = src; numdata < animdatalen; numdata++, src = next)
+                        {
+                            animdata[numdata] = strtod(src, &next);
+                            if(next <= src) break;
+                        }
+                    }
                     dualquat *frame = &animbones[tmp*skel->numbones];
                     loopv(basejoints)
                     {
                         md5hierarchy &h = hierarchy[i];
                         md5joint j = basejoints[i];
-                        float *jdata = &animdata[h.start];
-                        if(h.flags)
+                        if(h.start < animdatalen && h.flags)
                         {
+                            float *jdata = &animdata[h.start];
                             if(h.flags&1) j.pos.x = *jdata++;
                             if(h.flags&2) j.pos.y = -*jdata++;
                             if(h.flags&4) j.pos.z = *jdata++;
@@ -378,16 +362,24 @@ struct md5 : skelmodel
                             /*if(memcmp(&j, &basejoints[i], sizeof(j))) usedjoints[i] = 1; */
                         }
                         frame[i] = dualquat(j.orient, j.pos);
+                        frame[i].fixantipodal(skel->framebones[i]);
 #if 0
                         if(h.parent<0) frame[i] = dualquat(j.orient, j.pos); 
                         else (frame[i] = frame[h.parent]).mul(dualquat(j.orient, j.pos));
 #endif
                     }
+                    loopv(md5adjustments)
+                    {
+                        if(md5adjustments[i].yaw) frame[i].mulorient(quat(vec(0, 0, 1), md5adjustments[i].yaw*RAD));
+                        if(md5adjustments[i].pitch) frame[i].mulorient(quat(vec(0, -1, 0), md5adjustments[i].pitch*RAD));
+                        if(md5adjustments[i].roll) frame[i].mulorient(quat(vec(-1, 0, 0), md5adjustments[i].roll*RAD));
+                        if(!md5adjustments[i].translate.iszero()) frame[i].translate(md5adjustments[i].translate);
+                    }
                 }    
             }
 
             DELETEA(animdata);
-            fclose(f);
+            delete f;
 
 #if 0
             vector<dualquat> invbase;
@@ -421,30 +413,21 @@ struct md5 : skelmodel
             return sa;
         }
 
-        bool load(const char *meshfile)
+        bool load(const char *meshfile, float smooth)
         {
             name = newstring(meshfile);
 
-            if(!loadmd5mesh(meshfile)) return false;
+            if(!loadmd5mesh(meshfile, smooth)) return false;
             
             return true;
         }
     };            
 
-    void extendbb(int frame, vec &center, vec &radius, modelattach &a)
-    {
-        vec acenter, aradius;
-        a.m->boundbox(frame, acenter, aradius);
-        float margin = 2*max(aradius.x, max(aradius.y, aradius.z));
-        radius.x += margin;
-        radius.y += margin;
-    }
-
     meshgroup *loadmeshes(char *name, va_list args)
     {
         md5meshgroup *group = new md5meshgroup;
         group->shareskeleton(va_arg(args, char *));
-        if(!group->load(name)) { delete group; return NULL; }
+        if(!group->load(name, va_arg(args, double))) { delete group; return NULL; }
         return group;
     }
 
@@ -455,15 +438,16 @@ struct md5 : skelmodel
         mdl.model = this;
         mdl.index = 0;
         mdl.pitchscale = mdl.pitchoffset = mdl.pitchmin = mdl.pitchmax = 0;
+        md5adjustments.setsizenodelete(0);
         const char *fname = loadname + strlen(loadname);
         do --fname; while(fname >= loadname && *fname!='/' && *fname!='\\');
         fname++;
-        s_sprintfd(meshname)("packages/models/%s/%s.md5mesh", loadname, fname);
-        mdl.meshes = sharemeshes(path(meshname));
+        defformatstring(meshname)("packages/models/%s/%s.md5mesh", loadname, fname);
+        mdl.meshes = sharemeshes(path(meshname), NULL, 0);
         if(!mdl.meshes) return false;
         mdl.initanimparts();
         mdl.initskins();
-        s_sprintfd(animname)("packages/models/%s/%s.md5anim", loadname, fname);
+        defformatstring(animname)("packages/models/%s/%s.md5anim", loadname, fname);
         ((md5meshgroup *)mdl.meshes)->loadmd5anim(path(animname));
         return true;
     }
@@ -471,44 +455,57 @@ struct md5 : skelmodel
     bool load()
     {
         if(loaded) return true;
-        s_sprintf(md5dir)("packages/models/%s", loadname);
-        s_sprintfd(cfgname)("packages/models/%s/md5.cfg", loadname);
+        formatstring(md5dir)("packages/models/%s", loadname);
+        defformatstring(cfgname)("packages/models/%s/md5.cfg", loadname);
 
         loadingmd5 = this;
         persistidents = false;
-        if(execfile(cfgname) && parts.length()) // configured md5, will call the md5* commands below
+        if(execfile(cfgname, false) && parts.length()) // configured md5, will call the md5* commands below
         {
             persistidents = true;
             loadingmd5 = NULL;
             loopv(parts) if(!parts[i]->meshes) return false;
         }
-        else if(!loaddefaultparts()) // md5 without configuration, try default tris and skin
+        else // md5 without configuration, try default tris and skin 
         {
             persistidents = true;
+            if(!loaddefaultparts()) 
+            {
+                loadingmd5 = NULL;
+                return false;
+            }
             loadingmd5 = NULL;
-            return false;
         }
-        loadingmd5 = NULL;
+        scale /= 4;
+        parts[0]->translate = translate;
         loopv(parts) 
         {
             skelpart *p = (skelpart *)parts[i];
             p->endanimparts();
-            p->meshes = p->meshes->scaleverts(scale/4.0f, i ? vec(0, 0, 0) : vec(translate.x, translate.y, translate.z));
+            p->meshes->shared++;
         }
+        preloadshaders();
         return loaded = true;
     }
 };
 
-void md5load(char *meshfile, char *skelname)
+void setmd5dir(char *name)
 {
     if(!loadingmd5) { conoutf("not loading an md5"); return; }
-    s_sprintfd(filename)("%s/%s", md5dir, meshfile);
+    formatstring(md5dir)("packages/models/%s", name);
+}
+    
+void md5load(char *meshfile, char *skelname, float *smooth)
+{
+    if(!loadingmd5) { conoutf("not loading an md5"); return; }
+    defformatstring(filename)("%s/%s", md5dir, meshfile);
     md5::skelpart &mdl = *new md5::skelpart;
     loadingmd5->parts.add(&mdl);
     mdl.model = loadingmd5;
     mdl.index = loadingmd5->parts.length()-1;
     mdl.pitchscale = mdl.pitchoffset = mdl.pitchmin = mdl.pitchmax = 0;
-    mdl.meshes = loadingmd5->sharemeshes(path(filename), skelname[0] ? skelname : NULL);
+    md5adjustments.setsizenodelete(0);
+    mdl.meshes = loadingmd5->sharemeshes(path(filename), skelname[0] ? skelname : NULL, double(*smooth > 0 ? cos(clamp(*smooth, 0.0f, 90.0f)*RAD) : 2));
     if(!mdl.meshes) conoutf("could not load %s", filename); // ignore failure
     else 
     {
@@ -573,6 +570,18 @@ void md5pitch(char *name, float *pitchscale, float *pitchoffset, float *pitchmin
     }
 }
 
+void md5adjust(char *name, float *yaw, float *pitch, float *roll, float *tx, float *ty, float *tz)
+{
+    if(!loadingmd5 || loadingmd5->parts.empty()) { conoutf("not loading an md5"); return; }
+    md5::part &mdl = *loadingmd5->parts.last();
+
+    if(!name[0]) return;
+    int i = mdl.meshes ? ((md5::skelmeshgroup *)mdl.meshes)->skel->findbone(name) : -1;
+    if(i < 0) {  conoutf("could not find bone %s to adjust", name); return; }
+    while(!md5adjustments.inrange(i)) md5adjustments.add(md5adjustment(0, 0, 0, vec(0, 0, 0)));
+    md5adjustments[i] = md5adjustment(*yaw, *pitch, *roll, vec(*tx/4, *ty/4, *tz/4));
+}
+ 
 #define loopmd5meshes(meshname, m, body) \
     if(!loadingmd5 || loadingmd5->parts.empty()) { conoutf("not loading an md5"); return; } \
     md5::part &mdl = *loadingmd5->parts.last(); \
@@ -594,7 +603,7 @@ void md5skin(char *meshname, char *tex, char *masks, float *envmapmax, float *en
         s.tex = textureload(makerelpath(md5dir, tex), 0, true, false);
         if(*masks)
         {
-            s.masks = textureload(makerelpath(md5dir, masks, "<ffmask:25>"), 0, true, false);
+            s.masks = textureload(makerelpath(md5dir, masks, NULL, "<ffmask:25>"), 0, true, false);
             s.envmapmax = *envmapmax;
             s.envmapmin = *envmapmin;
         }
@@ -640,6 +649,11 @@ void md5alphablend(char *meshname, int *blend)
     loopmd5skins(meshname, s, s.alphablend = *blend!=0);
 }
 
+void md5cullface(char *meshname, int *cullface)
+{
+    loopmd5skins(meshname, s, s.cullface = *cullface!=0);
+}
+
 void md5envmap(char *meshname, char *envmap)
 {
     Texture *tex = cubemapload(envmap);
@@ -654,11 +668,6 @@ void md5bumpmap(char *meshname, char *normalmap, char *skin)
     loopmd5skins(meshname, s, { s.unlittex = skintex; s.normalmap = normalmaptex; m.calctangents(); });
 }
 
-void md5translucent(char *meshname, float *translucency)
-{
-    loopmd5skins(meshname, s, s.translucency = *translucency);
-}
-
 void md5fullbright(char *meshname, float *fullbright)
 {
     loopmd5skins(meshname, s, s.fullbright = *fullbright);
@@ -683,8 +692,9 @@ void md5anim(char *anim, char *animfile, float *speed, int *priority)
     if(anims.empty()) conoutf("could not find animation %s", anim);
     else 
     {
-        s_sprintfd(filename)("%s/%s", md5dir, animfile);
         md5::part *p = loadingmd5->parts.last();
+        if(!p->meshes) return;
+        defformatstring(filename)("%s/%s", md5dir, animfile);
         md5::skelanimspec *sa = ((md5::md5meshgroup *)p->meshes)->loadmd5anim(path(filename));
         if(!sa) conoutf("could not load md5anim file %s", filename);
         else loopv(anims)
@@ -729,9 +739,11 @@ void md5noclip(char *meshname, int *noclip)
     loopmd5meshes(meshname, m, m.noclip = *noclip!=0);
 }
 
-COMMAND(md5load, "ss");
+COMMANDN(md5dir, setmd5dir, "s");
+COMMAND(md5load, "ssf");
 COMMAND(md5tag, "ss");
 COMMAND(md5pitch, "sffff");
+COMMAND(md5adjust, "sffffff");
 COMMAND(md5skin, "sssff");
 COMMAND(md5spec, "si");
 COMMAND(md5ambient, "si");
@@ -739,9 +751,9 @@ COMMAND(md5glow, "si");
 COMMAND(md5glare, "sff");
 COMMAND(md5alphatest, "sf");
 COMMAND(md5alphablend, "si");
+COMMAND(md5cullface, "si");
 COMMAND(md5envmap, "ss");
 COMMAND(md5bumpmap, "sss");
-COMMAND(md5translucent, "sf");
 COMMAND(md5fullbright, "sf");
 COMMAND(md5shader, "ss");
 COMMAND(md5scroll, "sff");
