@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 
 extern int outline;
@@ -94,7 +93,7 @@ VARF(gridpower, 3-VVEC_FRAC, 3, VVEC_INT-1,
 {
     if(dragging) return;
     gridsize = 1<<gridpower;
-    if(gridsize>=hdr.worldsize) gridsize = hdr.worldsize/2;
+    if(gridsize>=worldsize) gridsize = worldsize/2;
     cancelsel();
 });
 
@@ -123,7 +122,7 @@ void cancelsel()
 void toggleedit(bool force)
 {
     if(player->state!=CS_ALIVE && player->state!=CS_DEAD && player->state!=CS_EDITING) return; // do not allow dead players to edit to avoid state confusion
-    if(!editmode && !cc->allowedittoggle()) return;         // not in most multiplayer modes
+    if(!editmode && !game::allowedittoggle()) return;         // not in most multiplayer modes
     if(!(editmode = !editmode))
     {
         player->state = player->editstate;
@@ -132,21 +131,22 @@ void toggleedit(bool force)
     }
     else
     {
-        cl->resetgamestate();
+        game::resetgamestate();
         player->editstate = player->state;
         player->state = CS_EDITING;
     }
     cancelsel();
+    stoppaintblendmap();
     keyrepeat(editmode);
     editing = entediting = editmode;
     extern int fullbright;
     if(fullbright) initlights();
-    if(!force) cc->edittoggled(editmode);
+    if(!force) game::edittoggled(editmode);
 }
 
-bool noedit(bool view)
+bool noedit(bool view, bool msg)
 {
-    if(!editmode) { conoutf(CON_ERROR, "operation only allowed in edit mode"); return true; }
+    if(!editmode) { if(msg) conoutf(CON_ERROR, "operation only allowed in edit mode"); return true; }
     if(view || haveselent()) return false;
     float r = 1.0f;
     vec o, s;
@@ -156,7 +156,7 @@ bool noedit(bool view)
     o.add(s);
     r = float(max(s.x, max(s.y, s.z)));
     bool viewable = (isvisiblesphere(r, o) != VFC_NOT_VISIBLE);
-    if(!viewable) conoutf(CON_ERROR, "selection not in view");
+    if(!viewable && msg) conoutf(CON_ERROR, "selection not in view");
     return !viewable;
 }
 
@@ -277,32 +277,27 @@ inline bool isheightmap(int orient, int d, bool empty, cube *c);
 extern void entdrag(const vec &ray);
 extern bool hoveringonent(int ent, int orient);
 extern void renderentselection(const vec &o, const vec &ray, bool entmoving);
-extern float rayent(const vec &o, const vec &ray, vec &hitpos, float radius, int mode, int size, int &orient, int &ent);
+extern float rayent(const vec &o, const vec &ray, float radius, int mode, int size, int &orient, int &ent);
 
 VAR(gridlookup, 0, 0, 1);
 VAR(passthroughcube, 0, 1, 1);
 
-void cursorupdate()
+void rendereditcursor()
 {
     if(sel.grid == 0) sel.grid = gridsize;
 
-    vec target(worldpos);
-    if(!insideworld(target)) loopi(3) 
-        target[i] = max(min(target[i], float(hdr.worldsize)), 0.0f);
-    vec ray(target);
-    ray.sub(player->o).normalize();
     int d   = dimension(sel.orient),
         od  = dimension(orient),
         odc = dimcoord(orient);
 
-    bool hidecursor = g3d_windowhit(true, false), hovering = false;
+    bool hidecursor = g3d_windowhit(true, false) || blendpaintmode, hovering = false;
     hmapsel = false;
            
     if(moving)
     {       
         ivec e;
         static vec v, handle;
-        editmoveplane(sel.o.tovec(), ray, od, sel.o[D[od]]+odc*sel.grid*sel.s[D[od]], handle, v, !havesel);
+        editmoveplane(sel.o.tovec(), camdir, od, sel.o[D[od]]+odc*sel.grid*sel.s[D[od]], handle, v, !havesel);
         if(!havesel)
         {
             v.add(handle);
@@ -317,22 +312,22 @@ void cursorupdate()
     else 
     if(entmoving)
     {
-        entdrag(ray);       
+        entdrag(camdir);       
     }
     else
     {  
-        vec v;
         ivec w;
         float sdist = 0, wdist = 0, t;
         int entorient = 0, ent = -1;
        
-        wdist = rayent(player->o, ray, v, 0, (editmode && showmat ? RAY_EDITMAT : 0)   // select cubes first
-                                           | (!dragging && entediting ? RAY_ENTS : 0)
-                                           | RAY_SKIPFIRST 
-                                           | (passthroughcube==1 ? RAY_PASS : 0), gridsize, entorient, ent);
+        wdist = rayent(player->o, camdir, 1e16f, 
+                       (editmode && showmat ? RAY_EDITMAT : 0)   // select cubes first
+                       | (!dragging && entediting ? RAY_ENTS : 0)
+                       | RAY_SKIPFIRST 
+                       | (passthroughcube==1 ? RAY_PASS : 0), gridsize, entorient, ent);
      
         if((havesel || dragging) && !passthroughsel && !hmapedit)     // now try selecting the selection
-            if(rayrectintersect(sel.o.tovec(), vec(sel.s.tovec()).mul(sel.grid), player->o, ray, sdist, orient))
+            if(rayrectintersect(sel.o.tovec(), vec(sel.s.tovec()).mul(sel.grid), player->o, camdir, sdist, orient))
             {   // and choose the nearest of the two
                 if(sdist < wdist) 
                 {
@@ -340,7 +335,7 @@ void cursorupdate()
                     ent   = -1;
                 }
             }
-
+       
         if((hovering = hoveringonent(hidecursor ? -1 : ent, entorient)))
         {
            if(!havesel) 
@@ -351,22 +346,28 @@ void cursorupdate()
         }
         else 
         {
-       
-            v = ray;
-            v.mul(wdist+0.1f);
-            v.add(player->o);
-            w = v;
-            cube *c = &lookupcube(w.x, w.y, w.z);            
+            vec w = vec(camdir).mul(wdist+0.05f).add(player->o);
+            if(!insideworld(w))
+            {
+                loopi(3) wdist = min(wdist, ((camdir[i] > 0 ? worldsize : 0) - player->o[i]) / camdir[i]);
+                w = vec(camdir).mul(wdist-0.05f).add(player->o);
+                if(!insideworld(w))
+                {
+                    wdist = 0;
+                    loopi(3) w[i] = clamp(player->o[i], 0.0f, float(worldsize));
+                }
+            }
+            cube *c = &lookupcube(int(w.x), int(w.y), int(w.z));            
             if(gridlookup && !dragging && !moving && !havesel && hmapedit!=1) gridsize = lusize;
             int mag = lusize / gridsize;
-            normalizelookupcube(w.x, w.y, w.z);
-            if(sdist == 0 || sdist > wdist) rayrectintersect(lu.tovec(), vec(gridsize), player->o, ray, t=0, orient); // just getting orient     
+            normalizelookupcube(int(w.x), int(w.y), int(w.z));
+            if(sdist == 0 || sdist > wdist) rayrectintersect(lu.tovec(), vec(gridsize), player->o, camdir, t=0, orient); // just getting orient     
             cur = lu;
-            cor = vec(v).mul(2).div(gridsize);
+            cor = vec(w).mul(2).div(gridsize);
             od = dimension(orient);
             d = dimension(sel.orient);
             
-            if(hmapedit==1 && dimcoord(horient) == (ray[dimension(horient)]<0))
+            if(hmapedit==1 && dimcoord(horient) == (camdir[dimension(horient)]<0))
             {
                 hmapsel = isheightmap(horient, dimension(horient), false, c);     
                 if(hmapsel)
@@ -413,7 +414,7 @@ void cursorupdate()
 
             sel.corner = (cor[R[d]]-(lu[R[d]]*2)/gridsize)+(cor[C[d]]-(lu[C[d]]*2)/gridsize)*2;
             selchildcount = 0;
-            countselchild(worldroot, ivec(0, 0, 0), hdr.worldsize/2);
+            countselchild(worldroot, ivec(0, 0, 0), worldsize/2);
             if(mag>1 && selchildcount==1) selchildcount = -mag;
         }
     }
@@ -423,7 +424,7 @@ void cursorupdate()
     
     // cursors    
 
-    renderentselection(player->o, ray, entmoving!=0);
+    renderentselection(player->o, camdir, entmoving!=0);
 
     enablepolygonoffset(GL_POLYGON_OFFSET_LINE);
 
@@ -464,9 +465,16 @@ void cursorupdate()
     glDisable(GL_BLEND);
 }
 
+void tryedit()
+{
+    extern int hidehud;
+    if(!editmode || hidehud || mainmenu) return;
+    if(blendpaintmode) trypaintblendmap();
+}
+
 //////////// ready changes to vertex arrays ////////////
 
-static bool invalidatedmerges = false, haschanged = false;
+static bool haschanged = false;
 
 void readychanges(block3 &b, cube *c, const ivec &cor, int size)
 {
@@ -480,15 +488,7 @@ void readychanges(block3 &b, cube *c, const ivec &cor, int size)
                 int hasmerges = c[i].ext->va->hasmerges;
                 destroyva(c[i].ext->va);
                 c[i].ext->va = NULL;
-                if(hasmerges) 
-                {
-                    if(!invalidatedmerges)
-                    {
-                        show_out_of_renderloop_progress(0, "invalidating merged surfaces...");
-                        invalidatedmerges = true;
-                    }
-                    invalidatemerges(c[i]);
-                }
+                if(hasmerges) invalidatemerges(c[i], true); 
             }
             freeoctaentities(c[i]);
             c[i].ext->tjoints = -1;
@@ -507,11 +507,10 @@ void readychanges(block3 &b, cube *c, const ivec &cor, int size)
     }
 }
 
-void commitchanges()
+void commitchanges(bool force)
 {
-    if(!haschanged) return;
+    if(!force && !haschanged) return;
     haschanged = false;
-    invalidatedmerges = false;
 
     extern vector<vtxarray *> valist;
     int oldlen = valist.length();
@@ -522,6 +521,7 @@ void commitchanges()
     setupmaterials(oldlen);
     invalidatepostfx();
     updatevabbs();
+    resetblobs();
 }
 
 void changed(const block3 &sel, bool commit = true)
@@ -534,7 +534,7 @@ void changed(const block3 &sel, bool commit = true)
     {
         b.o[i] -= 1;
         b.s[i] += 2;
-        readychanges(b, worldroot, ivec(0, 0, 0), hdr.worldsize/2);
+        readychanges(b, worldroot, ivec(0, 0, 0), worldsize/2);
         b.o[i] += 1;
         b.s[i] -= 2;
     }
@@ -586,7 +586,7 @@ void freeblock(block3 *b, bool alloced = true)
 
 void selgridmap(selinfo &sel, int *g)                           // generates a map of the cube sizes at each grid point
 {
-    loopxyz(sel, -sel.grid, (*g++ = lusize, c));
+    loopxyz(sel, -sel.grid, (*g++ = lusize, (void)c));
 }
 
 void freeundo(undoblock *u)
@@ -707,9 +707,11 @@ void addundo(undoblock *u)
     pruneundos(undomegs<<20);
 }
 
+VARP(nompedit, 0, 1, 1);
+
 void makeundoex(selinfo &s)
 {
-    if(multiplayer(false)) return;
+    if(nompedit && multiplayer(false)) return;
     undoblock *u = newundocube(s);
     addundo(u);
 }
@@ -723,7 +725,7 @@ void makeundo()                        // stores state of selected cubes before
 
 void swapundo(undolist &a, undolist &b, const char *s)
 {
-    if(noedit() || multiplayer()) return;
+    if(noedit() || (nompedit && multiplayer())) return;
     if(a.empty()) { conoutf(CON_WARN, "nothing more to %s", s); return; }	
 	int ts = a.last->timestamp;
     selinfo l = sel;
@@ -774,7 +776,7 @@ void freeeditinfo(editinfo *&e)
 
 void mpcopy(editinfo *&e, selinfo &sel, bool local)
 {
-    if(local) cl->edittrigger(sel, EDIT_COPY);
+    if(local) game::edittrigger(sel, EDIT_COPY);
     if(e==NULL) e = new editinfo;
     if(e->copy) freeblock(e->copy);
     e->copy = NULL;
@@ -785,7 +787,7 @@ void mpcopy(editinfo *&e, selinfo &sel, bool local)
 void mppaste(editinfo *&e, selinfo &sel, bool local)
 {
     if(e==NULL) return;
-    if(local) cl->edittrigger(sel, EDIT_PASTE);
+    if(local) game::edittrigger(sel, EDIT_PASTE);
     if(e->copy)
     {
         sel.s = e->copy->s;
@@ -1036,7 +1038,7 @@ namespace hmap
         if(biasup)
             pullhmap(0, >, <, 1, 0, -);
         else
-            pullhmap(hdr.worldsize, <, >, 0, 8, +);     
+            pullhmap(worldsize, <, >, 0, 8, +);     
    
         cube **c  = cmap[x][y];
         int e[2][2];
@@ -1128,7 +1130,7 @@ namespace hmap
         bool paintme = paintbrush;
         int cx = (sel.corner&1 ? 0 : -1);
         int cy = (sel.corner&2 ? 0 : -1);
-        hws= (hdr.worldsize>>gridpower);
+        hws= (worldsize>>gridpower);
         gx = (cur[R[d]] >> gridpower) + cx - MAXBRUSH2;
         gy = (cur[C[d]] >> gridpower) + cy - MAXBRUSH2;
         gz = (cur[D[d]] >> gridpower);
@@ -1154,7 +1156,7 @@ namespace hmap
             bnx = min(nx, brushmaxx-1);
             bny = min(ny, brushmaxy-1);   
         }
-        nz = hdr.worldsize-gridsize;
+        nz = worldsize-gridsize;
         mz = 0;
         hundo.s = ivec(d,1,1,5);
         hundo.orient = sel.orient;
@@ -1182,7 +1184,7 @@ namespace hmap
 }
 
 void edithmap(int dir, int mode) {    
-    if(multiplayer() || !hmapsel || gridsize < 8) return;    
+    if((nompedit && multiplayer()) || !hmapsel || gridsize < 8) return;    
     hmap::run(dir, mode);        
 }
 
@@ -1232,12 +1234,12 @@ void mpeditface(int dir, int mode, selinfo &sel, bool local)
     int seldir = dc ? -dir : dir;
 
     if(local)
-        cl->edittrigger(sel, EDIT_FACE, dir, mode);
+        game::edittrigger(sel, EDIT_FACE, dir, mode);
 
     if(mode==1)
     {
         int h = sel.o[d]+dc*sel.grid;
-        if(((dir>0) == dc && h<=0) || ((dir<0) == dc && h>=hdr.worldsize)) return;
+        if(((dir>0) == dc && h<=0) || ((dir<0) == dc && h>=worldsize)) return;
         if(dir<0) sel.o[d] += sel.grid * seldir;
     }
 
@@ -1325,7 +1327,7 @@ void pushsel(int *dir)
 
 void mpdelcube(selinfo &sel, bool local)
 {
-    if(local) cl->edittrigger(sel, EDIT_DELCUBE);
+    if(local) game::edittrigger(sel, EDIT_DELCUBE);
     loopselxyz(discardchildren(c); emptyfaces(c));
 }
 
@@ -1341,7 +1343,7 @@ COMMAND(delcube, "");
 
 /////////// texture editing //////////////////
 
-int curtexindex = -1, lasttex = 0;
+int curtexindex = -1, lasttex = 0, lasttexmillis = -1;
 int texpaneltimer = 0;
 vector<ushort> texmru;
 
@@ -1381,7 +1383,7 @@ void mpedittex(int tex, int allfaces, selinfo &sel, bool local)
 {
     if(local)
     {
-        cl->edittrigger(sel, EDIT_TEX, tex, allfaces);
+        game::edittrigger(sel, EDIT_TEX, tex, allfaces);
         if(allfaces || !(repsel == sel)) reptex = -1;
         repsel = sel;
     }
@@ -1398,11 +1400,15 @@ void filltexlist()
     }
 }
 
-void edittex(int i)
+void edittex(int i, bool save = true)
 {
-    curtexindex = i = min(max(i, 0), curtexnum-1);
-    int t = lasttex = texmru[i];    
-    mpedittex(t, allfaces, sel, true);
+    lasttex = i;
+    lasttexmillis = totalmillis;
+    if(save) 
+    {
+        loopvj(texmru) if(texmru[j]==lasttex) { curtexindex = j; break; }
+    }
+    mpedittex(i, allfaces, sel, true);
 }
 
 void edittex_(int *dir)
@@ -1411,7 +1417,8 @@ void edittex_(int *dir)
     filltexlist();
     texpaneltimer = 5000;
     if(!(lastsel==sel)) tofronttex();
-    edittex(curtexindex<0 ? 0 : curtexindex+*dir);
+    curtexindex = clamp(curtexindex<0 ? 0 : curtexindex+*dir, 0, curtexnum-1);
+    edittex(texmru[curtexindex], false);
 }
 
 void gettex()
@@ -1466,7 +1473,7 @@ void replacetexcube(cube &c, int oldtex, int newtex)
 
 void mpreplacetex(int oldtex, int newtex, selinfo &sel, bool local)
 {
-    if(local) cl->edittrigger(sel, EDIT_REPLACE, oldtex, newtex);
+    if(local) game::edittrigger(sel, EDIT_REPLACE, oldtex, newtex);
     loopi(8) replacetexcube(worldroot[i], oldtex, newtex);
     allchanged();
 }
@@ -1532,7 +1539,7 @@ void rotatecube(cube &c, int d)   // rotates cube clockwise. see pics in cvs for
 
 void mpflip(selinfo &sel, bool local)
 {
-    if(local) cl->edittrigger(sel, EDIT_FLIP);
+    if(local) game::edittrigger(sel, EDIT_FLIP);
     int zs = sel.s[dimension(sel.orient)];
     makeundo();
     loopxy(sel)
@@ -1556,7 +1563,7 @@ void flip()
 
 void mprotate(int cw, selinfo &sel, bool local)
 {
-    if(local) cl->edittrigger(sel, EDIT_ROTATE, cw);
+    if(local) game::edittrigger(sel, EDIT_ROTATE, cw);
     int d = dimension(sel.orient);
     if(!dimcoord(sel.orient)) cw = -cw;
     int m = sel.s[C[d]] < sel.s[R[d]] ? C[d] : R[d];
@@ -1585,44 +1592,66 @@ void rotate(int *cw)
 COMMAND(flip, "");
 COMMAND(rotate, "i");
 
-void setmat(cube &c, uchar mat, uchar matmask)
+void setmat(cube &c, uchar mat, uchar matmask, uchar filtermat, uchar filtermask)
 {
     if(c.children)
-        loopi(8) setmat(c.children[i], mat, matmask);
-    else if(mat!=MAT_AIR) 
+        loopi(8) setmat(c.children[i], mat, matmask, filtermat, filtermask);
+    else if(((c.ext ? c.ext->material : MAT_AIR)&filtermask) == filtermat)
     {
-        cubeext &e = ext(c);
-        e.material &= matmask;
-        e.material |= mat;
+        if(mat!=MAT_AIR) 
+        {
+            cubeext &e = ext(c);
+            e.material &= matmask;
+            e.material |= mat;
+        }
+        else if(c.ext) c.ext->material = MAT_AIR;
     }
-    else if(c.ext) c.ext->material = MAT_AIR;
 }
 
-void mpeditmat(int matid, selinfo &sel, bool local)
+void mpeditmat(int matid, int filter, selinfo &sel, bool local)
 {
-    if(local) cl->edittrigger(sel, EDIT_MAT, matid);
+    if(local) game::edittrigger(sel, EDIT_MAT, matid, filter);
 
-    uchar matmask = matid&MATF_VOLUME ? 0 : (matid&MATF_CLIP ? ~MATF_CLIP : 0xFF);
+    uchar matmask = matid&MATF_VOLUME ? 0 : (matid&MATF_CLIP ? ~MATF_CLIP : ~matid), 
+          filtermat = filter < 0 ? 0 : filter,
+          filtermask = filter < 0 ? 0 : (filter&MATF_VOLUME ? MATF_VOLUME : (filter&MATF_CLIP ? MATF_CLIP : filter));
     if(isclipped(matid&MATF_VOLUME)) matid |= MAT_CLIP;
     if(isdeadly(matid&MATF_VOLUME)) matid |= MAT_DEATH;
-    loopselxyz(setmat(c, matid, matmask));
+    if(matid < 0 && filter >= 0)
+    {
+        matid = 0;
+        matmask = filtermask;
+        if(isclipped(filter&MATF_VOLUME)) matmask &= ~MATF_CLIP; 
+        if(isdeadly(filter&MATF_VOLUME)) matmask &= ~MAT_DEATH; 
+    }
+    loopselxyz(setmat(c, matid, matmask, filtermat, filtermask));
 }
 
-void editmat(char *name)
+void editmat(char *name, char *filtername)
 {
     if(noedit()) return;
-    int id = findmaterial(name);
-    if(id<0) { conoutf(CON_ERROR, "unknown material \"%s\"", name); return; }
-    mpeditmat(id, sel, true);
+    int filter = -1;
+    if(filtername[0])
+    {
+        filter = findmaterial(filtername);
+        if(filter < 0) { conoutf(CON_ERROR, "unknown material \"%s\"", filtername); return; }
+    }
+    int id = -1;
+    if(name[0] || filter < 0)
+    {
+        id = findmaterial(name);
+        if(id<0) { conoutf(CON_ERROR, "unknown material \"%s\"", name); return; }
+    }
+    mpeditmat(id, filter, sel, true);
 }
 
-COMMAND(editmat, "s");
+COMMAND(editmat, "ss");
 
-#define TEXTURE_WIDTH 10
-#define TEXTURE_HEIGHT 7
+#define TEXGUI_WIDTH 10
+#define TEXGUI_HEIGHT 7
 extern int menudistance, menuautoclose;
 
-VAR(thumbtime, 0, 50, 1000);
+VARP(thumbtime, 0, 50, 1000);
 
 static int lastthumbnail = 0;
 
@@ -1632,36 +1661,42 @@ struct texturegui : g3d_callback
 {
     bool menuon;
     vec menupos;
-    int menustart;
-    
+    int menustart, menutab;
+   
+    texturegui() : menustart(-1) {} 
+
     void gui(g3d_gui &g, bool firstpass)
     {
-        int menutab = 1+curtexindex/(TEXTURE_WIDTH*TEXTURE_HEIGHT);        
-        int origtab = menutab;
+        int origtab = menutab, numtabs = max((curtexnum + TEXGUI_WIDTH*TEXGUI_HEIGHT - 1)/(TEXGUI_WIDTH*TEXGUI_HEIGHT), 1);
         g.start(menustart, 0.04f, &menutab);
-        loopi(1+curtexnum/(TEXTURE_WIDTH*TEXTURE_HEIGHT))
+        loopi(numtabs)
         {   
-            g.tab((i==0)?"Textures":NULL, 0xAAFFAA);
-            if(i != origtab-1) continue; //don't load textures on non-visible tabs!
-            loopj(TEXTURE_HEIGHT) 
+            g.tab(!i ? "Textures" : NULL, 0xAAFFAA);
+            if(i+1 != origtab) continue; //don't load textures on non-visible tabs!
+            loopj(TEXGUI_HEIGHT) 
             {
                 g.pushlist();
-                loopk(TEXTURE_WIDTH) 
+                loopk(TEXGUI_WIDTH) 
                 {
-                    int ti = (i*TEXTURE_HEIGHT+j)*TEXTURE_WIDTH+k;
+                    int ti = (i*TEXGUI_HEIGHT+j)*TEXGUI_WIDTH+k;
                     if(ti<curtexnum) 
                     {
-                        Texture *tex = notexture, *glowtex = NULL;
-                        Slot &slot = lookuptexture(texmru[ti], false);
+                        Texture *tex = notexture, *glowtex = NULL, *layertex = NULL;
+                        Slot &slot = lookuptexture(ti, false);
                         if(slot.sts.empty()) continue;
                         else if(slot.loaded) 
                         {
                             tex = slot.sts[0].t;
                             if(slot.texmask&(1<<TEX_GLOW)) { loopv(slot.sts) if(slot.sts[i].type==TEX_GLOW) { glowtex = slot.sts[i].t; break; } }
+                            if(slot.layer)
+                            {
+                                Slot &layer = lookuptexture(slot.layer);
+                                if(!layer.sts.empty()) layertex = layer.sts[0].t;
+                            }
                         }
                         else if(slot.thumbnail) tex = slot.thumbnail;
-                        else if(lastmillis-lastthumbnail>=thumbtime) { tex = loadthumbnail(slot); lastthumbnail = lastmillis; }
-                        if(g.texture(tex, 1.0, slot.rotation, slot.xoffset, slot.yoffset, glowtex, slot.glowcolor)&G3D_UP && (slot.loaded || tex!=notexture)) 
+                        else if(totalmillis-lastthumbnail>=thumbtime) { tex = loadthumbnail(slot); lastthumbnail = totalmillis; }
+                        if(g.texture(tex, 1.0, slot.rotation, slot.xoffset, slot.yoffset, glowtex, slot.glowcolor, layertex)&G3D_UP && (slot.loaded || tex!=notexture)) 
                             edittex(ti);
                     }
                     else
@@ -1671,19 +1706,24 @@ struct texturegui : g3d_callback
             }
         }
         g.end();
-        if(origtab != menutab) curtexindex = (menutab-1)*TEXTURE_WIDTH*TEXTURE_HEIGHT;
     }
 
     void showtextures(bool on)
     {
-        if(on != menuon && (menuon = on)) { menupos = menuinfrontofplayer(); menustart = starttime(); }
+        if(on != menuon && (menuon = on)) 
+        { 
+            if(menustart <= lasttexmillis) menutab = 1+clamp(lasttex, 0, curtexnum-1)/(TEXGUI_WIDTH*TEXGUI_HEIGHT);
+            menupos = menuinfrontofplayer(); 
+            menustart = starttime(); 
+        }
     }
 
     void show()
     {   
         if(!menuon) return;
         filltexlist();
-        if(!editmode || camera1->o.dist(menupos) > menuautoclose) menuon = false;
+        extern int usegui2d;
+        if(!editmode || ((!texgui2d || !usegui2d) && camera1->o.dist(menupos) > menuautoclose)) menuon = false;
         else g3d_addgui(this, menupos, texgui2d ? GUI_2D : 0);
     }
 } gui;
@@ -1702,14 +1742,14 @@ void showtexgui(int *n)
 // 0/noargs = toggle, 1 = on, other = off - will autoclose if too far away or exit editmode
 COMMAND(showtexgui, "i");
 
-void render_texture_panel(int w, int h)
+void rendertexturepanel(int w, int h)
 {
     if((texpaneltimer -= curtime)>0 && editmode)
     {
         glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-        glLoadIdentity();
-        int width = w*1800/h;
-        glOrtho(0, width, 1800, 0, -1, 1);
+
+        glPushMatrix();
+        glScalef(h/1800.0f, h/1800.0f, 1);
         int y = 50, gap = 10;
 
         static Shader *rgbonlyshader = NULL;
@@ -1723,13 +1763,18 @@ void render_texture_panel(int w, int h)
             if(ti>=0 && ti<curtexnum)
             {
                 Slot &slot = lookuptexture(texmru[ti]);
-                Texture *tex = slot.sts[0].t, *glowtex = NULL;
+                Texture *tex = slot.sts[0].t, *glowtex = NULL, *layertex = NULL;
                 if(slot.texmask&(1<<TEX_GLOW))
                 {
                     loopvj(slot.sts) if(slot.sts[j].type==TEX_GLOW) { glowtex = slot.sts[j].t; break; }
                 }
+                if(slot.layer)
+                {
+                    Slot &layer = lookuptexture(slot.layer);
+                    layertex = layer.sts[0].t;
+                }
                 float sx = min(1.0f, tex->xs/(float)tex->ys), sy = min(1.0f, tex->ys/(float)tex->xs);
-                int x = width-s-50, r = s;
+                int x = w*1800/h-s-50, r = s;
                 float tc[4][2] = { { 0, 0 }, { 1, 0 }, { 1, 1 }, { 0, 1 } };
                 float xoff = slot.xoffset, yoff = slot.yoffset;
                 if(slot.rotation)
@@ -1756,6 +1801,17 @@ void render_texture_panel(int w, int h)
                     glTexCoord2fv(tc[3]); glVertex2f(x,   y+r);
                     glEnd();
                     xtraverts += 4;
+                    if(j==1 && layertex)
+                    {
+                        glBindTexture(GL_TEXTURE_2D, layertex->id);
+                        glBegin(GL_QUADS);
+                        glTexCoord2fv(tc[0]); glVertex2f(x+r/2, y+r/2);
+                        glTexCoord2fv(tc[1]); glVertex2f(x+r,   y+r/2);
+                        glTexCoord2fv(tc[2]); glVertex2f(x+r,   y+r);
+                        glTexCoord2fv(tc[3]); glVertex2f(x+r/2, y+r);
+                        glEnd();
+                        xtraverts += 4;
+                    }
                     if(!j)
                     {
                         r -= 10;
@@ -1769,5 +1825,7 @@ void render_texture_panel(int w, int h)
         }
 
         defaultshader->set();
+
+        glPopMatrix();
     }
 }
