@@ -3,7 +3,6 @@
 // they "felt right", and have no basis in reality. Collision detection is simplistic but
 // very robust (uses discrete steps at fixed fps).
 
-#include "pch.h"
 #include "engine.h"
 
 const int MAXCLIPPLANES = 1024;
@@ -139,7 +138,7 @@ static float disttoent(octaentities *oc, octaentities *last, const vec &o, const
     int orient;
     float dist = 1e16f, f = 0.0f;
     if(oc == last || oc == NULL) return dist;
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
 
     #define entintersect(mask, type, func) {\
         if((mode&(mask))==(mask)) \
@@ -150,7 +149,8 @@ static float disttoent(octaentities *oc, octaentities *last, const vec &o, const
                     extentity &e = *ents[oc->type[i]]; \
                     if(!e.inoctanode || &e==t) continue; \
                     func; \
-                    if(f<dist && f>0) { \
+                    if(f<dist && f>0) \
+                    { \
                         hitentdist = dist = f; \
                         hitent = oc->type[i]; \
                         hitorient = orient; \
@@ -178,12 +178,34 @@ static float disttoent(octaentities *oc, octaentities *last, const vec &o, const
     return dist;
 }
 
+static float disttooutsideent(const vec &o, const vec &ray, float radius, int mode, extentity *t)
+{
+    vec eo, es;
+    int orient;
+    float dist = 1e16f, f = 0.0f;
+    const vector<extentity *> &ents = entities::getents();
+    loopv(outsideents)
+    {
+        extentity &e = *ents[outsideents[i]];
+        if(!e.inoctanode || &e == t) continue;
+        entselectionbox(e, eo, es);
+        if(!rayrectintersect(eo, es, o, ray, f, orient)) continue;
+        if(f<dist && f>0) 
+        {
+            hitentdist = dist = f;
+            hitent = outsideents[i];
+            hitorient = orient;
+        }
+    }
+    return dist;
+}
+         
 // optimized shadow version
 static float shadowent(octaentities *oc, octaentities *last, const vec &o, const vec &ray, float radius, int mode, extentity *t)
 {
     float dist = 1e16f, f = 0.0f;
     if(oc == last || oc == NULL) return dist;
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
     loopv(oc->mapmodels) if(!last || last->mapmodels.find(oc->mapmodels[i])<0)
     {
         extentity &e = *ents[oc->mapmodels[i]];
@@ -211,13 +233,13 @@ static float shadowent(octaentities *oc, octaentities *last, const vec &o, const
         loopi(3) \
         { \
             float c = v[i]; \
-            if(c<0 || c>=hdr.worldsize) \
+            if(c<0 || c>=worldsize) \
             { \
-                float d = ((invray[i]>0?0:hdr.worldsize)-c)*invray[i]; \
+                float d = ((invray[i]>0?0:worldsize)-c)*invray[i]; \
                 if(d<0) return (radius>0?radius:-1); \
                 disttoworld = max(disttoworld, 0.1f + d); \
             } \
-            float e = ((invray[i]>0?hdr.worldsize:0)-c)*invray[i]; \
+            float e = ((invray[i]>0?worldsize:0)-c)*invray[i]; \
             exitworld = min(exitworld, e); \
         } \
         if(disttoworld > exitworld) return (radius>0?radius:-1); \
@@ -259,7 +281,7 @@ static float shadowent(octaentities *oc, octaentities *last, const vec &o, const
         y = int(v.y); \
         z = int(v.z); \
         uint diff = uint(lo.x^x)|uint(lo.y^y)|uint(lo.z^z); \
-        if(diff >= uint(hdr.worldsize)) exitworld; \
+        if(diff >= uint(worldsize)) exitworld; \
         diff >>= lshift; \
         if(!diff) exitworld; \
         do \
@@ -343,21 +365,26 @@ float shadowray(const vec &o, const vec &ray, float radius, int mode, extentity
     }
 }
 
-float rayent(const vec &o, const vec &ray, vec &hitpos, float radius, int mode, int size, int &orient, int &ent)
+float rayent(const vec &o, const vec &ray, float radius, int mode, int size, int &orient, int &ent)
 {
     hitent = -1;
-    float d = raycubepos(o, ray, hitpos, hitentdist = radius, mode, size);
+    hitentdist = radius;
+    float dist = raycube(o, ray, radius, mode, size);
+    if((mode&RAY_ENTS) == RAY_ENTS)
+    {
+        float dent = disttooutsideent(o, ray, dist < 0 ? 1e16f : dist, mode, NULL);
+        if(dent < 1e15f && (dist < 0 || dent < dist)) dist = dent; 
+    }
     orient = hitorient;
-    ent = (hitentdist == d) ? hitent : -1;
-    return d;
+    ent = hitentdist == dist ? hitent : -1;
+    return dist;
 }
 
 float raycubepos(const vec &o, const vec &ray, vec &hitpos, float radius, int mode, int size)
 {
     hitpos = ray;
     float dist = raycube(o, ray, radius, mode, size);
-    hitpos.mul(dist); 
-    hitpos.add(o);
+    hitpos.mul(dist).add(o); 
     return dist; 
 }
 
@@ -393,28 +420,85 @@ const float FLOORZ = 0.867f;
 const float SLOPEZ = 0.5f;
 const float WALLZ = 0.2f;
 const float JUMPVEL = 125.0f;
-const float GRAVITY = 200.0f;
-const float STEPSPEED = 1.0f;
+extern const float GRAVITY = 200.0f;
 
-bool ellipsecollide(physent *d, const vec &dir, const vec &o, float yaw, float xr, float yr,  float hi, float lo)
+bool ellipserectcollide(physent *d, const vec &dir, const vec &o, const vec &center, float yaw, float xr, float yr, float hi, float lo)
 {
-    float below = (o.z-lo) - (d->o.z+d->aboveeye),
-          above = (d->o.z-d->eyeheight) - (o.z+hi);
+    float below = (o.z+center.z-lo) - (d->o.z+d->aboveeye),
+          above = (d->o.z-d->eyeheight) - (o.z+center.z+hi);
     if(below>=0 || above>=0) return true;
-    float x = o.x - d->o.x, y = o.y - d->o.y;
+
+    vec yo(d->o);
+    yo.sub(o);
+    yo.rotate_around_z(-yaw*RAD);
+    yo.sub(center);
+ 
+    float dx = clamp(yo.x, -xr, xr) - yo.x, dy = clamp(yo.y, -yr, yr) - yo.y,
+          dist = sqrtf(dx*dx + dy*dy) - d->radius;
+    if(dist < 0)
+    {
+        int sx = yo.x <= -xr ? -1 : (yo.x >= xr ? 1 : 0),
+            sy = yo.y <= -yr ? -1 : (yo.y >= yr ? 1 : 0);
+        if(dist > (yo.z < 0 ? below : above) && (sx || sy))
+        {
+            vec ydir(dir);
+            ydir.rotate_around_z(-yaw*RAD);
+            if(sx*yo.x - xr > sy*yo.y - yr)
+            {
+                if(dir.iszero() || sx*ydir.x < -1e-6f)
+                {
+                    wall = vec(sx, 0, 0);
+                    wall.rotate_around_z(yaw*RAD);
+                    return false;
+                }
+            }
+            else if(dir.iszero() || sy*ydir.y < -1e-6f)
+            { 
+                wall = vec(0, sy, 0);
+                wall.rotate_around_z(yaw*RAD);
+                return false;
+            }
+        }
+        if(yo.z < 0)
+        {
+            if(dir.iszero() || (dir.z > 0 && (d->type>=ENT_INANIMATE || below >= d->zmargin-(d->eyeheight+d->aboveeye)/4.0f)))
+            {
+                wall = vec(0, 0, -1);
+                return false;
+            }
+        }
+        else if(dir.iszero() || (dir.z < 0 && (d->type>=ENT_INANIMATE || above >= d->zmargin-(d->eyeheight+d->aboveeye)/3.0f)))
+        {
+            wall = vec(0, 0, 1);
+            return false;
+        }
+        inside = true;
+    }
+    return true;
+}
+
+bool ellipsecollide(physent *d, const vec &dir, const vec &o, const vec &center, float yaw, float xr, float yr, float hi, float lo)
+{
+    float below = (o.z+center.z-lo) - (d->o.z+d->aboveeye),
+          above = (d->o.z-d->eyeheight) - (o.z+center.z+hi);
+    if(below>=0 || above>=0) return true;
+    vec yo(center);
+    yo.rotate_around_z(yaw*RAD);
+    yo.add(o);
+    float x = yo.x - d->o.x, y = yo.y - d->o.y;
     float angle = atan2f(y, x), dangle = angle-(d->yaw+90)*RAD, eangle = angle-(yaw+90)*RAD;
     float dx = d->xradius*cosf(dangle), dy = d->yradius*sinf(dangle);
     float ex = xr*cosf(eangle), ey = yr*sinf(eangle);
     float dist = sqrtf(x*x + y*y) - sqrtf(dx*dx + dy*dy) - sqrtf(ex*ex + ey*ey);
     if(dist < 0)
     {
-        if(dist > (d->o.z < o.z ? below : above) && (dir.iszero() || x*dir.x + y*dir.y > 0))
+        if(dist > (d->o.z < yo.z ? below : above) && (dir.iszero() || x*dir.x + y*dir.y > 0))
         {
             wall = vec(-x, -y, 0);
             if(!wall.iszero()) wall.normalize();
             return false;
         }
-        if(d->o.z < o.z)
+        if(d->o.z < yo.z)
         {
             if(dir.iszero() || (dir.z > 0 && (d->type>=ENT_INANIMATE || below >= d->zmargin-(d->eyeheight+d->aboveeye)/4.0f)))
             {
@@ -434,7 +518,6 @@ bool ellipsecollide(physent *d, const vec &dir, const vec &o, float yaw, float x
 
 bool rectcollide(physent *d, const vec &dir, const vec &o, float xr, float yr,  float hi, float lo, uchar visible = 0xFF, bool collideonly = true, float cutoff = 0)
 {
-    if(collideonly && !visible) return true;
     vec s(d->o);
     s.sub(o);
     float dxr = d->collidetype==COLLIDE_ELLIPSE ? d->radius : d->xradius, dyr = d->collidetype==COLLIDE_ELLIPSE ? d->radius : d->yradius;
@@ -491,11 +574,11 @@ const vector<physent *> &checkdynentcache(int x, int y)
     dec.y = y;
     dec.frame = dynentframe;
     dec.dynents.setsize(0);
-    int numdyns = cl->numdynents(), dsize = 1<<dynentsize, dx = x<<dynentsize, dy = y<<dynentsize;
+    int numdyns = game::numdynents(), dsize = 1<<dynentsize, dx = x<<dynentsize, dy = y<<dynentsize;
     loopi(numdyns)
     {
-        dynent *d = cl->iterdynents(i);
-        if(!d || d->state != CS_ALIVE ||
+        dynent *d = game::iterdynents(i);
+        if(d->state != CS_ALIVE ||
            d->o.x+d->radius <= dx || d->o.x-d->radius >= dx+dsize ||
            d->o.y+d->radius <= dy || d->o.y-d->radius >= dy+dsize)
             continue;
@@ -505,8 +588,8 @@ const vector<physent *> &checkdynentcache(int x, int y)
 }
 
 #define loopdynentcache(curx, cury, o, radius) \
-    for(int curx = max(int(o.x-radius), 0)>>dynentsize, endx = min(int(o.x+radius), hdr.worldsize-1)>>dynentsize; curx <= endx; curx++) \
-    for(int cury = max(int(o.y-radius), 0)>>dynentsize, endy = min(int(o.y+radius), hdr.worldsize-1)>>dynentsize; cury <= endy; cury++)
+    for(int curx = max(int(o.x-radius), 0)>>dynentsize, endx = min(int(o.x+radius), worldsize-1)>>dynentsize; curx <= endx; curx++) \
+    for(int cury = max(int(o.y-radius), 0)>>dynentsize, endy = min(int(o.y+radius), worldsize-1)>>dynentsize; cury <= endy; cury++)
 
 void updatedynentcache(physent *d)
 {
@@ -551,7 +634,7 @@ bool plcollide(physent *d, const vec &dir)    // collide with player or monster
                     return false;
                 }
             }
-            else if(!ellipsecollide(d, dir, o->o, o->yaw, o->xradius, o->yradius, o->aboveeye, o->eyeheight))
+            else if(!ellipsecollide(d, dir, o->o, vec(0, 0, 0), o->yaw, o->xradius, o->yradius, o->aboveeye, o->eyeheight))
             {
                 hitplayer = o;
                 if((d->type==ENT_AI || d->type==ENT_INANIMATE) && wall.z>0) d->onplayer = o;
@@ -594,7 +677,7 @@ void rotatebb(vec &center, vec &radius, int yaw)
 
 bool mmcollide(physent *d, const vec &dir, octaentities &oc)               // collide with a mapmodel
 {   
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
     loopv(oc.mapmodels)
     {
         extentity &e = *ents[oc.mapmodels[i]];
@@ -603,12 +686,20 @@ bool mmcollide(physent *d, const vec &dir, octaentities &oc)               // co
         if(!m || !m->collide) continue;
         vec center, radius;
         m->collisionbox(0, center, radius);
-        if(!m->ellipsecollide || d->collidetype!=COLLIDE_ELLIPSE)
+        if(d->collidetype==COLLIDE_ELLIPSE)
+        {
+            float yaw = 180 + float((e.attr1+7)-(e.attr1+7)%15);
+            if(m->ellipsecollide)
+            {
+                if(!ellipsecollide(d, dir, e.o, center, yaw, radius.x, radius.y, radius.z, radius.z)) return false;
+            }
+            else if(!ellipserectcollide(d, dir, e.o, center, yaw, radius.x, radius.y, radius.z, radius.z)) return false;
+        } 
+        else
         {
             rotatebb(center, radius, e.attr1);
             if(!rectcollide(d, dir, center.add(e.o), radius.x, radius.y, radius.z, radius.z)) return false;
         }
-        else if(!ellipsecollide(d, dir, center.add(e.o), float((e.attr1+7)-(e.attr1+7)%15), radius.x, radius.y, radius.z, radius.z)) return false;
     }
     return true;
 }
@@ -707,7 +798,7 @@ static inline bool octacollide(physent *d, const vec &dir, float cutoff, const i
             if(c[i].ext) switch(c[i].ext->material&MATF_CLIP)
             {
                 case MAT_NOCLIP: continue;
-                case MAT_AICLIP: if(d->type==ENT_AI) solid = true; break;
+                case MAT_GAMECLIP: if(d->type==ENT_AI) solid = true; break;
                 case MAT_CLIP: if(isclipped(c[i].ext->material&MATF_VOLUME) || d->type<ENT_CAMERA) solid = true; break;
             }
             if(!solid && isempty(c[i])) continue;
@@ -721,8 +812,8 @@ static inline bool octacollide(physent *d, const vec &dir, float cutoff, const i
 {
     int diff = (bo.x^(bo.x+bs.x)) | (bo.y^(bo.y+bs.y)) | (bo.z^(bo.z+bs.z)),
         scale = worldscale-1;
-    if(diff&~((1<<scale)-1) || uint(bo.x|bo.y|bo.z|(bo.x+bs.x)|(bo.y+bs.y)|(bo.z+bs.z)) >= uint(hdr.worldsize))
-       return octacollide(d, dir, cutoff, bo, bs, worldroot, ivec(0, 0, 0), hdr.worldsize>>1);
+    if(diff&~((1<<scale)-1) || uint(bo.x|bo.y|bo.z|(bo.x+bs.x)|(bo.y+bs.y)|(bo.z+bs.z)) >= uint(worldsize))
+       return octacollide(d, dir, cutoff, bo, bs, worldroot, ivec(0, 0, 0), worldsize>>1);
     cube *c = &worldroot[octastep(bo.x, bo.y, bo.z, scale)];
     if(c->ext && c->ext->ents && !mmcollide(d, dir, *c->ext->ents)) return false;
     scale--;
@@ -737,7 +828,7 @@ static inline bool octacollide(physent *d, const vec &dir, float cutoff, const i
     if(c->ext) switch(c->ext->material&MATF_CLIP)
     {
         case MAT_NOCLIP: return true;
-        case MAT_AICLIP: if(d->type==ENT_AI) solid = true; break;
+        case MAT_GAMECLIP: if(d->type==ENT_AI) solid = true; break;
         case MAT_CLIP: if(isclipped(c->ext->material&MATF_VOLUME) || d->type<ENT_CAMERA) solid = true; break;
     }
     if(!solid && isempty(*c)) return true;
@@ -754,7 +845,7 @@ bool collide(physent *d, const vec &dir, float cutoff, bool playercol)
     ivec bo(int(d->o.x-d->radius), int(d->o.y-d->radius), int(d->o.z-d->eyeheight)),
          bs(int(d->radius)*2, int(d->radius)*2, int(d->eyeheight+d->aboveeye));
     bs.add(2);  // guard space for rounding errors
-    if(!octacollide(d, dir, cutoff, bo, bs)) return false;//, worldroot, ivec(0, 0, 0), hdr.worldsize>>1)) return false; // collide with world
+    if(!octacollide(d, dir, cutoff, bo, bs)) return false;//, worldroot, ivec(0, 0, 0), worldsize>>1)) return false; // collide with world
     return !playercol || plcollide(d, dir);
 }
 
@@ -770,10 +861,10 @@ void recalcdir(physent *d, const vec &oldvel, vec &dir)
     }
 }
 
-void slideagainst(physent *d, vec &dir, const vec &obstacle, bool foundfloor)
+void slideagainst(physent *d, vec &dir, const vec &obstacle, bool foundfloor, bool slidecollide)
 {
     vec wall(obstacle);
-    if(foundfloor && wall.z)
+    if(foundfloor ? wall.z > 0 : slidecollide)
     {
         wall.z = 0;
         if(!wall.iszero()) wall.normalize();
@@ -801,32 +892,83 @@ void switchfloor(physent *d, vec &dir, const vec &floor)
     recalcdir(d, oldvel, dir);
 }
 
-bool trystepup(physent *d, vec &dir, float maxstep)
+bool trystepup(physent *d, vec &dir, const vec &obstacle, float maxstep, const vec &floor)
 {
-    vec old(d->o);
+    vec old(d->o), stairdir = (obstacle.z >= 0 && obstacle.z < SLOPEZ ? vec(-obstacle.x, -obstacle.y, 0) : vec(dir.x, dir.y, 0)).rescale(1);
+    bool cansmooth = true;
     /* check if there is space atop the stair to move to */
     if(d->physstate != PHYS_STEP_UP)
     {
-        d->o.add(dir);
-        d->o.z += maxstep + 0.1f;
+        vec checkdir = stairdir;
+        checkdir.mul(0.1f);
+        checkdir.z += maxstep + 0.1f;
+        d->o.add(checkdir);
         if(!collide(d))
         {
             d->o = old;
-            return false;
+            if(collide(d, vec(0, 0, -1), SLOPEZ)) return false;
+            cansmooth = false;
         }
     }
+
+    if(cansmooth)
+    {
+        d->o = old;
+        vec checkdir = stairdir;
+        checkdir.z += 1;
+        checkdir.mul(maxstep);
+        d->o.add(checkdir);
+        if(!collide(d, checkdir))
+        {
+            if(collide(d, vec(0, 0, -1), SLOPEZ))
+            {
+                d->o = old;
+                return false;
+            }
+        }
+
+        /* try stepping up half as much as forward */
+        d->o = old;
+        vec smoothdir(dir.x, dir.y, 0);
+        float magxy = smoothdir.magnitude();
+        if(magxy > 1e-9f)
+        {
+            if(magxy > 2*dir.z) 
+            {
+                smoothdir.mul(1/magxy);
+                smoothdir.z = 0.5f;
+                smoothdir.mul(dir.magnitude()/smoothdir.magnitude());
+            }
+            else smoothdir.z = dir.z;
+            d->o.add(smoothdir);
+            d->o.z += maxstep + 0.1f;
+            if(collide(d, smoothdir))
+            {
+                d->o.z -= maxstep + 0.1f;
+                if(d->physstate == PHYS_FALL || d->floor != floor)
+                {
+                    d->timeinair = 0;
+                    d->floor = floor;
+                    switchfloor(d, dir, d->floor);
+                }
+                d->physstate = PHYS_STEP_UP;
+                return true;
+            }
+        }
+    }
+
     /* try stepping up */
-    d->o = old;
-    d->o.z += dir.magnitude()*STEPSPEED;
+    d->o = old;     
+    d->o.z += dir.magnitude();
     if(collide(d, vec(0, 0, 1)))
     {
-        if(d->physstate == PHYS_FALL)
+        if(d->physstate == PHYS_FALL || d->floor != floor)
         {
             d->timeinair = 0;
-            d->floor = vec(0, 0, 1);
+            d->floor = floor;
             switchfloor(d, dir, d->floor);
         }
-        d->physstate = PHYS_STEP_UP;
+        if(cansmooth) d->physstate = PHYS_STEP_UP;
         return true;
     }
     d->o = old;
@@ -877,7 +1019,7 @@ void falling(physent *d, vec &dir, const vec &floor)
     else d->physstate = PHYS_FALL;
 }
 
-void landing(physent *d, vec &dir, const vec &floor)
+void landing(physent *d, vec &dir, const vec &floor, bool collided)
 {
 #if 0
     if(d->physstate == PHYS_FALL)
@@ -888,8 +1030,8 @@ void landing(physent *d, vec &dir, const vec &floor)
 #endif
     switchfloor(d, dir, floor);
     d->timeinair = 0;
-    if(floor.z >= FLOORZ) d->physstate = PHYS_FLOOR;
-    else d->physstate = PHYS_SLOPE;
+    if(d->physstate!=PHYS_STEP_UP || !collided)
+        d->physstate = floor.z >= FLOORZ ? PHYS_FLOOR : PHYS_SLOPE;
     d->floor = floor;
 }
 
@@ -919,7 +1061,6 @@ bool findfloor(physent *d, bool collided, const vec &obstacle, bool &slide, vec
     }
     else if(d->physstate >= PHYS_SLOPE && d->floor.z < 1.0f)
     {
-        d->o.z -= d->radius;
         if(!collide(d, vec(d->floor).neg(), 0.95f) || !collide(d, vec(0, 0, -1)))
         {
             floor = wall;
@@ -939,9 +1080,9 @@ bool move(physent *d, vec &dir)
 {
     vec old(d->o);
 #if 0
-    if(d->physstate == PHYS_STEP_DOWN && dir.z <= 0.0f && cl->allowmove(pl) && (d->move || d->strafe))
+    if(d->physstate == PHYS_STEP_DOWN && dir.z <= 0.0f && game::allowmove(pl) && (d->move || d->strafe))
     {
-        float step = dir.magnitude()*STEPSPEED;
+        float step = dir.magnitude();
         if(trystepdown(d, dir, step, 0.75f, 0.25f)) return true;
         if(trystepdown(d, dir, step, 0.5f, 0.5f)) return true;
         if(trystepdown(d, dir, step, 0.25f, 0.75f)) return true;
@@ -953,7 +1094,7 @@ bool move(physent *d, vec &dir)
     bool collided = false, slidecollide = false;
     vec obstacle;
     d->o.add(dir);
-    if(!collide(d, d->type!=ENT_CAMERA ? dir : vec(0, 0, 0)) || ((d->type==ENT_AI || d->type==ENT_INANIMATE) && !collide(d)))
+    if(!collide(d, dir) || ((d->type==ENT_AI || d->type==ENT_INANIMATE) && !collide(d, vec(0, 0, 0), 0, false)))
     {
         obstacle = wall;
         /* check to see if there is an obstacle that would prevent this one from being used as a floor (or ceiling bump) */
@@ -963,16 +1104,13 @@ bool move(physent *d, vec &dir)
             obstacle = wall;
         }
         d->o = old;
-        if(d->type == ENT_CAMERA) return false;
-        float stepdist = (d->physstate >= PHYS_SLOPE && d->floor.z < 1.0f ? d->radius+0.1f : STAIRHEIGHT);
-        d->o.z -= stepdist;
-        d->zmargin = -stepdist;
+        d->o.z -= STAIRHEIGHT;
+        d->zmargin = -STAIRHEIGHT;
         if(d->physstate == PHYS_SLOPE || d->physstate == PHYS_FLOOR || (!collide(d, vec(0, 0, -1), SLOPEZ) && (d->physstate==PHYS_STEP_UP || wall.z>=FLOORZ)))
         {
             d->o = old;
             d->zmargin = 0;
-            float floorz = (d->physstate == PHYS_SLOPE || d->physstate == PHYS_FLOOR ? d->floor.z : wall.z);
-            if(trystepup(d, dir, floorz < 1.0f ? d->radius+0.1f : STAIRHEIGHT)) return true;
+            if(trystepup(d, dir, obstacle, STAIRHEIGHT, d->physstate == PHYS_SLOPE || d->physstate == PHYS_FLOOR ? d->floor : vec(wall))) return true;
         }
         else 
         {
@@ -982,25 +1120,32 @@ bool move(physent *d, vec &dir)
         /* can't step over the obstacle, so just slide against it */
         collided = true;
     }
+    else if(d->physstate == PHYS_STEP_UP) 
+    {
+        if(!collide(d, vec(0, 0, -1), SLOPEZ))
+        {
+            d->o = old;
+            if(trystepup(d, dir, vec(0, 0, 1), STAIRHEIGHT, vec(wall))) return true;
+            d->o.add(dir);
+        }
+    }
     vec floor(0, 0, 0);
     bool slide = collided,
          found = findfloor(d, collided, obstacle, slide, floor);
     if(slide || (!collided && floor.z > 0 && floor.z < WALLZ))
     {
-        slideagainst(d, dir, slide ? obstacle : floor, found || slidecollide);
+        slideagainst(d, dir, slide ? obstacle : floor, found, slidecollide);
         if(d->type == ENT_AI || d->type == ENT_INANIMATE) d->blocked = true;
     }
-    if(found)
-    {
-        if(d->type == ENT_CAMERA) return false;
-        landing(d, dir, floor);
-    }
+    if(found) landing(d, dir, floor, collided);
     else falling(d, dir, floor);
     return !collided;
 }
 
 bool bounce(physent *d, float secs, float elasticity, float waterfric)
 {
+    // make sure bouncers don't start inside geometry
+    if(d->physstate!=PHYS_BOUNCE && !collide(d, vec(0, 0, 0), 0, false)) return true;
     int mat = lookupmaterial(vec(d->o.x, d->o.y, d->o.z + (d->aboveeye - d->eyeheight)/2));
     bool water = isliquid(mat);
     if(water) 
@@ -1065,20 +1210,40 @@ void avoidcollision(physent *d, const vec &dir, physent *obstacle, float space)
     if(mindist >= 0.0f && mindist < 1e15f) d->o.add(vec(dir).mul(mindist));
 }
 
+bool movecamera(physent *pl, const vec &dir, float dist, float stepdist)
+{
+    int steps = (int)ceil(dist/stepdist);
+    if(steps <= 0) return true;
+
+    vec d(dir);
+    d.mul(dist/steps);
+    loopi(steps)
+    {
+        vec oldpos(pl->o);
+        pl->o.add(d);
+        if(!collide(pl, vec(0, 0, 0), 0, false))
+        {
+            pl->o = oldpos;
+            return false;
+        }
+    }
+    return true;
+}
+
 bool droptofloor(vec &o, float radius, float height)
 {
     if(!insideworld(o)) return false;
     vec v(0.0001f, 0.0001f, -1);
     v.normalize();
-    if(raycube(o, v, hdr.worldsize) >= hdr.worldsize) return false;
+    if(raycube(o, v, worldsize) >= worldsize) return false;
     physent d;
     d.type = ENT_CAMERA;
     d.o = o;
     d.vel = vec(0, 0, -1);
-    d.radius = radius;
+    d.radius = d.xradius = d.yradius = radius;
     d.eyeheight = height;
     d.aboveeye = radius;
-    loopi(hdr.worldsize) if(!move(&d, v))
+    if(!movecamera(&d, d.vel, worldsize, 1))
     {
         o = d.o;
         return true;
@@ -1086,9 +1251,21 @@ bool droptofloor(vec &o, float radius, float height)
     return false;
 }
 
+float dropheight(entity &e)
+{
+    switch(e.type)
+    {
+        case ET_PARTICLES:
+        case ET_MAPMODEL: return 0.0f;
+        default:
+            if(e.type >= ET_GAMESPECIFIC) return entities::dropheight(e);
+            return 4.0f;
+    }
+}
+
 void dropenttofloor(entity *e)
 {
-    droptofloor(e->o, 1.0f, et->dropheight(*e));
+    droptofloor(e->o, 1.0f, dropheight(*e));
 }
 
 void phystest()
@@ -1131,29 +1308,30 @@ void vectoyawpitch(const vec &v, float &yaw, float &pitch)
 }
 
 VARP(maxroll, 0, 3, 20);
+FVAR(straferoll, 0, 0.033f, 90);
 VAR(floatspeed, 10, 100, 1000);
 
 void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curtime)
 {
     if(floating)
     {
-        if(pl->jumpnext)
+        if(pl->jumping)
         {
-            pl->jumpnext = false;
+            pl->jumping = false;
             pl->vel.z = JUMPVEL;
         }
     }
     else if(pl->physstate >= PHYS_SLOPE || water)
     {
-        if(pl->type != ENT_CAMERA && water && !pl->inwater) pl->vel.div(8);
-        if(pl->jumpnext)
+        if(water && !pl->inwater) pl->vel.div(8);
+        if(pl->jumping)
         {
-            pl->jumpnext = false;
+            pl->jumping = false;
 
             pl->vel.z = JUMPVEL; // physics impulse upwards
             if(water) { pl->vel.x /= 8.0f; pl->vel.y /= 8.0f; } // dampen velocity change even harder, gives correct water feel
 
-            cl->physicstrigger(pl, local, 1, 0);
+            game::physicstrigger(pl, local, 1, 0);
         }
     }
     if(!floating && pl->physstate == PHYS_FALL) pl->timeinair += curtime;
@@ -1165,7 +1343,7 @@ void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curt
         if(d->rotspeed && d->yaw!=d->targetyaw)
         {
             float oldyaw = d->yaw, diff = d->rotspeed*curtime/1000.0f, maxdiff = fabs(d->targetyaw-d->yaw);
-            if(diff >= maxdiff) 
+            if(diff >= maxdiff)
             {
                 d->yaw = d->targetyaw;
                 d->rotspeed = 0;
@@ -1182,18 +1360,17 @@ void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curt
         }
     }
 
-    if(m.iszero() && cl->allowmove(pl) && (pl->move || pl->strafe))
+    if(m.iszero() && game::allowmove(pl) && (pl->move || pl->strafe))
     {
         vecfromyawpitch(pl->yaw, floating || water || pl->type==ENT_CAMERA ? pl->pitch : 0, pl->move, pl->strafe, m);
 
-        if(!floating && pl->physstate >= PHYS_SLIDE)
+        if(!floating && pl->physstate >= PHYS_SLOPE)
         {
             /* move up or down slopes in air
              * but only move up slopes in water
              */
             float dz = -(m.x*pl->floor.x + m.y*pl->floor.y)/pl->floor.z;
-            if(water) m.z = max(m.z, dz);
-            else if(pl->floor.z >= WALLZ) m.z = dz;
+            m.z = water ? max(m.z, dz) : dz;
         }
 
         m.normalize();
@@ -1201,17 +1378,20 @@ void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curt
 
     vec d(m);
     d.mul(pl->maxspeed);
-    if(floating) 
+    if(pl->type==ENT_PLAYER)
     {
-        if(pl==player) d.mul(floatspeed/100.0f);
+        if(floating) 
+        {
+            if(pl==player) d.mul(floatspeed/100.0f);
+        }
+        else if(!water && game::allowmove(pl)) d.mul((pl->move && !pl->strafe ? 1.3f : 1.0f) * (pl->physstate < PHYS_SLOPE ? 1.3f : 1.0f)); // EXPERIMENTAL
     }
-    else if(!water && cl->allowmove(pl)) d.mul((pl->move && !pl->strafe ? 1.3f : 1.0f) * (pl->physstate < PHYS_SLOPE ? 1.3f : 1.0f)); // EXPERIMENTAL
-    float friction = water && !floating ? 20.0f : (pl->physstate >= PHYS_SLOPE || floating ? 6.0f : 30.0f);
-    float fpsfric = friction/curtime*20.0f;
-
-    pl->vel.mul(fpsfric-1);
-    pl->vel.add(d);
-    pl->vel.div(fpsfric);
+    float fric = water && !floating ? 20.0f : (pl->physstate >= PHYS_SLOPE || floating ? 6.0f : 30.0f);
+    pl->vel.lerp(d, pl->vel, pow(1 - 1/fric, curtime/20.0f));
+// old fps friction
+//    float friction = water && !floating ? 20.0f : (pl->physstate >= PHYS_SLOPE || floating ? 6.0f : 30.0f);
+//    float fpsfric = min(curtime/(20.0f*friction), 1.0f);
+//    pl->vel.lerp(pl->vel, d, fpsfric);
 }
 
 void modifygravity(physent *pl, bool water, int curtime)
@@ -1226,14 +1406,18 @@ void modifygravity(physent *pl, bool water, int curtime)
         g.normalize();
         g.mul(GRAVITY*secs);
     }
-    if(!water || !cl->allowmove(pl) || (!pl->move && !pl->strafe)) pl->falling.add(g);
+    if(!water || !game::allowmove(pl) || (!pl->move && !pl->strafe)) pl->falling.add(g);
 
     if(water || pl->physstate >= PHYS_SLOPE)
     {
-        float friction = water ? 2.0f : 6.0f,
-              fpsfric = friction/curtime*20.0f,
+        float fric = water ? 2.0f : 6.0f,
               c = water ? 1.0f : clamp((pl->floor.z - SLOPEZ)/(FLOORZ-SLOPEZ), 0.0f, 1.0f);
-        pl->falling.mul(1 - c/fpsfric);
+        pl->falling.mul(pow(1 - c/fric, curtime/20.0f));
+// old fps friction
+//        float friction = water ? 2.0f : 6.0f,
+//              fpsfric = friction/curtime*20.0f,
+//              c = water ? 1.0f : clamp((pl->floor.z - SLOPEZ)/(FLOORZ-SLOPEZ), 0.0f, 1.0f);
+//        pl->falling.mul(1 - c/fpsfric);
     }
 }
 
@@ -1245,16 +1429,16 @@ bool moveplayer(physent *pl, int moveres, bool local, int curtime)
 {
     int material = lookupmaterial(vec(pl->o.x, pl->o.y, pl->o.z + (3*pl->aboveeye - pl->eyeheight)/4));
     bool water = isliquid(material&MATF_VOLUME);
-    bool floating = pl->state==CS_EDITING || (pl->type!=ENT_CAMERA && pl->state==CS_SPECTATOR);
+    bool floating = pl->type==ENT_PLAYER && (pl->state==CS_EDITING || pl->state==CS_SPECTATOR);
     float secs = curtime/1000.f;
 
     // apply gravity
-    if(!floating && pl->type!=ENT_CAMERA) modifygravity(pl, water, curtime);
+    if(!floating) modifygravity(pl, water, curtime);
     // apply any player generated changes in velocity
     modifyvelocity(pl, local, water, floating, curtime);
 
     vec d(pl->vel), oldpos(pl->o);
-    if(!floating && pl->type!=ENT_CAMERA && water) d.mul(0.5f);
+    if(!floating && water) d.mul(0.5f);
     d.add(pl->falling);
     d.mul(secs);
 
@@ -1279,14 +1463,14 @@ bool moveplayer(physent *pl, int moveres, bool local, int curtime)
         int collisions = 0;
 
         d.mul(f);
-        loopi(moveres) if(!move(pl, d)) { if(pl->type==ENT_CAMERA) return false; if(++collisions<5) i--; } // discrete steps collision detection & sliding
+        loopi(moveres) if(!move(pl, d) && ++collisions<5) i--; // discrete steps collision detection & sliding
         if(timeinair > 800 && !pl->timeinair && !water) // if we land after long time must have been a high jump, make thud sound
         {
-            cl->physicstrigger(pl, local, -1, 0);
+            game::physicstrigger(pl, local, -1, 0);
         }
     }
 
-    if(pl->type!=ENT_CAMERA && pl->state==CS_ALIVE) updatedynentcache(pl);
+    if(pl->state==CS_ALIVE) updatedynentcache(pl);
 
     if(!pl->timeinair && pl->physstate >= PHYS_FLOOR && pl->vel.squaredlen() < 1e-4f) pl->moving = false;
 
@@ -1301,26 +1485,23 @@ bool moveplayer(physent *pl, int moveres, bool local, int curtime)
     }
     else
     {
-        pl->roll += pl->strafe*curtime/-30.0f;
-        if(pl->roll>maxroll) pl->roll = (float)maxroll;
-        if(pl->roll<-maxroll) pl->roll = (float)-maxroll;
+        pl->roll -= pl->strafe*curtime*straferoll;
+        if(pl->roll > maxroll) pl->roll = maxroll;
+        else if(pl->roll < -maxroll) pl->roll = -maxroll;
     }
 
     // play sounds on water transitions
 
-    if(pl->type!=ENT_CAMERA)
+    if(pl->inwater && !water)
     {
-        if(pl->inwater && !water)
-        {
-            material = lookupmaterial(vec(pl->o.x, pl->o.y, pl->o.z + (pl->aboveeye - pl->eyeheight)/2));
-            water = isliquid(material&MATF_VOLUME);
-        }
-        if(!pl->inwater && water) cl->physicstrigger(pl, local, 0, -1, material&MATF_VOLUME);
-        else if(pl->inwater && !water) cl->physicstrigger(pl, local, 0, 1, pl->inwater);
-        pl->inwater = water ? material&MATF_VOLUME : MAT_AIR;
-
-        if(pl->state==CS_ALIVE && (pl->o.z < 0 || material&MAT_DEATH)) cl->suicide(pl);
+        material = lookupmaterial(vec(pl->o.x, pl->o.y, pl->o.z + (pl->aboveeye - pl->eyeheight)/2));
+        water = isliquid(material&MATF_VOLUME);
     }
+    if(!pl->inwater && water) game::physicstrigger(pl, local, 0, -1, material&MATF_VOLUME);
+    else if(pl->inwater && !water) game::physicstrigger(pl, local, 0, 1, pl->inwater);
+    pl->inwater = water ? material&MATF_VOLUME : MAT_AIR;
+
+    if(pl->state==CS_ALIVE && (pl->o.z < 0 || material&MAT_DEATH)) game::suicide(pl);
 
     return true;
 }
@@ -1331,7 +1512,7 @@ int physsteps = 0, physframetime = PHYSFRAMETIME, lastphysframe = 0;
 
 void physicsframe()          // optimally schedule physics frames inside the graphics frames
 {
-    int diff = lastmillis + curtime - lastphysframe;
+    int diff = lastmillis - lastphysframe;
     if(diff <= 0) physsteps = 0;
     else
     {
@@ -1349,7 +1530,7 @@ void interppos(physent *pl)
 {
     pl->o = pl->newpos;
 
-    int diff = lastphysframe - (lastmillis + curtime); 
+    int diff = lastphysframe - lastmillis;
     if(diff <= 0 || !physinterp) return;
         
     vec deltapos(pl->deltapos);
@@ -1412,25 +1593,19 @@ void updatephysstate(physent *d)
     {
         case PHYS_SLOPE:
         case PHYS_FLOOR:
-            d->o.z -= 0.1f;
+            d->o.z -= 0.15f;
             if(!collide(d, vec(0, 0, -1), d->physstate == PHYS_SLOPE ? SLOPEZ : FLOORZ))
                 d->floor = wall;
-            else if(d->physstate == PHYS_SLOPE)
-            {
-                d->o.z -= d->radius;
-                if(!collide(d, vec(0, 0, -1), SLOPEZ))
-                    d->floor = wall;
-            }
             break;
 
         case PHYS_STEP_UP:
-            d->o.z -= STAIRHEIGHT+0.1f;
+            d->o.z -= STAIRHEIGHT+0.15f;
             if(!collide(d, vec(0, 0, -1), SLOPEZ))
                 d->floor = wall;
             break;
 
         case PHYS_SLIDE:
-            d->o.z -= d->radius+0.1f;
+            d->o.z -= 0.15f;
             if(!collide(d, vec(0, 0, -1)) && wall.z < SLOPEZ)
                 d->floor = wall;
             break;
@@ -1439,35 +1614,6 @@ void updatephysstate(physent *d)
     d->o = old;
 }
 
-bool intersect(physent *d, vec &from, vec &to)   // if lineseg hits entity bounding box
-{
-    vec v = to, w = d->o, *p;
-    v.sub(from);
-    w.sub(from);
-    float c1 = w.dot(v);
-
-    if(c1<=0) p = &from;
-    else
-    {
-        float c2 = v.dot(v);
-        if(c2<=c1) p = &to;
-        else
-        {
-            float f = c1/c2;
-            v.mul(f);
-            v.add(from);
-            p = &v;
-        }
-    }
-
-    return p->x <= d->o.x+d->radius
-        && p->x >= d->o.x-d->radius
-        && p->y <= d->o.y+d->radius
-        && p->y >= d->o.y-d->radius
-        && p->z <= d->o.z+d->aboveeye
-        && p->z >= d->o.z-d->eyeheight;
-}
-
 const float PLATFORMMARGIN = 0.2f;
 const float PLATFORMBORDER = 10.0f;
 
@@ -1490,28 +1636,28 @@ bool platformcollide(physent *d, physent *o, const vec &dir, float margin = 0)
             o->eyeheight + margin))
             return true;
     }
-    else if(ellipsecollide(d, dir, o->o, o->yaw, o->xradius, o->yradius, o->aboveeye, o->eyeheight + margin))
+    else if(ellipsecollide(d, dir, o->o, vec(0, 0, 0), o->yaw, o->xradius, o->yradius, o->aboveeye, o->eyeheight + margin))
         return true;
     return false;
 }
 
 bool moveplatform(physent *p, const vec &dir)
-{   
+{
     if(!insideworld(p->newpos)) return false;
 
     vec oldpos(p->o);
     (p->o = p->newpos).add(dir);
     if(!collide(p, dir, 0, dir.z<=0))
     {
-        p->o = oldpos; 
+        p->o = oldpos;
         return false;
     }
-    p->o = oldpos; 
+    p->o = oldpos;
 
     static vector<physent *> candidates;
     candidates.setsizenodelete(0);
-    for(int x = int(max(p->o.x-p->radius-PLATFORMBORDER, 0.0f))>>dynentsize, ex = int(min(p->o.x+p->radius+PLATFORMBORDER, hdr.worldsize-1.0f))>>dynentsize; x <= ex; x++)
-    for(int y = int(max(p->o.y-p->radius-PLATFORMBORDER, 0.0f))>>dynentsize, ey = int(min(p->o.y+p->radius+PLATFORMBORDER, hdr.worldsize-1.0f))>>dynentsize; y <= ey; y++)
+    for(int x = int(max(p->o.x-p->radius-PLATFORMBORDER, 0.0f))>>dynentsize, ex = int(min(p->o.x+p->radius+PLATFORMBORDER, worldsize-1.0f))>>dynentsize; x <= ex; x++)
+    for(int y = int(max(p->o.y-p->radius-PLATFORMBORDER, 0.0f))>>dynentsize, ey = int(min(p->o.y+p->radius+PLATFORMBORDER, worldsize-1.0f))>>dynentsize; y <= ey; y++)
     {
         const vector<physent *> &dynents = checkdynentcache(x, y);
         loopv(dynents)
@@ -1614,8 +1760,8 @@ dir(forward,  move,    1, k_up,    k_down);
 dir(left,     strafe,  1, k_left,  k_right);
 dir(right,    strafe, -1, k_right, k_left);
 
-ICOMMAND(jump,   "D", (int *down), { if(cl->canjump()) player->jumpnext = *down!=0; });
-ICOMMAND(attack, "D", (int *down), { cl->doattack(*down!=0); });
+ICOMMAND(jump,   "D", (int *down), { if(!*down || game::canjump()) player->jumping = *down!=0; });
+ICOMMAND(attack, "D", (int *down), { game::doattack(*down!=0); });
 
 bool entinmap(dynent *d, bool avoidplayers)        // brute force but effective way to find a free spawn spot in the map
 {
