@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 #include "SDL_thread.h"
 
@@ -68,7 +67,7 @@ static bool mergepvsnodes(pvsnode &p, pvsnode *children)
     return true;
 }
 
-static void genpvsnodes(cube *c, int parent = 0, const ivec &co = ivec(0, 0, 0), int size = hdr.worldsize/2)
+static void genpvsnodes(cube *c, int parent = 0, const ivec &co = ivec(0, 0, 0), int size = worldsize/2)
 {
     int index = origpvsnodes.length();
     loopi(8)
@@ -329,7 +328,7 @@ struct pvsworker
     int hasvoxel(const ivec &p, int coord, int dir, int ocoord = 0, int odir = 0, int *omin = NULL)
     {
         uint diff = (origin.x^p.x)|(origin.y^p.y)|(origin.z^p.z);
-        if(diff >= uint(hdr.worldsize)) return 0;
+        if(diff >= uint(worldsize)) return 0;
         diff >>= curlevel;
         while(diff)
         {
@@ -396,7 +395,7 @@ struct pvsworker
         if(p.edges.x!=0xFF) p.flags |= PVS_HIDE_GEOM;
     }
 
-    void shaftcullpvs(shaft &s, pvsnode &p, const ivec &co = ivec(0, 0, 0), int size = hdr.worldsize)
+    void shaftcullpvs(shaft &s, pvsnode &p, const ivec &co = ivec(0, 0, 0), int size = worldsize)
     {
         if(p.flags&PVS_HIDE_BB) return;
         shaftbb bb(co, size);
@@ -422,7 +421,7 @@ struct pvsworker
 
     ringbuf<shaftbb, 32> prevblockers;
 
-    void cullpvs(pvsnode &p, const ivec &co = ivec(0, 0, 0), int size = hdr.worldsize)
+    void cullpvs(pvsnode &p, const ivec &co = ivec(0, 0, 0), int size = worldsize)
     {
         if(p.flags&(PVS_HIDE_BB | PVS_HIDE_GEOM) || genpvs_canceled) return;
         if(p.children && !(p.flags&PVS_HIDE_BB))
@@ -702,7 +701,7 @@ struct pvsworker
             bbsize[R[dim]] = m.rsize;
             bborigin[dim] -= 2;
             bbsize[dim] = 2;
-            if(!materialoccluded(pvsnodes[0], vec(0, 0, 0), hdr.worldsize/2, bborigin, bbsize)) return false;
+            if(!materialoccluded(pvsnodes[0], vec(0, 0, 0), worldsize/2, bborigin, bbsize)) return false;
         }
         return true;
     }
@@ -732,7 +731,7 @@ struct pvsworker
         waterbytes = 0;
         loopi(4) if(wateroccluded&(0xFF<<(i*8))) waterbytes = i+1;
 
-        compresspvs(pvsnodes[0], hdr.worldsize, pvsleafsize);
+        compresspvs(pvsnodes[0], worldsize, pvsleafsize);
         outbuf.setsizenodelete(0);
         serializepvs(pvsnodes[0]);
     }
@@ -820,13 +819,11 @@ static int totalviewcells = 0;
 
 static void show_genpvs_progress(int unique = pvs.length(), int processed = numviewcells)
 {
-    float bar1 = float(processed) / float(totalviewcells>0 ? totalviewcells : 1),
-          bar2 = float(unique) / float(totalviewcells>0 ? totalviewcells : 1);
+    float bar1 = float(processed) / float(totalviewcells>0 ? totalviewcells : 1);
 
-    s_sprintfd(text1)("%d%% - %d of %d view cells", int(bar1 * 100), processed, totalviewcells);
-    s_sprintfd(text2)("%d unique view cells", unique);
+    defformatstring(text1)("%d%% - %d of %d view cells (%d unique)", int(bar1 * 100), processed, totalviewcells, unique);
 
-    show_out_of_renderloop_progress(bar1, text1, bar2, text2);
+    renderprogress(bar1, text1);
 
     if(interceptkey(SDLK_ESCAPE)) genpvs_canceled = true;
     check_genpvs_progress = false;
@@ -929,7 +926,7 @@ static int curwaterpvs = 0, lockedwaterpvs = 0;
 static inline pvsdata *lookupviewcell(const vec &p)
 {
     uint x = uint(floor(p.x)), y = uint(floor(p.y)), z = uint(floor(p.z));
-    if(!viewcells || (x|y|z)>=uint(hdr.worldsize)) return NULL;
+    if(!viewcells || (x|y|z)>=uint(worldsize)) return NULL;
     viewcellnode *vc = viewcells;
     for(int scale = worldscale-1; scale>=0; scale--)
     {
@@ -1076,17 +1073,17 @@ COMMAND(testpvs, "i");
 
 void genpvs(int *viewcellsize)
 {
-    if(hdr.worldsize > 1<<15)
+    if(worldsize > 1<<15)
     {
         conoutf(CON_ERROR, "map is too large for PVS");
         return;
     }
 
-    computescreen("generating PVS (esc to abort)");
+    renderbackground("generating PVS (esc to abort)");
     genpvs_canceled = false;
     Uint32 start = SDL_GetTicks();
 
-    show_out_of_renderloop_progress(0, "finding view cells");
+    renderprogress(0, "finding view cells");
 
     clearpvs();
     calcpvsbounds();
@@ -1098,7 +1095,7 @@ void genpvs(int *viewcellsize)
     root.children = 0;
     genpvsnodes(worldroot);
 
-    totalviewcells = countviewcells(worldroot, ivec(0, 0, 0), hdr.worldsize>>1, *viewcellsize>0 ? *viewcellsize : 32);
+    totalviewcells = countviewcells(worldroot, ivec(0, 0, 0), worldsize>>1, *viewcellsize>0 ? *viewcellsize : 32);
     numviewcells = 0;
     genpvs_canceled = false;
     check_genpvs_progress = false;
@@ -1109,14 +1106,14 @@ void genpvs(int *viewcellsize)
         timer = SDL_AddTimer(500, genpvs_timer, NULL);
     }
     viewcells = new viewcellnode;
-    genviewcells(*viewcells, worldroot, ivec(0, 0, 0), hdr.worldsize>>1, *viewcellsize>0 ? *viewcellsize : 32);
+    genviewcells(*viewcells, worldroot, ivec(0, 0, 0), worldsize>>1, *viewcellsize>0 ? *viewcellsize : 32);
     if(pvsthreads<=1)
     {
         SDL_RemoveTimer(timer);
     }
     else
     {
-        show_out_of_renderloop_progress(0, "creating threads");
+        renderprogress(0, "creating threads");
         if(!pvsmutex) pvsmutex = SDL_CreateMutex();
         if(!viewcellmutex) viewcellmutex = SDL_CreateMutex();
         loopi(pvsthreads)
@@ -1220,91 +1217,63 @@ bool waterpvsoccluded(int height)
     return false;
 }
 
-void saveviewcells(gzFile f, viewcellnode &p)
+void saveviewcells(stream *f, viewcellnode &p)
 {
-    gzputc(f, p.leafmask);
+    f->putchar(p.leafmask);
     loopi(8)
     {
-        if(p.leafmask&(1<<i))
-        {
-            int pvsindex = p.children[i].pvs;
-            endianswap(&pvsindex, sizeof(int), 1);
-            gzwrite(f, &pvsindex, sizeof(int));
-        }
+        if(p.leafmask&(1<<i)) f->putlil<int>(p.children[i].pvs);
         else saveviewcells(f, *p.children[i].node);
     }
 }
 
-void savepvs(gzFile f)
+void savepvs(stream *f)
 {
     uint totallen = pvsbuf.length() | (numwaterplanes>0 ? 0x80000000U : 0);
-    endianswap(&totallen, sizeof(uint), 1);
-    gzwrite(f, &totallen, sizeof(uint));
+    f->putlil<uint>(totallen);
     if(numwaterplanes>0)
     {
-        uint numwp = numwaterplanes;
-        endianswap(&numwp, sizeof(uint), 1);
-        gzwrite(f, &numwp, sizeof(uint));
+        f->putlil<uint>(numwaterplanes);
         loopi(numwaterplanes)
         {
-            int height = waterplanes[i].height;
-            endianswap(&height, sizeof(int), 1);
-            gzwrite(f, &height, sizeof(int));
+            f->putlil<int>(waterplanes[i].height);
             if(waterplanes[i].height < 0) break;
         }
     }
-    loopv(pvs)
-    {
-        ushort len = pvs[i].len;
-        endianswap(&len, sizeof(ushort), 1);
-        gzwrite(f, &len, sizeof(ushort));
-    }
-    gzwrite(f, pvsbuf.getbuf(), pvsbuf.length());
+    loopv(pvs) f->putlil<ushort>(pvs[i].len);
+    f->write(pvsbuf.getbuf(), pvsbuf.length());
     saveviewcells(f, *viewcells);
 }
 
-viewcellnode *loadviewcells(gzFile f)
+viewcellnode *loadviewcells(stream *f)
 {
     viewcellnode *p = new viewcellnode;
-    p->leafmask = gzgetc(f);
+    p->leafmask = f->getchar();
     loopi(8)
     {
-        if(p->leafmask&(1<<i))
-        {
-            gzread(f, &p->children[i].pvs, sizeof(int));
-            endianswap(&p->children[i].pvs, sizeof(int), 1);
-        }
+        if(p->leafmask&(1<<i)) p->children[i].pvs = f->getlil<int>();
         else p->children[i].node = loadviewcells(f);
     }
     return p;
 }
 
-void loadpvs(gzFile f)
+void loadpvs(stream *f, int numpvs)
 {
-    uint totallen = pvsbuf.length();
-    gzread(f, &totallen, sizeof(uint));
-    endianswap(&totallen, sizeof(uint), 1);
+    uint totallen = f->getlil<uint>();
     if(totallen & 0x80000000U)
     {
         totallen &= ~0x80000000U;
-        gzread(f, &numwaterplanes, sizeof(uint));
-        endianswap(&numwaterplanes, sizeof(uint), 1);
-        loopi(numwaterplanes)
-        {
-            gzread(f, &waterplanes[i].height, sizeof(int));
-            endianswap(&waterplanes[i].height, sizeof(int), 1);
-        }
+        numwaterplanes = f->getlil<uint>();
+        loopi(numwaterplanes) waterplanes[i].height = f->getlil<int>();
     }
     int offset = 0;
-    loopi(hdr.numpvs)
+    loopi(numpvs)
     {
-        ushort len;
-        gzread(f, &len, sizeof(ushort));
-        endianswap(&len, sizeof(ushort), 1);
+        ushort len = f->getlil<ushort>();
         pvs.add(pvsdata(offset, len));
         offset += len;
     }
-    gzread(f, pvsbuf.reserve(totallen).buf, totallen);
+    f->read(pvsbuf.reserve(totallen).buf, totallen);
     pvsbuf.advance(totallen);
     viewcells = loadviewcells(f);
 }
