@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 
 VARP(oqdynent, 0, 1, 1);
@@ -6,6 +5,7 @@ VARP(animationinterpolationtime, 0, 150, 1000);
 
 model *loadingmodel = NULL;
 
+#include "ragdoll.h"
 #include "animmodel.h"
 #include "vertmodel.h"
 #include "skelmodel.h"
@@ -19,7 +19,7 @@ model *loadingmodel = NULL;
 void mdlcullface(int *cullface)
 {
     checkmdl;
-    loadingmodel->cullface = *cullface!=0;
+    loadingmodel->setcullface(*cullface!=0);
 }
 
 COMMAND(mdlcullface, "i");
@@ -78,6 +78,14 @@ void mdlalphablend(int *blend)
 
 COMMAND(mdlalphablend, "i");
 
+void mdlalphadepth(int *depth)
+{
+    checkmdl;
+    loadingmodel->alphadepth = *depth!=0;
+}
+
+COMMAND(mdlalphadepth, "i");
+
 void mdlglow(int *percent)
 {
     checkmdl;
@@ -105,14 +113,6 @@ void mdlenvmap(float *envmapmax, float *envmapmin, char *envmap)
 
 COMMAND(mdlenvmap, "ffs");
 
-void mdltranslucent(float *translucency)
-{
-    checkmdl;
-    loadingmodel->settranslucency(*translucency);
-}
-
-COMMAND(mdltranslucent, "f");
-
 void mdlfullbright(float *fullbright)
 {
     checkmdl;
@@ -129,13 +129,14 @@ void mdlshader(char *shader)
 
 COMMAND(mdlshader, "s");
 
-void mdlspin(float *rate)
+void mdlspin(float *yaw, float *pitch)
 {
     checkmdl;
-    loadingmodel->spin = *rate;
+    loadingmodel->spinyaw = *yaw;
+    loadingmodel->spinpitch = *pitch;
 }
 
-COMMAND(mdlspin, "f");
+COMMAND(mdlspin, "ff");
 
 void mdlscale(int *percent)
 {
@@ -190,6 +191,14 @@ void mdlbb(float *rad, float *h, float *eyeheight)
 
 COMMAND(mdlbb, "fff");
 
+void mdlextendbb(float *x, float *y, float *z)
+{
+    checkmdl;
+    loadingmodel->bbextend = vec(*x, *y, *z);
+}
+
+COMMAND(mdlextendbb, "fff");
+
 void mdlname()
 {
     checkmdl;
@@ -198,6 +207,76 @@ void mdlname()
 
 COMMAND(mdlname, "");
 
+#define checkragdoll \
+    skelmodel *m = dynamic_cast<skelmodel *>(loadingmodel); \
+    if(!m) { conoutf(CON_ERROR, "not loading a skeletal model"); return; } \
+    skelmodel::skelmeshgroup *meshes = (skelmodel::skelmeshgroup *)m->parts.last()->meshes; \
+    if(!meshes) return; \
+    skelmodel::skeleton *skel = meshes->skel; \
+    if(!skel->ragdoll) skel->ragdoll = new ragdollskel; \
+    ragdollskel *ragdoll = skel->ragdoll; \
+    if(ragdoll->loaded) return;
+    
+
+void rdvert(float *x, float *y, float *z)
+{
+    checkragdoll;
+    ragdollskel::vert &v = ragdoll->verts.add();
+    v.pos = vec(*x, *y, *z);
+}
+COMMAND(rdvert, "fff");
+
+void rdeye(int *v)
+{
+    checkragdoll;
+    ragdoll->eye = *v;
+}
+COMMAND(rdeye, "i");
+
+void rdtri(int *v1, int *v2, int *v3)
+{
+    checkragdoll;
+    ragdollskel::tri &t = ragdoll->tris.add();
+    t.vert[0] = *v1;
+    t.vert[1] = *v2;
+    t.vert[2] = *v3;
+}
+COMMAND(rdtri, "iii");
+
+void rdjoint(int *n, int *t, char *v1, char *v2, char *v3)
+{
+    checkragdoll;
+    ragdollskel::joint &j = ragdoll->joints.add();
+    j.bone = *n;
+    j.tri = *t;
+    j.vert[0] = v1[0] ? atoi(v1) : -1;
+    j.vert[1] = v2[0] ? atoi(v2) : -1;
+    j.vert[2] = v3[0] ? atoi(v3) : -1;
+}
+COMMAND(rdjoint, "iisss");
+   
+void rdlimitdist(int *v1, int *v2, float *mindist, float *maxdist)
+{
+    checkragdoll;
+    ragdollskel::distlimit &d = ragdoll->distlimits.add();
+    d.vert[0] = *v1;
+    d.vert[1] = *v2;
+    d.mindist = *mindist;
+    d.maxdist = max(*maxdist, *mindist);
+}
+COMMAND(rdlimitdist, "iiff");
+
+void rdlimitrot(int *t1, int *t2, float *maxangle, float *qx, float *qy, float *qz, float *qw)
+{
+    checkragdoll;
+    ragdollskel::rotlimit &r = ragdoll->rotlimits.add();
+    r.tri[0] = *t1;
+    r.tri[1] = *t2;
+    r.maxangle = *maxangle * RAD;
+    r.middle = matrix3x3(quat(*qx, *qy, *qz, *qw));
+}
+COMMAND(rdlimitrot, "iifffff");
+ 
 // mapmodels
 
 vector<mapmodelinfo> mapmodels;
@@ -205,7 +284,7 @@ vector<mapmodelinfo> mapmodels;
 void mmodel(char *name)
 {
     mapmodelinfo &mmi = mapmodels.add();
-    s_strcpy(mmi.name, name);
+    copystring(mmi.name, name);
     mmi.m = NULL;
 }
 
@@ -214,7 +293,11 @@ void mapmodelcompat(int *rad, int *h, int *tex, char *name, char *shadow)
     mmodel(name);
 }
 
-void mapmodelreset() { mapmodels.setsize(0); }
+void mapmodelreset() 
+{ 
+    if(!overrideidents && !game::allowedittoggle()) return;
+    mapmodels.setsize(0); 
+}
 
 mapmodelinfo &getmminfo(int i) { return mapmodels.inrange(i) ? mapmodels[i] : *(mapmodelinfo *)0; }
 const char *mapmodelname(int i) { return mapmodels.inrange(i) ? mapmodels[i].name : NULL; }
@@ -228,6 +311,24 @@ ICOMMAND(nummapmodels, "", (), { intret(mapmodels.length()); });
 // model registry
 
 hashtable<const char *, model *> mdllookup;
+vector<const char *> preloadmodels;
+
+void preloadmodel(const char *name)
+{
+    if(mdllookup.access(name)) return;
+    preloadmodels.add(newstring(name));
+}
+
+void flushpreloadedmodels()
+{
+    loopv(preloadmodels)
+    {
+        loadprogress = float(i+1)/preloadmodels.length();
+        loadmodel(preloadmodels[i], -1, true);
+    }
+    preloadmodels.deletecontentsa();
+    loadprogress = 0;
+}
 
 model *loadmodel(const char *name, int i, bool msg)
 {
@@ -245,8 +346,8 @@ model *loadmodel(const char *name, int i, bool msg)
     { 
         if(msg)
         {
-            s_sprintfd(filename)("packages/models/%s", name);
-            show_out_of_renderloop_progress(0, filename);
+            defformatstring(filename)("packages/models/%s", name);
+            renderprogress(loadprogress, filename);
         }
         m = new md2(name);
         loadingmodel = m;
@@ -275,12 +376,18 @@ model *loadmodel(const char *name, int i, bool msg)
             }
         }
         loadingmodel = NULL;
-        mdllookup.access(m->name(), &m);
+        mdllookup.access(m->name(), m);
     }
     if(mapmodels.inrange(i) && !mapmodels[i].m) mapmodels[i].m = m;
     return m;
 }
 
+void preloadmodelshaders()
+{
+    if(initing) return;
+    enumerate(mdllookup, model *, m, m->preloadshaders());
+}
+
 void clear_mdls()
 {
     enumerate(mdllookup, model *, m, delete m);
@@ -295,6 +402,7 @@ void clearmodel(char *name)
 {
     model **m = mdllookup.access(name);
     if(!m) { conoutf("model %s is not loaded", name); return; }
+    loopv(mapmodels) if(mapmodels[i].m==*m) mapmodels[i].m = NULL;
     mdllookup.remove(name);
     (*m)->cleanup();
     delete *m;
@@ -358,116 +466,12 @@ void renderellipse(vec &o, float xradius, float yradius, float yaw)
     glEnable(GL_TEXTURE_2D);
 }
 
-void setshadowmatrix(const plane &p, const vec &dir)
-{
-    float d = p.dot(dir);
-    GLfloat m[16] =
-    {
-        d-dir.x*p.x,       -dir.y*p.x,       -dir.z*p.x,      0, 
-         -dir.x*p.y,      d-dir.y*p.y,       -dir.z*p.y,      0,
-         -dir.x*p.z,       -dir.y*p.z,      d-dir.z*p.z,      0,
-         -dir.x*p.offset,  -dir.y*p.offset,  -dir.z*p.offset, d
-    };
-    glMultMatrixf(m);
-}
-
-VARP(bounddynshadows, 0, 1, 1);
-VARP(dynshadow, 0, 60, 100);
-
-void rendershadow(vec &dir, model *m, int anim, const vec &o, vec center, float radius, float yaw, float pitch, float speed, int basetime, dynent *d, int cull, modelattach *a)
-{
-    vec floor;
-    float dist = rayfloor(center, floor, 0, center.z);
-    if(dist<=0 || dist>=center.z) return;
-    center.z -= dist;
-    if((cull&MDL_CULL_VFC) && refracting<0 && center.z>=reflectz) return;
-    if(vec(center).sub(camera1->o).dot(floor)>0) return;
-
-    vec shaddir; 
-    if(cull&MDL_DYNSHADOW) 
-    {
-        extern vec shadowdir;
-        shaddir = shadowdir;
-        shaddir.normalize();
-    }
-    else
-    {
-        shaddir = dir;
-        shaddir.z = 0;
-        if(!shaddir.iszero()) shaddir.normalize();
-        shaddir.z = 1.5f*(dir.z*0.5f+1);
-        shaddir.normalize();
-    }
-
-    glDisable(GL_TEXTURE_2D);
-    glDepthMask(GL_FALSE);
-    
-    if(!hasFBO || !reflecting || !refracting || hasDS) 
-    {
-        glEnable(GL_STENCIL_TEST);
-
-        if(bounddynshadows)
-        {    
-            nocolorshader->set();
-            glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
-            glStencilFunc(GL_ALWAYS, 1, 1);
-            glStencilOp(GL_KEEP, GL_REPLACE, GL_ZERO);
-
-            vec below(center);
-            below.z -= 1.0f;
-            glPushMatrix();
-            setshadowmatrix(plane(floor, -floor.dot(below)), shaddir);
-            glBegin(GL_QUADS);
-            loopi(6) if((shaddir[dimension(i)]>0)==dimcoord(i)) loopj(4)
-            {
-                const ivec &cc = cubecoords[fv[i][j]];
-                glVertex3f(center.x + (cc.x ? 1.5f : -1.5f)*radius,
-                           center.y + (cc.y ? 1.5f : -1.5f)*radius,
-                           cc.z ? center.z + dist + radius : below.z);
-                xtraverts += 4;
-            }
-            glEnd();
-            glPopMatrix();
-
-            glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, fading ? GL_FALSE : GL_TRUE);
-        }
-    }
-
-    float intensity = dynshadow/100.0f;
-    if(fogging)
-    {
-        if(renderpath!=R_FIXEDFUNCTION) setfogplane(0, max(0.1f, reflectz-center.z));
-        else intensity *= 1.0f - max(0.0f, min(1.0f, (reflectz - center.z)/waterfog));
-    }
-    if((anim&ANIM_INDEX)==ANIM_DYING) intensity *= max(1.0f - (lastmillis - basetime)/1000.0f, 0.0f);
-    glColor4f(0, 0, 0, intensity);
-
-    if(!hasFBO || !reflecting || !refracting || hasDS)
-    {
-        glStencilFunc(GL_NOTEQUAL, bounddynshadows ? 0 : 1, 1);
-        glStencilOp(GL_KEEP, GL_REPLACE, GL_REPLACE);
-    }
-
-    vec above(center);
-    above.z += 0.25f;
-
-    glPushMatrix();
-    setshadowmatrix(plane(floor, -floor.dot(above)), shaddir);
-    m->render(anim|ANIM_NOSKIN|ANIM_SHADOW, speed, basetime, o, yaw, pitch, d, a);
-    glPopMatrix();
-
-    glEnable(GL_TEXTURE_2D);
-    glDepthMask(GL_TRUE);
-    
-    if(!hasFBO || !reflecting || !refracting || hasDS) glDisable(GL_STENCIL_TEST);
-}
-
 struct batchedmodel
 {
     vec pos, color, dir;
     int anim;
-    float yaw, pitch, speed;
-    int basetime, cull;
+    float yaw, pitch, transparent;
+    int basetime, basetime2, flags;
     dynent *d;
     int attached;
     occludequery *query;
@@ -475,6 +479,7 @@ struct batchedmodel
 struct modelbatch
 {
     model *m;
+    int flags;
     vector<batchedmodel> batched;
 };  
 static vector<modelbatch *> batches;
@@ -488,7 +493,7 @@ void startmodelbatches()
     modelattached.setsizenodelete(0);
 }
 
-batchedmodel &addbatchedmodel(model *m)
+modelbatch &addbatchedmodel(model *m)
 {
     modelbatch *b = NULL;
     if(m->batch>=0 && m->batch<numbatches && batches[m->batch]->m==m) b = batches[m->batch];
@@ -501,50 +506,39 @@ batchedmodel &addbatchedmodel(model *m)
         }
         else b = batches.add(new modelbatch);
         b->m = m;
+        b->flags = 0;
         m->batch = numbatches++;
     }
-    batchedmodel &bm = b->batched.add();
-    bm.query = modelquery;
-    return bm;
+    return *b;
 }
 
 void renderbatchedmodel(model *m, batchedmodel &b)
 {
     modelattach *a = NULL;
     if(b.attached>=0) a = &modelattached[b.attached];
-    if((!shadowmap || renderpath==R_FIXEDFUNCTION) && (b.cull&(MDL_SHADOW|MDL_DYNSHADOW)) && dynshadow && hasstencil && !reflecting && refracting<=0)
-    {
-        vec center;
-        float radius = m->boundsphere(0/*frame*/, center, a); // FIXME
-        center.add(b.pos);
-        rendershadow(b.dir, m, b.anim, b.pos, center, radius, b.yaw, b.pitch, b.speed, b.basetime, b.d, b.cull, a);
-        if((b.cull&MDL_CULL_VFC) && refracting<0 && center.z-radius>=reflectz) return;
-    }
 
     int anim = b.anim;
     if(shadowmapping) 
     {
         anim |= ANIM_NOSKIN; 
-        setenvparamf("shadowintensity", SHPARAM_VERTEX, 1,
-            (anim&ANIM_INDEX)==ANIM_DYING ? max(1.0f - (lastmillis - b.basetime)/1000.0f, 0.0f) : 1.0f);
+        if(renderpath!=R_FIXEDFUNCTION) setenvparamf("shadowintensity", SHPARAM_VERTEX, 1, b.transparent);
     }
     else 
     {
-        if(b.cull&MDL_TRANSLUCENT) anim |= ANIM_TRANSLUCENT;
-        if(b.cull&MDL_FULLBRIGHT) anim |= ANIM_FULLBRIGHT;
+        if(b.flags&MDL_FULLBRIGHT) anim |= ANIM_FULLBRIGHT;
     }
 
-    m->render(anim, b.speed, b.basetime, b.pos, b.yaw, b.pitch, b.d, a, b.color, b.dir);
+    m->render(anim, b.basetime, b.basetime2, b.pos, b.yaw, b.pitch, b.d, a, b.color, b.dir, b.transparent);
 }
 
-struct translucentmodel
+struct transparentmodel
 {
     model *m;
     batchedmodel *batched;
     float dist;
 };
 
-static int sorttranslucentmodels(const translucentmodel *x, const translucentmodel *y)
+static int sorttransparentmodels(const transparentmodel *x, const transparentmodel *y)
 {
     if(x->dist > y->dist) return -1;
     if(x->dist < y->dist) return 1;
@@ -553,28 +547,41 @@ static int sorttranslucentmodels(const translucentmodel *x, const translucentmod
 
 void endmodelbatches()
 {
-    vector<translucentmodel> translucent;
+    vector<transparentmodel> transparent;
     loopi(numbatches)
     {
         modelbatch &b = *batches[i];
         if(b.batched.empty()) continue;
+        if(b.flags&(MDL_SHADOW|MDL_DYNSHADOW))
+        {
+            vec center, bbradius;
+            b.m->boundbox(0/*frame*/, center, bbradius); // FIXME
+            loopvj(b.batched)
+            {
+                batchedmodel &bm = b.batched[j];
+                if(bm.flags&(MDL_SHADOW|MDL_DYNSHADOW))
+                    renderblob(bm.flags&MDL_DYNSHADOW ? BLOB_DYNAMIC : BLOB_STATIC, bm.d && bm.d->ragdoll ? bm.d->ragdoll->center : bm.pos, bm.d ? bm.d->radius : max(bbradius.x, bbradius.y), bm.transparent);
+            }
+            flushblobs();
+        }
         bool rendered = false;
         occludequery *query = NULL;
         loopvj(b.batched) 
         {
             batchedmodel &bm = b.batched[j];
+            if(bm.flags&MDL_CULL_VFC) continue;
             if(bm.query!=query)
             {
                 if(query) endquery(query);
                 query = bm.query;
                 if(query) startquery(query);
             }
-            if(bm.cull&MDL_TRANSLUCENT && (!query || query->owner==bm.d))
+            if(bm.transparent < 1 && (!query || query->owner==bm.d) && !shadowmapping)
             {
-                translucentmodel &tm = translucent.add();
+                transparentmodel &tm = transparent.add();
                 tm.m = b.m;
                 tm.batched = &bm;
-                tm.dist = camera1->o.dist(bm.pos);
+                tm.dist = camera1->o.dist(bm.d && bm.d->ragdoll ? bm.d->ragdoll->center : bm.pos);
                 continue;
             }
             if(!rendered) { b.m->startrender(); rendered = true; }
@@ -583,14 +590,14 @@ void endmodelbatches()
         if(query) endquery(query);
         if(rendered) b.m->endrender();
     }
-    if(translucent.length())
+    if(transparent.length())
     {
-        translucent.sort(sorttranslucentmodels);
+        transparent.sort(sorttransparentmodels);
         model *lastmodel = NULL;
         occludequery *query = NULL;
-        loopv(translucent)
+        loopv(transparent)
         {
-            translucentmodel &tm = translucent[i];
+            transparentmodel &tm = transparent[i];
             if(lastmodel!=tm.m)
             {
                 if(lastmodel) lastmodel->endrender();
@@ -677,22 +684,31 @@ void rendermodelquery(model *m, dynent *d, const vec &center, float radius)
 
 extern int oqfrags;
 
-void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, float yaw, float pitch, int cull, dynent *d, modelattach *a, int basetime, float speed)
+void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, float yaw, float pitch, int flags, dynent *d, modelattach *a, int basetime, int basetime2, float trans)
 {
-    if(shadowmapping && !(cull&(MDL_SHADOW|MDL_DYNSHADOW))) return;
+    if(shadowmapping && !(flags&(MDL_SHADOW|MDL_DYNSHADOW))) return;
     model *m = loadmodel(mdl); 
     if(!m) return;
-    vec center;
+    vec center, bbradius;
     float radius = 0;
-    bool shadow = (!shadowmap || renderpath==R_FIXEDFUNCTION) && (cull&(MDL_SHADOW|MDL_DYNSHADOW)) && dynshadow && hasstencil,
-         doOQ = cull&MDL_CULL_QUERY && hasOQ && oqfrags && oqdynent;
-    if(cull&(MDL_CULL_VFC|MDL_CULL_DIST|MDL_CULL_OCCLUDED|MDL_CULL_QUERY|MDL_SHADOW|MDL_DYNSHADOW))
+    bool shadow = !shadowmap && !glaring && (flags&(MDL_SHADOW|MDL_DYNSHADOW)) && showblobs,
+         doOQ = flags&MDL_CULL_QUERY && hasOQ && oqfrags && oqdynent;
+    if(flags&(MDL_CULL_VFC|MDL_CULL_DIST|MDL_CULL_OCCLUDED|MDL_CULL_QUERY|MDL_SHADOW|MDL_DYNSHADOW))
     {
-        radius = m->boundsphere(0/*frame*/, center, a); // FIXME
-        center.rotate_around_z((-180-yaw)*RAD);
-        center.add(o);
-        if(cull&MDL_CULL_DIST && center.dist(camera1->o)/radius>maxmodelradiusdistance) return;
-        if(cull&MDL_CULL_VFC)
+        m->boundbox(0/*frame*/, center, bbradius); // FIXME
+        radius = bbradius.magnitude();
+        if(d && d->ragdoll)
+        {
+            radius = max(radius, d->ragdoll->radius);
+            center = d->ragdoll->center;
+        }
+        else
+        {
+            center.rotate_around_z((-180-yaw)*RAD);
+            center.add(o);
+        }
+        if(flags&MDL_CULL_DIST && center.dist(camera1->o)/radius>maxmodelradiusdistance) return;
+        if(flags&MDL_CULL_VFC)
         {
             if(reflecting || refracting)
             {
@@ -714,12 +730,12 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
         {
             if(d)
             {
-                if(cull&MDL_CULL_OCCLUDED && d->occluded>=OCCLUDE_PARENT) return;
+                if(flags&MDL_CULL_OCCLUDED && d->occluded>=OCCLUDE_PARENT) return;
                 if(doOQ && d->occluded+1>=OCCLUDE_BB && d->query && d->query->owner==d && checkquery(d->query)) return;
             }
             if(!addshadowmapcaster(center, radius, radius)) return;
         }
-        else if(cull&MDL_CULL_OCCLUDED && modeloccluded(center, radius))
+        else if(flags&MDL_CULL_OCCLUDED && modeloccluded(center, radius))
         {
             if(!reflecting && !refracting && d)
             {
@@ -738,7 +754,9 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
             return;
         }
     }
-    if(showboundingbox && !shadowmapping && !reflecting && !refracting)
+
+    if(flags&MDL_NORENDER) anim |= ANIM_NORENDER;
+    else if(showboundingbox && !shadowmapping && !reflecting && !refracting)
     {
         if(d && showboundingbox==1) 
         {
@@ -759,39 +777,45 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
     vec lightcolor(1, 1, 1), lightdir(0, 0, 1);
     if(!shadowmapping)
     {
+        vec pos = o;
         if(d) 
         {
             if(!reflecting && !refracting) d->occluded = OCCLUDE_NOTHING;
             if(!light) light = &d->light;
-            if(cull&MDL_LIGHT && light->millis!=lastmillis)
+            if(flags&MDL_LIGHT && light->millis!=lastmillis)
             {
-                lightreaching(d->o, light->color, light->dir);
-                dynlightreaching(o, light->color, light->dir);
-                cl->lighteffects(d, light->color, light->dir);
+                if(d->ragdoll)
+                {
+                    pos = d->ragdoll->center;
+                    pos.z += radius/2;
+                }
+                lightreaching(pos, light->color, light->dir);
+                dynlightreaching(pos, light->color, light->dir);
+                game::lighteffects(d, light->color, light->dir);
                 light->millis = lastmillis;
             }
         }
-        else if(cull&MDL_LIGHT)
+        else if(flags&MDL_LIGHT)
         {
             if(!light) 
             {
-                lightreaching(o, lightcolor, lightdir);
-                dynlightreaching(o, lightcolor, lightdir);
+                lightreaching(pos, lightcolor, lightdir);
+                dynlightreaching(pos, lightcolor, lightdir);
             }
             else if(light->millis!=lastmillis)
             {
-                lightreaching(o, light->color, light->dir);
-                dynlightreaching(o, light->color, light->dir);
+                lightreaching(pos, light->color, light->dir);
+                dynlightreaching(pos, light->color, light->dir);
                 light->millis = lastmillis;
             }
         }
         if(light) { lightcolor = light->color; lightdir = light->dir; }
-        if(cull&MDL_DYNLIGHT) dynlightreaching(o, lightcolor, lightdir);
+        if(flags&MDL_DYNLIGHT) dynlightreaching(pos, lightcolor, lightdir);
     }
 
-    if(a) for(int i = 0; a[i].name; i++)
+    if(a) for(int i = 0; a[i].tag; i++)
     {
-        a[i].m = loadmodel(a[i].name);
+        if(a[i].name) a[i].m = loadmodel(a[i].name);
         //if(a[i].m && a[i].m->type()!=m->type()) a[i].m = NULL;
     }
 
@@ -799,41 +823,49 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
   
     if(numbatches>=0)
     {
-        batchedmodel &b = addbatchedmodel(m);
+        modelbatch &mb = addbatchedmodel(m);
+        batchedmodel &b = mb.batched.add();
+        b.query = modelquery;
         b.pos = o;
         b.color = lightcolor;
         b.dir = lightdir;
         b.anim = anim;
         b.yaw = yaw;
         b.pitch = pitch;
-        b.speed = speed;
         b.basetime = basetime;
-        b.cull = cull;
+        b.basetime2 = basetime2;
+        b.transparent = trans;
+        b.flags = flags & ~(MDL_CULL_VFC | MDL_CULL_DIST | MDL_CULL_OCCLUDED);
+        if(!shadow || reflecting || refracting>0) 
+        {
+            b.flags &= ~(MDL_SHADOW|MDL_DYNSHADOW);
+            if((flags&MDL_CULL_VFC) && refracting<0 && center.z-radius>=reflectz) b.flags |= MDL_CULL_VFC;
+        }
+        mb.flags |= b.flags;
         b.d = d;
         b.attached = a ? modelattached.length() : -1;
-        if(a) for(int i = 0;; i++) { modelattached.add(a[i]); if(!a[i].name) break; }
+        if(a) for(int i = 0;; i++) { modelattached.add(a[i]); if(!a[i].tag) break; }
         if(doOQ) d->query = b.query = newquery(d);
         return;
     }
 
-    m->startrender();
-
     if(shadow && !reflecting && refracting<=0)
     {
-        rendershadow(lightdir, m, anim, o, center, radius, yaw, pitch, speed, basetime, d, cull, a);
-        if((cull&MDL_CULL_VFC) && refracting<0 && center.z-radius>=reflectz) { m->endrender(); return; }
+        renderblob(flags&MDL_DYNSHADOW ? BLOB_DYNAMIC : BLOB_STATIC, d && d->ragdoll ? center : o, d ? d->radius : max(bbradius.x, bbradius.y), trans);
+        flushblobs();
+        if((flags&MDL_CULL_VFC) && refracting<0 && center.z-radius>=reflectz) return;
     }
 
+    m->startrender();
+
     if(shadowmapping)
     {
         anim |= ANIM_NOSKIN;
-        setenvparamf("shadowintensity", SHPARAM_VERTEX, 1,
-            (anim&ANIM_INDEX)==ANIM_DYING ? max(1.0f - (lastmillis - basetime)/1000.0f, 0.0f) : 1.0f);
+        if(renderpath!=R_FIXEDFUNCTION) setenvparamf("shadowintensity", SHPARAM_VERTEX, 1, trans);
     }
-    else
+    else 
     {
-        if(cull&MDL_TRANSLUCENT) anim |= ANIM_TRANSLUCENT;
-        if(cull&MDL_FULLBRIGHT) anim |= ANIM_FULLBRIGHT;
+        if(flags&MDL_FULLBRIGHT) anim |= ANIM_FULLBRIGHT;
     }
 
     if(doOQ)
@@ -842,7 +874,7 @@ void rendermodel(entitylight *light, const char *mdl, int anim, const vec &o, fl
         if(d->query) startquery(d->query);
     }
 
-    m->render(anim, speed, basetime, o, yaw, pitch, d, a, lightcolor, lightdir);
+    m->render(anim, basetime, basetime2, o, yaw, pitch, d, a, lightcolor, lightdir, trans);
 
     if(doOQ && d->query) endquery(d->query);
 
@@ -887,12 +919,14 @@ void findanims(const char *pattern, vector<int> &anims)
     { 
         "dead", "dying", "idle", 
         "forward", "backward", "left", "right", 
-        "punch", "shoot", "pain", 
+        "hold 1", "hold 2", "hold 3", "hold 4", "hold 5", "hold 6", "hold 7",
+        "attack 1", "attack 2", "attack 3", "attack 4", "attack 5", "attack 6", "attack 7",
+        "pain", 
         "jump", "sink", "swim", 
         "edit", "lag", "taunt", "win", "lose", 
-        "gun shoot", "gun idle",
-        "vwep", "shield", "powerup", 
-        "mapmodel", "trigger" 
+        "gun idle", "gun shoot",
+        "vwep idle", "vwep shoot", "shield", "powerup", 
+        "mapmodel", "trigger"
     };
     loopi(sizeof(names)/sizeof(names[0])) if(matchanim(names[i], pattern)) anims.add(i);
 }
@@ -900,20 +934,20 @@ void findanims(const char *pattern, vector<int> &anims)
 void loadskin(const char *dir, const char *altdir, Texture *&skin, Texture *&masks) // model skin sharing
 {
 #define ifnoload(tex, path) if((tex = textureload(path, 0, true, false))==notexture)
-#define tryload(tex, prefix, name) \
-    ifnoload(tex, makerelpath(mdir, name ".jpg", prefix)) \
+#define tryload(tex, cmd, name) \
+    ifnoload(tex, makerelpath(mdir, name ".jpg", NULL, cmd)) \
     { \
-        ifnoload(tex, makerelpath(mdir, name ".png", prefix)) \
+        ifnoload(tex, makerelpath(mdir, name ".png", NULL, cmd)) \
         { \
-            ifnoload(tex, makerelpath(maltdir, name ".jpg", prefix)) \
+            ifnoload(tex, makerelpath(maltdir, name ".jpg", NULL, cmd)) \
             { \
-                ifnoload(tex, makerelpath(maltdir, name ".png", prefix)) return; \
+                ifnoload(tex, makerelpath(maltdir, name ".png", NULL, cmd)) return; \
             } \
         } \
     }
    
-    s_sprintfd(mdir)("packages/models/%s", dir);
-    s_sprintfd(maltdir)("packages/models/%s", altdir);
+    defformatstring(mdir)("packages/models/%s", dir);
+    defformatstring(maltdir)("packages/models/%s", altdir);
     masks = notexture;
     tryload(skin, NULL, "skin");
     tryload(masks, "<ffmask:25>", "masks");
@@ -925,58 +959,63 @@ VAR(animoverride, -1, 0, NUMANIMS-1);
 VAR(testanims, 0, 0, 1);
 VAR(testpitch, -90, 0, 90);
 
-void renderclient(dynent *d, const char *mdlname, modelattach *attachments, int attack, int attackdelay, int lastaction, int lastpain, float sink)
+void renderclient(dynent *d, const char *mdlname, modelattach *attachments, int hold, int attack, int attackdelay, int lastaction, int lastpain, float fade, bool ragdoll)
 {
-    int anim = ANIM_IDLE|ANIM_LOOP;
-    float yaw = d->yaw, pitch = d->pitch;
-    vec o(d->o);
-    o.z -= d->eyeheight + sink;
+    int anim = hold ? hold : ANIM_IDLE|ANIM_LOOP;
+    float yaw = testanims && d==player ? 0 : d->yaw+90,
+          pitch = testpitch && d==player ? testpitch : d->pitch;
+    vec o = d->feetpos();
     int basetime = 0;
     if(animoverride) anim = (animoverride<0 ? ANIM_ALL : animoverride)|ANIM_LOOP;
     else if(d->state==CS_DEAD)
     {
-        pitch = 0;
         anim = ANIM_DYING;
         basetime = lastpain;
-        int t = lastmillis-lastpain;
-        if(t<0 || t>20000) return;
-        if(t>1000) { anim = ANIM_DEAD|ANIM_LOOP; if(t>1600) { t -= 1600; o.z -= t*t/10000000000.0f*t/16.0f; } }
-        if(o.z<-1000) return;
+        if(ragdoll)
+        {
+            if(!d->ragdoll || d->ragdoll->millis < basetime) anim |= ANIM_RAGDOLL;
+        }
+        else 
+        {
+            pitch *= max(1.0f - (lastmillis-basetime)/500.0f, 0.0f);
+            if(lastmillis-basetime>1000) anim = ANIM_DEAD|ANIM_LOOP;
+        }
     }
     else if(d->state==CS_EDITING || d->state==CS_SPECTATOR) anim = ANIM_EDIT|ANIM_LOOP;
     else if(d->state==CS_LAGGED)                            anim = ANIM_LAG|ANIM_LOOP;
     else
     {
-        if(lastmillis-lastpain<300) 
+        if(lastmillis-lastpain < 300) 
         { 
             anim = ANIM_PAIN;
             basetime = lastpain;
         }
-        else if(lastpain < lastaction && (attack<0 || (d->type!=ENT_AI && lastmillis-lastaction<attackdelay)))
+        else if(lastpain < lastaction && (attack < 0 || (d->type != ENT_AI && lastmillis-lastaction < attackdelay)))
         { 
-            anim = attack<0 ? -attack : attack; 
+            anim = attack < 0 ? -attack : attack; 
             basetime = lastaction; 
         }
 
-        if(d->inwater && d->physstate<=PHYS_FALL) anim |= (((cl->allowmove(d) && (d->move || d->strafe)) || d->vel.z+d->falling.z>0 ? ANIM_SWIM : ANIM_SINK)|ANIM_LOOP)<<ANIM_SECONDARY;
+        if(d->inwater && d->physstate<=PHYS_FALL) anim |= (((game::allowmove(d) && (d->move || d->strafe)) || d->vel.z+d->falling.z>0 ? ANIM_SWIM : ANIM_SINK)|ANIM_LOOP)<<ANIM_SECONDARY;
         else if(d->timeinair>100) anim |= (ANIM_JUMP|ANIM_END)<<ANIM_SECONDARY;
-        else if(cl->allowmove(d)) 
+        else if(game::allowmove(d) && (d->move || d->strafe)) 
         {
             if(d->move>0) anim |= (ANIM_FORWARD|ANIM_LOOP)<<ANIM_SECONDARY;
             else if(d->strafe) anim |= ((d->strafe>0 ? ANIM_LEFT : ANIM_RIGHT)|ANIM_LOOP)<<ANIM_SECONDARY;
             else if(d->move<0) anim |= (ANIM_BACKWARD|ANIM_LOOP)<<ANIM_SECONDARY;
         }
-
+        
         if((anim&ANIM_INDEX)==ANIM_IDLE && (anim>>ANIM_SECONDARY)&ANIM_INDEX) anim >>= ANIM_SECONDARY;
     }
+    if(d->ragdoll && (!ragdoll || anim!=ANIM_DYING)) DELETEP(d->ragdoll);
     if(!((anim>>ANIM_SECONDARY)&ANIM_INDEX)) anim |= (ANIM_IDLE|ANIM_LOOP)<<ANIM_SECONDARY;
     int flags = MDL_LIGHT;
-    if(d!=player) flags |= MDL_CULL_VFC | MDL_CULL_OCCLUDED | MDL_CULL_QUERY;
+    if(d!=player && !(anim&ANIM_RAGDOLL)) flags |= MDL_CULL_VFC | MDL_CULL_OCCLUDED | MDL_CULL_QUERY;
     if(d->type==ENT_PLAYER) flags |= MDL_FULLBRIGHT;
     else flags |= MDL_CULL_DIST;
-    if(d->state==CS_LAGGED) flags |= MDL_TRANSLUCENT;
-    else if((anim&ANIM_INDEX)!=ANIM_DEAD) flags |= MDL_DYNSHADOW;
-    rendermodel(NULL, mdlname, anim, o, testanims && d==player ? 0 : yaw+90, testpitch && d==player ? testpitch : pitch, flags, d, attachments, basetime);
+    if(d->state==CS_LAGGED) fade = min(fade, 0.3f);
+    else flags |= MDL_DYNSHADOW;
+    rendermodel(NULL, mdlname, anim, o, yaw, pitch, flags, d, attachments, basetime, 0, fade);
 }
 
 void setbbfrommodel(dynent *d, const char *mdl)
