@@ -1,6 +1,5 @@
 // renderva.cpp: handles the occlusion and rendering of vertex arrays
 
-#include "pch.h"
 #include "engine.h"
 
 static inline void drawtris(GLsizei numindices, const GLvoid *indices, ushort minvert, ushort maxvert)
@@ -76,7 +75,7 @@ void addvisibleva(vtxarray *va)
     float dist = vadist(va, camera1->o);
     va->distance = int(dist); /*cv.dist(camera1->o) - va->size*SQRT3/2*/
 
-    int hash = min(int(dist*VASORTSIZE/hdr.worldsize), VASORTSIZE-1);
+    int hash = min(int(dist*VASORTSIZE/worldsize), VASORTSIZE-1);
     vtxarray **prev = &vasort[hash], *cur = vasort[hash];
 
     while(cur && va->distance >= cur->distance)
@@ -282,6 +281,7 @@ void clearqueries()
 
 VAR(oqfrags, 0, 8, 64);
 VAR(oqreflect, 0, 4, 64);
+VAR(oqwait, 0, 1, 1);
 
 bool checkquery(occludequery *query, bool nowait)
 {
@@ -289,7 +289,7 @@ bool checkquery(occludequery *query, bool nowait)
     if(query->fragments >= 0) fragments = query->fragments;
     else
     {
-        if(nowait)
+        if(nowait || !oqwait)
         {
             GLint avail;
             glGetQueryObjectiv_(query->id, GL_QUERY_RESULT_AVAILABLE, &avail);
@@ -358,7 +358,7 @@ void findvisiblemms(const vector<extentity *> &ents)
                 loopv(oe->mapmodels)
                 {
                     extentity &e = *ents[oe->mapmodels[i]];
-                    if(e.visible || (e.attr3 && e.triggerstate == TRIGGER_DISAPPEARED)) continue;
+                    if(e.attr3 && e.triggerstate == TRIGGER_DISAPPEARED) continue;
                     e.visible = true;
                     ++visible;
                 }
@@ -401,27 +401,30 @@ void rendermapmodel(extentity &e)
 
 extern int reflectdist;
 
-static vector<octaentities *> renderedmms;
-
 vtxarray *reflectedva;
 
 void renderreflectedmapmodels()
 {
-    vector<octaentities *> reflectedmms;
-    vector<octaentities *> &mms = reflecting ? reflectedmms : renderedmms;
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
 
+    octaentities *mms = visiblemms;
     if(reflecting)
     {
+        octaentities **lastmms = &mms;
         for(vtxarray *va = reflectedva; va; va = va->rnext)
         {
             if(!va->mapmodels || va->distance > reflectdist) continue;
-            loopv(*va->mapmodels) reflectedmms.add((*va->mapmodels)[i]);
+            loopv(*va->mapmodels) 
+            {
+                octaentities *oe = (*va->mapmodels)[i];
+                *lastmms = oe;
+                lastmms = &oe->rnext;
+            }
         }
+        *lastmms = NULL;
     }
-    loopv(mms)
+    for(octaentities *oe = mms; oe; oe = reflecting ? oe->rnext : oe->next)
     {
-        octaentities *oe = mms[i];
         if(reflecting || refracting>0 ? oe->bbmax.z <= reflectz : oe->bbmin.z >= reflectz) continue;
         if(isvisiblecube(oe->o, oe->size) >= VFC_FOGGED) continue;
         loopv(oe->mapmodels)
@@ -431,12 +434,11 @@ void renderreflectedmapmodels()
            e.visible = true;
         }
     }
-    if(mms.length())
+    if(mms)
     {
         startmodelbatches();
-        loopv(mms)
+        for(octaentities *oe = mms; oe; oe = reflecting ? oe->rnext : oe->next)
         {
-            octaentities *oe = mms[i];
             loopv(oe->mapmodels)
             {
                 extentity &e = *ents[oe->mapmodels[i]];
@@ -451,7 +453,7 @@ void renderreflectedmapmodels()
 
 void rendermapmodels()
 {
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
 
     visiblemms = NULL;
     lastvisiblemms = &visiblemms;
@@ -460,24 +462,24 @@ void rendermapmodels()
     static int skipoq = 0;
     bool doquery = hasOQ && oqfrags && oqmm;
 
-    renderedmms.setsizenodelete(0);
     startmodelbatches();
     for(octaentities *oe = visiblemms; oe; oe = oe->next) if(oe->distance>=0)
     {
+        bool rendered = false;
         loopv(oe->mapmodels)
         {
             extentity &e = *ents[oe->mapmodels[i]];
-            if(!e.visible || (e.attr3 && e.triggerstate == TRIGGER_DISAPPEARED)) continue;
-            if(renderedmms.empty() || renderedmms.last()!=oe)
+            if(!e.visible) continue;
+            if(!rendered)
             {
-                renderedmms.add(oe);
+                rendered = true;
                 oe->query = doquery && oe->distance>0 && !(++skipoq%oqmm) ? newquery(oe) : NULL;
                 if(oe->query) startmodelquery(oe->query);
             }        
             rendermapmodel(e);
             e.visible = false;
         }
-        if(renderedmms.length() && renderedmms.last()==oe && oe->query) endmodelquery();
+        if(rendered && oe->query) endmodelquery();
     }
     endmodelbatches();
 
@@ -536,7 +538,7 @@ bool bboccluded(const ivec &bo, const ivec &br)
     if(c->ext && c->ext->va)
     {
         vtxarray *va = c->ext->va;
-        if(va->curvfc >= VFC_FOGGED || va->occluded >= OCCLUDE_BB) return true;
+        if(va->curvfc >= VFC_FOGGED || (va->occluded >= OCCLUDE_BB && bbinsideva(bo, br, va))) return true;
     }
     scale--;
     while(c->children && !(diff&(1<<scale)))
@@ -545,7 +547,7 @@ bool bboccluded(const ivec &bo, const ivec &br)
         if(c->ext && c->ext->va)
         {
             vtxarray *va = c->ext->va;
-            if(va->curvfc >= VFC_FOGGED || va->occluded >= OCCLUDE_BB) return true;
+            if(va->curvfc >= VFC_FOGGED || (va->occluded >= OCCLUDE_BB && bbinsideva(bo, br, va))) return true;
         }
         scale--;
     }
@@ -553,7 +555,31 @@ bool bboccluded(const ivec &bo, const ivec &br)
     return false;
 }
 
-VAR(outline, 0, 0, 0xFFFFFF);
+extern int ati_texgen_bug;
+
+static void setuptexgen(int dims = 2)
+{
+    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
+    glEnable(GL_TEXTURE_GEN_S);
+    if(dims>=2)
+    {
+        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
+        glEnable(GL_TEXTURE_GEN_T);
+        if(ati_texgen_bug) glEnable(GL_TEXTURE_GEN_R);     // should not be needed, but apparently makes some ATI drivers happy
+    }
+}
+
+static void disabletexgen(int dims = 2)
+{
+    glDisable(GL_TEXTURE_GEN_S);
+    if(dims>=2)
+    {
+        glDisable(GL_TEXTURE_GEN_T);
+        if(ati_texgen_bug) glDisable(GL_TEXTURE_GEN_R);
+    }
+}
+
+HVAR(outline, 0, 0, 0xFFFFFF);
 VAR(dtoutline, 0, 1, 1);
 
 void renderoutline()
@@ -614,6 +640,90 @@ void renderoutline()
     defaultshader->set();
 }
 
+HVAR(blendbrushcolor, 0, 0x0000C0, 0xFFFFFF);
+
+void renderblendbrush(GLuint tex, float x, float y, float w, float h)
+{
+    static Shader *blendbrushshader = NULL;
+    if(!blendbrushshader) blendbrushshader = lookupshaderbyname("blendbrush");
+    blendbrushshader->set();
+
+    glEnableClientState(GL_VERTEX_ARRAY);
+
+    glPushMatrix();
+
+    glDepthFunc(GL_LEQUAL);
+
+    glEnable(GL_BLEND);
+    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
+
+    glEnable(GL_TEXTURE_2D); 
+    glBindTexture(GL_TEXTURE_2D, tex);
+    glColor4ub((blendbrushcolor>>16)&0xFF, (blendbrushcolor>>8)&0xFF, blendbrushcolor&0xFF, 0x40);
+
+    GLfloat s[4] = { 0, 0, 0, 0 }, t[4] = { 0, 0, 0, 0 };
+    if(renderpath==R_FIXEDFUNCTION) setuptexgen();
+
+    resetorigin();
+    vtxarray *prev = NULL;
+    for(vtxarray *va = visibleva; va; va = va->next)
+    {
+        if(!va->texs || va->occluded >= OCCLUDE_GEOM) continue;
+        if(va->o.x + va->size <= x || va->o.y + va->size <= y || va->o.x >= x + w || va->o.y >= y + h) continue;
+
+        if(!prev || va->vbuf != prev->vbuf)
+        {
+            if(setorigin(va))
+            {
+                s[0] = 1.0f / (w*(1<<VVEC_FRAC));
+                s[3] = (vaorigin.x - x) / w;
+                t[1] = 1.0f / (h*(1<<VVEC_FRAC));
+                t[3] = (vaorigin.y - y) / h;
+                if(renderpath==R_FIXEDFUNCTION)
+                {
+                    glTexGenfv(GL_S, GL_OBJECT_PLANE, s);
+                    glTexGenfv(GL_T, GL_OBJECT_PLANE, t);
+                }
+                else
+                {
+                    setlocalparamfv("texgenS", SHPARAM_VERTEX, 0, s);
+                    setlocalparamfv("texgenT", SHPARAM_VERTEX, 1, t);
+                }
+            }
+        
+            if(hasVBO)
+            {
+                glBindBuffer_(GL_ARRAY_BUFFER_ARB, va->vbuf);
+                glBindBuffer_(GL_ELEMENT_ARRAY_BUFFER_ARB, va->ebuf);
+            }
+            glVertexPointer(3, floatvtx ? GL_FLOAT : GL_SHORT, VTXSIZE, &va->vdata[0].x);
+        }
+
+        drawvatris(va, 3*va->tris, va->edata);
+        xtravertsva += va->verts;
+
+        prev = va;
+    }
+
+    if(renderpath==R_FIXEDFUNCTION) disabletexgen();
+
+    glDisable(GL_TEXTURE_2D);
+    glDisable(GL_BLEND);
+
+    glDepthFunc(GL_LESS);
+
+    glPopMatrix();
+
+    if(hasVBO)
+    {
+        glBindBuffer_(GL_ARRAY_BUFFER_ARB, 0);
+        glBindBuffer_(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
+    }
+    glDisableClientState(GL_VERTEX_ARRAY);
+
+    notextureshader->set();
+}
+ 
 void rendershadowmapreceivers()
 {
     if(!hasBE) return;
@@ -774,7 +884,7 @@ float orientation_binormal[6][3][4] =
 
 struct renderstate
 {
-    bool colormask, depthmask, mtglow, skippedglow;
+    bool colormask, depthmask, blending, mtglow, skippedglow;
     GLuint vbuf;
     float fogplane;
     int diffusetmu, lightmaptmu, glowtmu, fogtmu, causticstmu;
@@ -787,8 +897,10 @@ struct renderstate
     bool mttexgen;
     int visibledynlights;
     uint dynlightmask;
+    vec dynlightpos;
+    float dynlightradius;
 
-    renderstate() : colormask(true), depthmask(true), mtglow(false), skippedglow(false), vbuf(0), fogplane(-1), diffusetmu(0), lightmaptmu(1), glowtmu(-1), fogtmu(-1), causticstmu(-1), glowcolor(1, 1, 1), slot(NULL), texgendim(-1), mttexgen(false), visibledynlights(0), dynlightmask(0)
+    renderstate() : colormask(true), depthmask(true), blending(false), mtglow(false), skippedglow(false), vbuf(0), fogplane(-1), diffusetmu(0), lightmaptmu(1), glowtmu(-1), fogtmu(-1), causticstmu(-1), glowcolor(1, 1, 1), slot(NULL), texgendim(-1), mttexgen(false), visibledynlights(0), dynlightmask(0)
     {
         loopk(4) color[k] = 1;
         loopk(8) textures[k] = 0;
@@ -818,7 +930,10 @@ enum
     RENDERPASS_Z,
     RENDERPASS_GLOW,
     RENDERPASS_CAUSTICS,
-    RENDERPASS_FOG
+    RENDERPASS_FOG,
+    RENDERPASS_SHADOWMAP,
+    RENDERPASS_DYNLIGHT,
+    RENDERPASS_LIGHTMAP_BLEND
 };
 
 struct geombatch
@@ -838,10 +953,10 @@ struct geombatch
     {
         if(va->vbuf < b.va->vbuf) return -1;
         if(va->vbuf > b.va->vbuf) return 1;
-        if(va->dynlightmask < b.va->dynlightmask) return -1;
-        if(va->dynlightmask > b.va->dynlightmask) return 1;
         if(renderpath!=R_FIXEDFUNCTION)
         {
+            if(va->dynlightmask < b.va->dynlightmask) return -1;
+            if(va->dynlightmask > b.va->dynlightmask) return 1;
             if(slot.shader < b.slot.shader) return -1;
             if(slot.shader > b.slot.shader) return 1;
             if(slot.params.length() < b.slot.params.length()) return -1;
@@ -956,7 +1071,7 @@ static void changefogplane(renderstate &cur, int pass, vtxarray *va)
 {
     if(renderpath!=R_FIXEDFUNCTION)
     {
-        if(fading || fogging)
+        if((fading && !cur.blending) || fogging)
         {
             float fogplane = reflectz - vaorigin.z;
             if(cur.fogplane!=fogplane)
@@ -967,20 +1082,33 @@ static void changefogplane(renderstate &cur, int pass, vtxarray *va)
             }
         }
     }
-    else if(pass==RENDERPASS_FOG || (cur.fogtmu>=0 && (pass==RENDERPASS_LIGHTMAP || pass==RENDERPASS_GLOW)))
+    else if(pass==RENDERPASS_FOG || (cur.fogtmu>=0 && (pass==RENDERPASS_LIGHTMAP || pass==RENDERPASS_GLOW || pass==RENDERPASS_SHADOWMAP)))
     {
         if(pass==RENDERPASS_LIGHTMAP) glActiveTexture_(GL_TEXTURE0_ARB+cur.fogtmu);
-        else if(pass==RENDERPASS_GLOW) glActiveTexture_(GL_TEXTURE1_ARB);
+        else if(pass==RENDERPASS_GLOW || pass==RENDERPASS_SHADOWMAP) glActiveTexture_(GL_TEXTURE1_ARB);
         GLfloat s[4] = { 0, 0, -1.0f/(waterfog<<VVEC_FRAC), (reflectz - vaorigin.z)/waterfog };
         glTexGenfv(GL_S, GL_OBJECT_PLANE, s);
         if(pass==RENDERPASS_LIGHTMAP) glActiveTexture_(GL_TEXTURE0_ARB+cur.diffusetmu);
-        else if(pass==RENDERPASS_GLOW) glActiveTexture_(GL_TEXTURE0_ARB);
+        else if(pass==RENDERPASS_GLOW || pass==RENDERPASS_SHADOWMAP) glActiveTexture_(GL_TEXTURE0_ARB);
     }
 }
 
+static void changedynlightpos(renderstate &cur, vtxarray *va)
+{
+    GLfloat tx[4] = { 0.5f/(cur.dynlightradius * (1<<VVEC_FRAC)), 0, 0, 0.5f + 0.5f*(vaorigin.x - cur.dynlightpos.x)/cur.dynlightradius },
+            ty[4] = { 0, 0.5f/(cur.dynlightradius * (1<<VVEC_FRAC)), 0, 0.5f + 0.5f*(vaorigin.y - cur.dynlightpos.y)/cur.dynlightradius },
+            tz[4] = { 0, 0, 0.5f/(cur.dynlightradius * (1<<VVEC_FRAC)), 0.5f + 0.5f*(vaorigin.z - cur.dynlightpos.z)/cur.dynlightradius };
+    glActiveTexture_(GL_TEXTURE0_ARB);
+    glTexGenfv(GL_S, GL_OBJECT_PLANE, tx);
+    glTexGenfv(GL_T, GL_OBJECT_PLANE, ty);
+    glActiveTexture_(GL_TEXTURE1_ARB);
+    glTexGenfv(GL_S, GL_OBJECT_PLANE, tz);
+    glActiveTexture_(GL_TEXTURE2_ARB);
+}
+    
 static void changevbuf(renderstate &cur, int pass, vtxarray *va)
 {
-    if(setorigin(va, pass==RENDERPASS_LIGHTMAP && !envmapping))
+    if(setorigin(va, renderpath!=R_FIXEDFUNCTION ? pass==RENDERPASS_LIGHTMAP && !envmapping && !glaring : pass==RENDERPASS_SHADOWMAP))
     {
         cur.visibledynlights = 0;
         cur.dynlightmask = 0;
@@ -1016,7 +1144,8 @@ static void changebatchtmus(renderstate &cur, int pass, geombatch &b)
 {
     bool changed = false;
     extern bool brightengeom;
-    int lmid = brightengeom ? LMID_BRIGHT : b.es.lmid; 
+    extern int fullbright;
+    int lmid = brightengeom && (b.es.lmid < LMID_RESERVED || (fullbright && editmode)) ? LMID_BRIGHT : b.es.lmid; 
     if(cur.textures[cur.lightmaptmu]!=lightmaptexs[lmid].id)
     {
         glActiveTexture_(GL_TEXTURE0_ARB+cur.lightmaptmu);
@@ -1071,15 +1200,15 @@ static void changeglow(renderstate &cur, int pass, Slot &slot)
             if(color==vec(1, 1, 1)) 
             {
                 glActiveTexture_(GL_TEXTURE0_ARB+cur.glowtmu);
-                setuptmu(cur.glowtmu, "P + T");
+                setuptmu(cur.glowtmu, "P + T", "= Pa");
             }
             else if(hasTE3 || hasTE4)
             {
                 glActiveTexture_(GL_TEXTURE0_ARB+cur.glowtmu);
                 if(cur.glowcolor==vec(1, 1, 1))
                 {
-                    if(hasTE3) setuptmu(cur.glowtmu, "TPK3");
-                    else if(hasTE4) setuptmu(cur.glowtmu, "TKP14");
+                    if(hasTE3) setuptmu(cur.glowtmu, "TPK3", "= Pa");
+                    else if(hasTE4) setuptmu(cur.glowtmu, "TKP14", "= Pa");
                 }
                 colortmu(cur.glowtmu, color.x, color.y, color.z);
             }
@@ -1109,7 +1238,7 @@ static void changeglow(renderstate &cur, int pass, Slot &slot)
 
 static void changeslottmus(renderstate &cur, int pass, Slot &slot)
 {
-    if(pass==RENDERPASS_LIGHTMAP || pass==RENDERPASS_COLOR) 
+    if(pass==RENDERPASS_LIGHTMAP || pass==RENDERPASS_COLOR || pass==RENDERPASS_DYNLIGHT) 
     {
         GLuint diffusetex = slot.sts.empty() ? notexture->id : slot.sts[0].t->id;
         if(cur.textures[cur.diffusetmu]!=diffusetex)
@@ -1197,23 +1326,20 @@ static void changeshader(renderstate &cur, Shader *s, Slot &slot, bool shadowed)
 {
     if(glaring)
     {
-        Shader *g = s->hasvariant(min(s->variants[4].length()-1, cur.visibledynlights), 4);
-        if(g) g->set(&slot);
-        else
-        {
-            static Shader *noglareshader = NULL;
-            if(!noglareshader) noglareshader = lookupshaderbyname("noglareworld");
-            noglareshader->set(&slot);
-        }
+        static Shader *noglareshader = NULL, *noglareblendshader = NULL;
+        if(!noglareshader) noglareshader = lookupshaderbyname("noglareworld");
+        if(!noglareblendshader) noglareblendshader = lookupshaderbyname("noglareblendworld");
+        if(s->hasoption(4)) s->setvariant(cur.visibledynlights, 4, &slot, cur.blending ? noglareblendshader : noglareshader);
+        else s->setvariant(cur.blending ? 1 : 0, 4, &slot, cur.blending ? noglareblendshader : noglareshader);
     }
-    else if(fading)
+    else if(fading && !cur.blending)
     {
-        if(shadowed) s->variant(min(s->variants[3].length()-1, cur.visibledynlights), 3)->set(&slot);
-        else s->variant(min(s->variants[2].length()-1, cur.visibledynlights), 2)->set(&slot);
+        if(shadowed) s->setvariant(cur.visibledynlights, 3, &slot);
+        else s->setvariant(cur.visibledynlights, 2, &slot);
     }
-    else if(shadowed) s->variant(min(s->variants[1].length()-1, cur.visibledynlights), 1)->set(&slot);
+    else if(shadowed) s->setvariant(cur.visibledynlights, 1, &slot);
     else if(!cur.visibledynlights) s->set(&slot);
-    else s->variant(min(s->variants[0].length()-1, cur.visibledynlights-1))->set(&slot);
+    else s->setvariant(cur.visibledynlights-1, 0, &slot);
     if(s->type&SHADER_GLSLANG) cur.texgendim = -1;
 }
 
@@ -1264,8 +1390,8 @@ static void changetexgen(renderstate &cur, Slot &slot, int dim)
             glTexGenfv(GL_S, GL_OBJECT_PLANE, sgen);
             glTexGenfv(GL_T, GL_OBJECT_PLANE, tgen);
             glActiveTexture_(GL_TEXTURE0_ARB+cur.diffusetmu);
-            cur.mttexgen = cur.mtglow;
         }
+        cur.mttexgen = cur.mtglow;
     }
     else
     {
@@ -1367,7 +1493,8 @@ static void renderbatches(renderstate &cur, int pass)
         if(cur.vbuf != b.va->vbuf) 
         {
             changevbuf(cur, pass, b.va);
-            changefogplane(cur, pass, b.va);
+            if(pass == RENDERPASS_DYNLIGHT) changedynlightpos(cur, b.va);
+            else changefogplane(cur, pass, b.va);
         }
         if(pass == RENDERPASS_LIGHTMAP) 
         {
@@ -1470,9 +1597,7 @@ void renderfoggedvas(renderstate &cur, bool doquery = false)
 
     glDisable(GL_TEXTURE_2D);
         
-    uchar wcol[3];
-    getwatercolour(wcol);
-    glColor3ubv(wcol);
+    glColor3ubv(watercolor.v);
 
     loopv(foggedvas)
     {
@@ -1490,6 +1615,41 @@ void renderfoggedvas(renderstate &cur, bool doquery = false)
     foggedvas.setsizenodelete(0);
 }
 
+void rendershadowmappass(renderstate &cur, vtxarray *va)
+{
+    if(cur.vbuf!=va->vbuf) 
+    {
+        changevbuf(cur, RENDERPASS_SHADOWMAP, va);
+        changefogplane(cur, RENDERPASS_SHADOWMAP, va);
+    }
+
+    elementset *texs = va->eslist;
+    ushort *edata = va->edata;
+    loopi(va->texs)
+    {
+        elementset &es = texs[i];
+        int len = es.length[1] - es.length[0];
+        if(len > 0) 
+        {
+            drawtris(len, &edata[es.length[0]], es.minvert[1], es.maxvert[1]);
+            vtris += len/3;
+        }
+        len = es.length[3] - es.length[2];
+        if(len > 0) 
+        {
+            drawtris(len, &edata[es.length[2]], es.minvert[3], es.maxvert[3]);
+            vtris += len/3;
+        }
+        len = es.length[5] - es.length[4];
+        if(len > 0) 
+        {
+            drawtris(len, &edata[es.length[4]], es.minvert[5], es.maxvert[5]);
+            vtris += len/3;
+        }
+        edata += es.length[5];
+    }
+}
+
 VAR(batchgeom, 0, 1, 1);
 
 void renderva(renderstate &cur, vtxarray *va, int pass = RENDERPASS_LIGHTMAP, bool fogpass = false, bool doquery = false)
@@ -1512,7 +1672,7 @@ void renderva(renderstate &cur, vtxarray *va, int pass = RENDERPASS_LIGHTMAP, bo
                 foggedvas.add(va);
                 break;
             }
-            if(renderpath!=R_FIXEDFUNCTION && !envmapping)
+            if(renderpath!=R_FIXEDFUNCTION && !envmapping && !glaring)
             {
                 va->shadowed = isshadowmapreceiver(va);
                 calcdynlightmask(va);
@@ -1523,6 +1683,24 @@ void renderva(renderstate &cur, vtxarray *va, int pass = RENDERPASS_LIGHTMAP, bo
             else if(!batchgeom && geombatches.length()) renderbatches(cur, pass);
             break;
 
+        case RENDERPASS_LIGHTMAP_BLEND:
+        {
+            if(doquery) startvaquery(va, { if(geombatches.length()) renderbatches(cur, RENDERPASS_LIGHTMAP); });
+            ushort *edata = va->edata;
+            loopi(va->texs) edata += va->eslist[i].length[5];
+            mergetexs(va, &va->eslist[va->texs], va->blends, edata);
+            if(doquery) endvaquery(va, { if(geombatches.length()) renderbatches(cur, RENDERPASS_LIGHTMAP); });
+            else if(!batchgeom && geombatches.length()) renderbatches(cur, RENDERPASS_LIGHTMAP);
+            break;
+        }
+
+        case RENDERPASS_DYNLIGHT:
+            if(cur.dynlightpos.dist_to_bb(va->geommin, va->geommax) >= cur.dynlightradius) break;
+            vverts += va->verts;
+            mergetexs(va);
+            if(!batchgeom && geombatches.length()) renderbatches(cur, pass);
+            break;
+
         case RENDERPASS_FOG:
             if(cur.vbuf!=va->vbuf)
             {
@@ -1532,7 +1710,11 @@ void renderva(renderstate &cur, vtxarray *va, int pass = RENDERPASS_LIGHTMAP, bo
             drawvatris(va, 3*va->tris, va->edata);
             xtravertsva += va->verts;
             break;
- 
+
+        case RENDERPASS_SHADOWMAP:
+            if(isshadowmapreceiver(va)) rendershadowmappass(cur, va);
+            break;
+
         case RENDERPASS_CAUSTICS:
             if(cur.vbuf!=va->vbuf) changevbuf(cur, pass, va);
             drawvatris(va, 3*va->tris, va->edata);
@@ -1551,30 +1733,6 @@ VAR(oqdist, 0, 256, 1024);
 VAR(zpass, 0, 1, 1);
 VAR(glowpass, 0, 1, 1);
 
-extern int ati_texgen_bug;
-
-static void setuptexgen(int dims = 2)
-{
-    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
-    glEnable(GL_TEXTURE_GEN_S);
-    if(dims>=2) 
-    {
-        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
-        glEnable(GL_TEXTURE_GEN_T);
-        if(ati_texgen_bug) glEnable(GL_TEXTURE_GEN_R);     // should not be needed, but apparently makes some ATI drivers happy
-    }
-}
-
-static void disabletexgen(int dims = 2)
-{
-    glDisable(GL_TEXTURE_GEN_S);
-    if(dims>=2)
-    {
-        glDisable(GL_TEXTURE_GEN_T);
-        if(ati_texgen_bug) glDisable(GL_TEXTURE_GEN_R);
-    }
-}
-
 GLuint fogtex = 0;
 
 void createfogtex()
@@ -1583,23 +1741,57 @@ void createfogtex()
     uchar buf[2*256] = { 255, 0, 255, 255 };
     if(!bilinear) loopi(256) { buf[2*i] = 255; buf[2*i+1] = i; }
     glGenTextures(1, &fogtex);
-    createtexture(fogtex, bilinear ? 2 : 256, 1, buf, 3, false, GL_LUMINANCE_ALPHA, GL_TEXTURE_1D);
+    createtexture(fogtex, bilinear ? 2 : 256, 1, buf, 3, 1, GL_LUMINANCE_ALPHA, GL_TEXTURE_1D);
+}
+
+GLuint attenxytex = 0, attenztex = 0;
+
+static GLuint createattenxytex(int size)
+{
+    uchar *data = new uchar[size*size], *dst = data;
+    loop(y, size) loop(x, size)
+    {
+        float dx = 2*float(x)/(size-1) - 1, dy = 2*float(y)/(size-1) - 1;
+        float atten = max(0.0f, 1.0f - dx*dx - dy*dy);
+        *dst++ = uchar(atten*255);
+    }
+    GLuint tex = 0;
+    glGenTextures(1, &tex);
+    createtexture(tex, size, size, data, 3, 1, GL_ALPHA);
+    delete[] data;
+    return tex;
+}
+
+static GLuint createattenztex(int size)
+{
+    uchar *data = new uchar[size], *dst = data;
+    loop(z, size) 
+    {
+        float dz = 2*float(z)/(size-1) - 1;
+        float atten = dz*dz;
+        *dst++ = uchar(atten*255);
+    }
+    GLuint tex = 0;
+    glGenTextures(1, &tex);
+    createtexture(tex, size, 1, data, 3, 1, GL_ALPHA, GL_TEXTURE_1D);
+    delete[] data;
+    return tex;
 }
 
 #define NUMCAUSTICS 32
 
-VARR(causticscale, 0, 100, 10000);
-VARR(causticmillis, 0, 75, 1000);
-VARP(caustics, 0, 1, 1);
-
 static Texture *caustictex[NUMCAUSTICS] = { NULL };
 
-void loadcaustics()
+void loadcaustics(bool force)
 {
+    static bool needcaustics = false;
+    if(force) needcaustics = true;
+    if(!caustics || !needcaustics) return;
+    useshaderbyname("caustic");
     if(caustictex[0]) return;
     loopi(NUMCAUSTICS)
     {
-        s_sprintfd(name)(
+        defformatstring(name)(
             renderpath==R_FIXEDFUNCTION ? 
                 "<mad:0.6,0.4>packages/caustics/caust%.2d.png" :
                 "<mad:-0.6,0.6>packages/caustics/caust%.2d.png",
@@ -1610,16 +1802,20 @@ void loadcaustics()
 
 void cleanupva()
 {
-    vaclearc(worldroot);
+    clearvas(worldroot);
     clearqueries();
     if(fogtex) { glDeleteTextures(1, &fogtex); fogtex = 0; }
+    if(attenxytex) { glDeleteTextures(1, &attenxytex); attenxytex = 0; }
+    if(attenztex) { glDeleteTextures(1, &attenztex); attenztex = 0; }
     loopi(NUMCAUSTICS) caustictex[i] = NULL;
 }
 
+VARR(causticscale, 0, 100, 10000);
+VARR(causticmillis, 0, 75, 1000);
+VARFP(caustics, 0, 1, 1, loadcaustics());
+
 void setupcaustics(int tmu, float blend, GLfloat *color = NULL)
 {
-    if(!caustictex[0]) loadcaustics();
-
     GLfloat s[4] = { 0.011f, 0, 0.0066f, 0 };
     GLfloat t[4] = { 0, 0.011f, 0.0066f, 0 };
     loopk(3)
@@ -1684,7 +1880,15 @@ void setupTMUs(renderstate &cur, float causticspass, bool fogpass)
                 cur.lightmaptmu = 3;
                 if(maxtmus>=5)
                 {
-                    if(fogpass) cur.fogtmu = 4;
+                    if(fogpass) 
+                    {
+                        if(glowpass && maxtmus>=6) 
+                        { 
+                            cur.fogtmu = 5;
+                            cur.glowtmu = 4;
+                        }
+                        else cur.fogtmu = 4;
+                    } 
                     else if(glowpass) cur.glowtmu = 4;
                 }
             }
@@ -1695,19 +1899,17 @@ void setupTMUs(renderstate &cur, float causticspass, bool fogpass)
         {
             glActiveTexture_(GL_TEXTURE0_ARB+cur.glowtmu);
             setuptexgen();
-            setuptmu(cur.glowtmu, "P + T");
+            setuptmu(cur.glowtmu, "P + T", "= Pa");
         }
         if(cur.fogtmu>=0)
         {
             glActiveTexture_(GL_TEXTURE0_ARB+cur.fogtmu);
             glEnable(GL_TEXTURE_1D);
             setuptexgen(1);
-            setuptmu(cur.fogtmu, "C , P @ Ta");
+            setuptmu(cur.fogtmu, "C , P @ Ta", "= Pa");
             if(!fogtex) createfogtex();
             glBindTexture(GL_TEXTURE_1D, fogtex);
-            uchar wcol[3];
-            getwatercolour(wcol);
-            loopk(3) cur.color[k] = wcol[k]/255.0f;
+            loopk(3) cur.color[k] = watercolor[k]/255.0f;
         }
         if(cur.causticstmu>=0) setupcaustics(cur.causticstmu, causticspass, cur.color);
     }
@@ -1718,7 +1920,7 @@ void setupTMUs(renderstate &cur, float causticspass, bool fogpass)
         glEnableClientState(GL_COLOR_ARRAY);
         loopi(8-2) { glActiveTexture_(GL_TEXTURE2_ARB+i); glEnable(GL_TEXTURE_2D); }
         glActiveTexture_(GL_TEXTURE0_ARB);
-        setenvparamf("ambient", SHPARAM_PIXEL, 5, hdr.ambient/255.0f, hdr.ambient/255.0f, hdr.ambient/255.0f);
+        setenvparamf("ambient", SHPARAM_PIXEL, 5, ambientcolor[0]/255.0f, ambientcolor[1]/255.0f, ambientcolor[2]/255.0f);
         setenvparamf("millis", SHPARAM_VERTEX, 6, lastmillis/1000.0f, lastmillis/1000.0f, lastmillis/1000.0f);
     }
  
@@ -1729,7 +1931,7 @@ void setupTMUs(renderstate &cur, float causticspass, bool fogpass)
         glActiveTexture_(GL_TEXTURE0_ARB+cur.lightmaptmu);
         glClientActiveTexture_(GL_TEXTURE0_ARB+cur.lightmaptmu);
 
-        setuptmu(cur.lightmaptmu, "P * T x 2");
+        setuptmu(cur.lightmaptmu, "P * T x 2", "= Ta");
         glEnable(GL_TEXTURE_2D);
         glEnableClientState(GL_TEXTURE_COORD_ARRAY);
         glMatrixMode(GL_TEXTURE);
@@ -1807,7 +2009,7 @@ void cleanupTMUs(renderstate &cur)
 #define FIRSTVA (reflecting ? reflectedva : visibleva)
 #define NEXTVA (reflecting ? va->rnext : va->next)
 
-void rendergeommultipass(renderstate &cur, int pass, bool fogpass)
+static void rendergeommultipass(renderstate &cur, int pass, bool fogpass)
 {
     cur.vbuf = 0;
     for(vtxarray *va = FIRSTVA; va; va = NEXTVA)
@@ -1832,6 +2034,10 @@ void rendergeommultipass(renderstate &cur, int pass, bool fogpass)
 VAR(oqgeom, 0, 1, 1);
 VAR(oqbatch, 0, 1, 1);
 
+VAR(dbgffsm, 0, 0, 1);
+VAR(dbgffdl, 0, 0, 1);
+VAR(ffdlscissor, 0, 1, 1);
+
 void rendergeom(float causticspass, bool fogpass)
 {
     renderstate cur;
@@ -1850,10 +2056,10 @@ void rendergeom(float causticspass, bool fogpass)
     if(!doOQ) 
     {
         setupTMUs(cur, causticspass, fogpass);
-        if(shadowmap) pushshadowmap();
+        if(shadowmap && !envmapping && !glaring && renderpath!=R_FIXEDFUNCTION) pushshadowmap();
     }
 
-    finddynlights();
+    int hasdynlights = finddynlights();
 
     glPushMatrix();
 
@@ -1861,6 +2067,7 @@ void rendergeom(float causticspass, bool fogpass)
 
     resetbatches();
 
+    int blends = 0;
     for(vtxarray *va = FIRSTVA; va; va = NEXTVA)
     {
         if(!va->texs) continue;
@@ -1941,6 +2148,7 @@ void rendergeom(float causticspass, bool fogpass)
             if(va->occluded >= OCCLUDE_GEOM) continue;
         }
 
+        if(!doOQ) blends += va->blends;
         renderva(cur, va, doOQ ? RENDERPASS_Z : (nolights ? RENDERPASS_COLOR : RENDERPASS_LIGHTMAP), fogpass, true);
     }
 
@@ -1952,7 +2160,7 @@ void rendergeom(float causticspass, bool fogpass)
     if(doOQ)
     {
         setupTMUs(cur, causticspass, fogpass);
-        if(shadowmap) 
+        if(shadowmap && !envmapping && !glaring && renderpath!=R_FIXEDFUNCTION) 
         {
             glPopMatrix();
             glPushMatrix();
@@ -1991,6 +2199,7 @@ void rendergeom(float causticspass, bool fogpass)
                 if(va->occluded >= OCCLUDE_GEOM) continue;
             }
             
+            blends += va->blends;
             renderva(cur, va, nolights ? RENDERPASS_COLOR : RENDERPASS_LIGHTMAP, fogpass);
         }
         if(geombatches.length()) renderbatches(cur, nolights ? RENDERPASS_COLOR : RENDERPASS_LIGHTMAP);
@@ -2009,6 +2218,8 @@ void rendergeom(float causticspass, bool fogpass)
                 va->occluded = pvsoccluded(va->geommin, va->geommax) ? OCCLUDE_GEOM : OCCLUDE_NOTHING;
                 if(va->occluded >= OCCLUDE_GEOM) continue;
             }
+
+            blends += va->blends;
             renderva(cur, va, nolights ? RENDERPASS_COLOR : RENDERPASS_LIGHTMAP, fogpass);
         }
         if(geombatches.length()) renderbatches(cur, nolights ? RENDERPASS_COLOR : RENDERPASS_LIGHTMAP);
@@ -2016,7 +2227,48 @@ void rendergeom(float causticspass, bool fogpass)
         if(foggedvas.empty()) glDepthFunc(GL_LESS);
     }
 
-    if(shadowmap) popshadowmap();
+    if(blends && (renderpath!=R_FIXEDFUNCTION || !nolights))
+    {
+        if(shadowmap && !envmapping && !glaring && renderpath!=R_FIXEDFUNCTION)
+        {
+            glPopMatrix();
+            glPushMatrix();
+            resetorigin();
+        }
+        if(foggedvas.empty()) glDepthFunc(GL_LEQUAL);
+        glDepthMask(GL_FALSE);
+        glEnable(GL_BLEND);
+        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_FALSE);
+
+        cur.vbuf = 0;
+        cur.blending = true;
+        for(vtxarray **prevva = &FIRSTVA, *va = FIRSTVA; va; prevva = &NEXTVA, va = NEXTVA)
+        {
+            if(!va->blends || va->occluded >= OCCLUDE_GEOM) continue;
+            if(refracting)
+            {
+                if(refracting < 0 ? va->geommin.z > reflectz : va->geommax.z <= reflectz) continue;
+                if(isvisiblecube(va->o, va->size) >= VFC_NOT_VISIBLE) continue;
+                if((!hasOQ || !oqfrags) && va->distance > reflectdist) break;
+            }
+            else if(reflecting)
+            {
+                if(va->geommax.z <= reflectz || (va->rquery && checkquery(va->rquery))) continue;
+            }
+            if(fogpass ? va->geommax.z <= reflectz-waterfog : va->curvfc==VFC_FOGGED) continue;
+            renderva(cur, va, RENDERPASS_LIGHTMAP_BLEND, fogpass);
+        }
+        if(geombatches.length()) renderbatches(cur, RENDERPASS_LIGHTMAP);
+        cur.blending = false;
+
+        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
+        glDisable(GL_BLEND);
+        glDepthMask(GL_TRUE);
+        if(foggedvas.empty()) glDepthFunc(GL_LESS);
+    }
+
+    if(shadowmap && !envmapping && !glaring && renderpath!=R_FIXEDFUNCTION) popshadowmap();
 
     cleanupTMUs(cur);
 
@@ -2026,7 +2278,7 @@ void rendergeom(float causticspass, bool fogpass)
         if(doOQ) glDepthFunc(GL_LESS);
     }
 
-    if(renderpath==R_FIXEDFUNCTION ? (glowpass && cur.skippedglow) || (causticspass>=1 && cur.causticstmu<0) || (fogpass && cur.fogtmu<0) : causticspass)
+    if(renderpath==R_FIXEDFUNCTION ? (glowpass && cur.skippedglow) || (causticspass>=1 && cur.causticstmu<0) || (fogpass && cur.fogtmu<0) || (shadowmap && shadowmapcasters) || hasdynlights : causticspass)
     {
         glDepthFunc(GL_LEQUAL);
         glDepthMask(GL_FALSE);
@@ -2060,7 +2312,7 @@ void rendergeom(float causticspass, bool fogpass)
                 resettmu(0);
                 glActiveTexture_(GL_TEXTURE1_ARB);
                 resettmu(1);
-                disabletexgen();
+                disabletexgen(1);
                 glDisable(GL_TEXTURE_1D);
                 glActiveTexture_(GL_TEXTURE0_ARB);
             } 
@@ -2087,7 +2339,101 @@ void rendergeom(float causticspass, bool fogpass)
             }
             glActiveTexture_(GL_TEXTURE0_ARB);
         }
-    
+
+        if(renderpath==R_FIXEDFUNCTION && shadowmap && shadowmapcasters)
+        {
+            glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
+            glFogfv(GL_FOG_COLOR, zerofog);
+            glPopMatrix();
+            glPushMatrix();
+            pushshadowmap();
+            resetorigin();
+            if(cur.fogtmu>=0)
+            {
+                setuptmu(0, "C * T");
+                glActiveTexture_(GL_TEXTURE1_ARB);
+                glEnable(GL_TEXTURE_1D);
+                setuptexgen(1);
+                setuptmu(1, "P * T~a");
+                if(!fogtex) createfogtex();
+                glBindTexture(GL_TEXTURE_1D, fogtex);
+                glActiveTexture_(GL_TEXTURE0_ARB);
+            }
+            if(dbgffsm) { glDisable(GL_BLEND); glDisable(GL_TEXTURE_2D); glColor3f(1, 0, 1); }
+            rendergeommultipass(cur, RENDERPASS_SHADOWMAP, fogpass);
+            if(dbgffsm) { glEnable(GL_BLEND); glEnable(GL_TEXTURE_2D); }
+            popshadowmap();
+            if(cur.fogtmu>=0)
+            {
+                resettmu(0);
+                glActiveTexture_(GL_TEXTURE1_ARB);
+                resettmu(1);
+                disabletexgen();
+                glDisable(GL_TEXTURE_1D);
+                glActiveTexture_(GL_TEXTURE0_ARB);
+            }
+        }
+   
+        if(renderpath==R_FIXEDFUNCTION && hasdynlights)
+        {
+            glBlendFunc(GL_SRC_ALPHA, dbgffdl ? GL_ZERO : GL_ONE);
+            glFogfv(GL_FOG_COLOR, zerofog);
+
+            if(!attenxytex) attenxytex = createattenxytex(64);
+            glBindTexture(GL_TEXTURE_2D, attenxytex);
+
+            setuptmu(0, "= C", "= Ta");
+            setuptexgen();
+
+            glActiveTexture_(GL_TEXTURE1_ARB);
+            setuptmu(1, "= P", "Pa - Ta");
+            setuptexgen(1);
+            if(!attenztex) attenztex = createattenztex(64);
+            glBindTexture(GL_TEXTURE_1D, attenztex);
+            glEnable(GL_TEXTURE_1D);
+ 
+            glActiveTexture_(GL_TEXTURE2_ARB);
+            cur.diffusetmu = 2;
+            setuptmu(2, "P * T x 4", "= Pa");
+            setuptexgen();
+            glEnable(GL_TEXTURE_2D);
+
+            vec lightcolor;
+            for(int n = 0; getdynlight(n, cur.dynlightpos, cur.dynlightradius, lightcolor); n++)
+            {
+                lightcolor.mul(0.5f);
+                if(fogpass && cur.fogtmu>=0)
+                {
+                    float fog = (reflectz - cur.dynlightpos.z)/waterfog;
+                    if(fog >= 1.0f) continue;
+                    lightcolor.mul(1.0f - max(fog, 0.0f));
+                }
+                glColor3f(lightcolor.x, lightcolor.y, lightcolor.z);
+                if(ffdlscissor)
+                {
+                    float sx1, sy1, sx2, sy2;
+                    calcspherescissor(cur.dynlightpos, cur.dynlightradius, sx1, sy1, sx2, sy2);
+                    pushscissor(sx1, sy1, sx2, sy2);
+                }
+                resetorigin();
+                rendergeommultipass(cur, RENDERPASS_DYNLIGHT, fogpass);
+                if(ffdlscissor) popscissor();
+            }
+
+            glDisable(GL_TEXTURE_2D);
+            disabletexgen();
+            resettmu(2);
+
+            glActiveTexture_(GL_TEXTURE1_ARB);
+            glDisable(GL_TEXTURE_1D);
+            resettmu(1);
+            disabletexgen(1);
+            
+            glActiveTexture_(GL_TEXTURE0_ARB);
+            resettmu(0);
+            disabletexgen();
+        }
+
         if(renderpath==R_FIXEDFUNCTION && fogpass && cur.fogtmu<0)
         {
             glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
@@ -2097,15 +2443,13 @@ void rendergeom(float causticspass, bool fogpass)
             if(!fogtex) createfogtex();
             glBindTexture(GL_TEXTURE_1D, fogtex);
             setuptexgen(1);
-            uchar wcol[3];
-            getwatercolour(wcol);
-            glColor3ubv(wcol);
+            glColor3ubv(watercolor.v);
             rendergeommultipass(cur, RENDERPASS_FOG, fogpass);
             disabletexgen(1);
             glDisable(GL_TEXTURE_1D);
             glEnable(GL_TEXTURE_2D);
         }
- 
+
         glFogfv(GL_FOG_COLOR, oldfogc);
         glDisable(GL_BLEND);
         glDepthFunc(GL_LESS);
