@@ -1,4 +1,3 @@
-#include "pch.h"
 #include "engine.h"
 #include "rendertarget.h"
 
@@ -8,10 +7,17 @@ extern void cleanshadowmap();
 VARFP(shadowmapsize, 7, 9, 11, cleanshadowmap());
 VARP(shadowmapradius, 64, 96, 256);
 VAR(shadowmapheight, 0, 32, 128);
+VARP(ffshadowmapdist, 128, 1024, 4096);
 VARP(shadowmapdist, 128, 256, 512);
 VARFP(fpshadowmap, 0, 0, 1, cleanshadowmap());
 VARFP(shadowmapprecision, 0, 0, 1, cleanshadowmap());
-VARR(shadowmapambient, 0, 0, 0xFFFFFF);
+bvec shadowmapambientcolor(0, 0, 0);
+HVARFR(shadowmapambient, 0, 0, 0xFFFFFF,
+{
+    if(shadowmapambient <= 255) shadowmapambient |= (shadowmapambient<<8) | (shadowmapambient<<16);
+    shadowmapambientcolor = bvec((shadowmapambient>>16)&0xFF, (shadowmapambient>>8)&0xFF, shadowmapambient&0xFF);
+});
+VARP(shadowmapintensity, 0, 40, 100);
 
 VARP(blurshadowmap, 0, 1, 3);
 VARP(blursmsigma, 1, 100, 200);
@@ -20,6 +26,7 @@ VARP(blursmsigma, 1, 100, 200);
 
 vec shadowoffset(0, 0, 0), shadowfocus(0, 0, 0), shadowdir(0, SHADOWSKEW, 1);
 VAR(shadowmapcasters, 1, 0, 0);
+float shadowmapmaxz = 0;
 
 void setshadowdir(int angle)
 {
@@ -34,7 +41,7 @@ void guessshadowdir()
     if(shadowmapangle) return;
     vec lightpos(0, 0, 0), casterpos(0, 0, 0);
     int numlights = 0, numcasters = 0;
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
     loopv(ents)
     {
         extentity &e = *ents[i];
@@ -71,7 +78,7 @@ void guessshadowdir()
 
 bool shadowmapping = false;
 
-static GLdouble shadowmapprojection[16], shadowmapmodelview[16];
+static glmatrixf shadowmapmatrix;
 
 VARP(shadowmapbias, 0, 5, 1024);
 VARP(shadowmappeelbias, 0, 20, 1024);
@@ -80,14 +87,25 @@ VAR(smoothshadowmappeel, 1, 0, 0);
 
 static struct shadowmaptexture : rendertarget
 {
-    const GLenum *colorformats() const
+    GLenum attachment() const
     {
-        static const GLenum colorfmts[] = { GL_RGBA16F_ARB, GL_RGBA16, GL_RGBA, GL_RGBA8, GL_FALSE };
-        int offset = fpshadowmap && hasTF && hasFBO ? 0 : (shadowmapprecision && hasFBO ? 1 : 2);
-        return &colorfmts[offset];
+        return renderpath==R_FIXEDFUNCTION ? GL_DEPTH_ATTACHMENT_EXT : GL_COLOR_ATTACHMENT0_EXT;
     }
 
-    bool swaptexs() const { return true; }
+    const GLenum *colorformats() const
+    {
+        if(renderpath==R_FIXEDFUNCTION) 
+        {
+            static const GLenum depthtexfmts[] = { GL_DEPTH_COMPONENT16, GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT32, GL_FALSE };
+            return depthtexfmts;
+        }
+       
+        static const GLenum rgbfmts[] = { GL_RGB, GL_RGB8, GL_FALSE }, rgbafmts[] = { GL_RGBA16F_ARB, GL_RGBA16, GL_RGBA, GL_RGBA8, GL_FALSE };
+        return hasFBO ? &rgbafmts[fpshadowmap && hasTF ? 0 : (shadowmapprecision ? 1 : 2)] : rgbfmts;
+    }
+
+    bool filter() const { return renderpath!=R_FIXEDFUNCTION || hasNVPCF; }
+    bool swaptexs() const { return renderpath!=R_FIXEDFUNCTION; }
 
     bool scissorblur(int &x, int &y, int &w, int &h)
     {
@@ -108,8 +126,14 @@ static struct shadowmaptexture : rendertarget
 
     void doclear()
     {
+        if(!hasFBO && rtscissor)
+        {
+            glEnable(GL_SCISSOR_TEST);
+            glScissor(screen->w-vieww, screen->h-viewh, vieww, viewh);
+        }
         glClearColor(0, 0, 0, 0);
-        glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
+        glClear(GL_DEPTH_BUFFER_BIT | (renderpath!=R_FIXEDFUNCTION ? GL_COLOR_BUFFER_BIT : 0));
+        if(!hasFBO && rtscissor) glDisable(GL_SCISSOR_TEST);
     }
 
     bool dorender()
@@ -120,7 +144,9 @@ static struct shadowmaptexture : rendertarget
         glMatrixMode(GL_PROJECTION);
         glPushMatrix();
         glLoadIdentity();
-        glOrtho(-shadowmapradius, shadowmapradius, -shadowmapradius, shadowmapradius, -shadowmapdist, shadowmapdist);
+        glOrtho(-shadowmapradius, shadowmapradius, -shadowmapradius, shadowmapradius, 
+            renderpath==R_FIXEDFUNCTION ? 0 : -shadowmapdist, 
+            renderpath==R_FIXEDFUNCTION ? ffshadowmapdist : shadowmapdist);
 
         glMatrixMode(GL_MODELVIEW);
 
@@ -152,21 +178,34 @@ static struct shadowmaptexture : rendertarget
         glTranslatef(-camera1->o.x, -camera1->o.y, -camera1->o.z);
         shadowfocus = camera1->o;
         shadowfocus.add(dir);
-        shadowfocus.add(vec(-shadowdir.x, -shadowdir.y, 1).mul(shadowmapheight));
+        shadowfocus.add(vec(shadowdir).mul(shadowmapheight));
         shadowfocus.add(dirx.mul(shadowoffset.x));
         shadowfocus.add(diry.mul(shadowoffset.y));
 
-        glGetDoublev(GL_PROJECTION_MATRIX, shadowmapprojection);
-        glGetDoublev(GL_MODELVIEW_MATRIX, shadowmapmodelview);
+        glmatrixf proj, mv;
+        glGetFloatv(GL_PROJECTION_MATRIX, proj.v);
+        glGetFloatv(GL_MODELVIEW_MATRIX, mv.v);
+        shadowmapmatrix.mul(proj, mv);
+        if(renderpath==R_FIXEDFUNCTION) shadowmapmatrix.projective();
+        else shadowmapmatrix.projective(-1, 1-shadowmapbias/float(shadowmapdist));
 
-        setenvparamf("shadowmapbias", SHPARAM_VERTEX, 0, -shadowmapbias/float(shadowmapdist), 1 - (shadowmapbias + (smoothshadowmappeel ? 0 : shadowmappeelbias))/float(shadowmapdist));
+        glColor3f(0, 0, 0);
+        glDisable(GL_TEXTURE_2D);
+
+        if(renderpath!=R_FIXEDFUNCTION) setenvparamf("shadowmapbias", SHPARAM_VERTEX, 0, -shadowmapbias/float(shadowmapdist), 1 - (shadowmapbias + (smoothshadowmappeel ? 0 : shadowmappeelbias))/float(shadowmapdist));
+        else glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
 
         shadowmapcasters = 0;
+        shadowmapmaxz = shadowfocus.z - shadowmapdist;
         shadowmapping = true;
         rendergame();
         shadowmapping = false;
+        shadowmapmaxz = min(shadowmapmaxz, shadowfocus.z);
 
-        if(shadowmapcasters && smdepthpeel) 
+        glEnable(GL_TEXTURE_2D);
+
+        if(renderpath==R_FIXEDFUNCTION) glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
+        else if(shadowmapcasters && smdepthpeel) 
         {
             int sx, sy, sw, sh;
             bool scissoring = rtscissor && scissorblur(sx, sy, sw, sh) && sw > 0 && sh > 0;
@@ -211,6 +250,66 @@ void cleanshadowmap()
     shadowmaptex.cleanup(true);
 }
 
+VAR(ffsmscissor, 0, 1, 1);
+
+static void calcscissorbox()
+{
+    int smx, smy, smw, smh;
+    shadowmaptex.scissorblur(smx, smy, smw, smh);
+
+    vec forward, right;
+    vecfromyawpitch(camera1->yaw, 0, -1, 0, forward);
+    vecfromyawpitch(camera1->yaw, 0, 0, -1, right);
+    forward.mul(shadowmapradius*2.0f/shadowmaptex.viewh);
+    right.mul(shadowmapradius*2.0f/shadowmaptex.vieww);
+
+    vec bottom(shadowfocus);
+    bottom.sub(vec(shadowdir).mul(shadowmapdist));
+    bottom.add(vec(forward).mul(smy - shadowmaptex.viewh/2)).add(vec(right).mul(smx - shadowmaptex.vieww/2));
+    vec top(bottom);
+    top.add(vec(shadowdir).mul(shadowmapmaxz - (shadowfocus.z - shadowmapdist)));
+   
+    vec4 v[8];
+    float sx1 = 1, sy1 = 1, sx2 = -1, sy2 = -1;
+    loopi(8)
+    {
+        vec c = i&4 ? top : bottom;
+        if(i&1) c.add(vec(right).mul(smw));
+        if(i&2) c.add(vec(forward).mul(smh));
+        if(reflecting) c.z = 2*reflectz - c.z;
+        vec4 &p = v[i];
+        mvpmatrix.transform(c, p);
+        if(p.z >= 0)
+        {
+            float x = p.x / p.w, y = p.y / p.w;
+            sx1 = min(sx1, x);
+            sy1 = min(sy1, y);
+            sx2 = max(sx2, x);
+            sy2 = max(sy2, y);
+        }
+    }
+    if(sx1 >= sx2 || sy1 >= sy2) return;
+    loopi(8)
+    {
+        const vec4 &p = v[i];
+        if(p.z >= 0) continue;
+        loopj(3)
+        {
+            const vec4 &o = v[i^(1<<j)];
+            if(o.z <= 0) continue;
+            float t = p.z/(p.z - o.z),
+                  w = p.w + t*(o.w - p.w),
+                  x = (p.x + t*(o.x - p.x))/w,
+                  y = (p.y + t*(o.y - p.y))/w;
+            sx1 = min(sx1, x);
+            sy1 = min(sy1, y);
+            sx2 = max(sx2, x);
+            sy2 = max(sy2, y);
+        }
+    }
+    pushscissor(sx1, sy1, sx2, sy2);
+}
+
 void calcshadowmapbb(const vec &o, float xyrad, float zrad, float &x1, float &y1, float &x2, float &y2)
 {
     vec skewdir(shadowdir);
@@ -239,6 +338,8 @@ bool addshadowmapcaster(const vec &o, float xyrad, float zrad)
 {
     if(o.z + zrad <= shadowfocus.z - shadowmapdist || o.z - zrad >= shadowfocus.z) return false;
 
+    shadowmapmaxz = max(shadowmapmaxz, o.z + zrad);
+
     float x1, y1, x2, y2;
     calcshadowmapbb(o, xyrad, zrad, x1, y1, x2, y2);
 
@@ -250,9 +351,9 @@ bool addshadowmapcaster(const vec &o, float xyrad, float zrad)
 
 bool isshadowmapreceiver(vtxarray *va)
 {
-    if(!shadowmap || renderpath==R_FIXEDFUNCTION || !shadowmapcasters) return false;
+    if(!shadowmap || !shadowmapcasters) return false;
 
-    if(va->shadowmapmax.z <= shadowfocus.z - shadowmapdist || va->shadowmapmin.z >= shadowfocus.z) return false;
+    if(va->shadowmapmax.z <= shadowfocus.z - shadowmapdist || va->shadowmapmin.z >= shadowmapmaxz) return false;
 
     float xyrad = SQRT2*0.5f*max(va->shadowmapmax.x-va->shadowmapmin.x, va->shadowmapmax.y-va->shadowmapmin.y),
           zrad = 0.5f*(va->shadowmapmax.z-va->shadowmapmin.z),
@@ -268,10 +369,10 @@ bool isshadowmapreceiver(vtxarray *va)
 #if 0
     // cheaper inexact test
     float dz = va->o.z + va->size/2 - shadowfocus.z;
-    float cx = shadowfocus.x - dz*shadowdir.x, cy = shadowfocus.y - dz*shadowdir.y;
+    float cx = shadowfocus.x + dz*shadowdir.x, cy = shadowfocus.y + dz*shadowdir.y;
     float skew = va->size/2*SHADOWSKEW;
     if(!shadowmap || !shadowmaptex ||
-       va->o.z + va->size <= shadowfocus.z - shadowmapdist || va->o.z >= shadowfocus.z ||
+       va->o.z + va->size <= shadowfocus.z - shadowmapdist || va->o.z >= shadowmapmaxz ||
        va->o.x + va->size <= cx - shadowmapradius-skew || va->o.x >= cx + shadowmapradius+skew || 
        va->o.y + va->size <= cy - shadowmapradius-skew || va->o.y >= cy + shadowmapradius+skew) 
         return false;
@@ -283,7 +384,7 @@ bool isshadowmapcaster(const vec &o, float rad)
 {
     // cheaper inexact test
     float dz = o.z - shadowfocus.z;
-    float cx = shadowfocus.x - dz*shadowdir.x, cy = shadowfocus.y - dz*shadowdir.y;
+    float cx = shadowfocus.x + dz*shadowdir.x, cy = shadowfocus.y + dz*shadowdir.y;
     float skew = rad*SHADOWSKEW;
     if(!shadowmapping ||
        o.z + rad <= shadowfocus.z - shadowmapdist || o.z - rad >= shadowfocus.z ||
@@ -297,36 +398,77 @@ void pushshadowmap()
 {
     if(!shadowmap || !shadowmaptex.rendertex) return;
 
+    if(renderpath==R_FIXEDFUNCTION)
+    {
+        static GLfloat texgenS[4] = { 1, 0, 0, 0 },
+                       texgenT[4] = { 0, 1, 0, 0 },
+                       texgenR[4] = { 0, 0, 1, 0 };
+
+        glBindTexture(GL_TEXTURE_2D, shadowmaptex.rendertex);
+
+        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
+        glTexGenfv(GL_S, GL_OBJECT_PLANE, texgenS);
+        glEnable(GL_TEXTURE_GEN_S);
+
+        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
+        glTexGenfv(GL_T, GL_OBJECT_PLANE, texgenT);
+        glEnable(GL_TEXTURE_GEN_T);
+
+        glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
+        glTexGenfv(GL_R, GL_OBJECT_PLANE, texgenR);
+        glEnable(GL_TEXTURE_GEN_R);
+
+        // intel driver bug workaround: when R texgen is enabled, it uses the value of Q, even if not enabled!
+        // MUST set Q with glTexCoord4f, glTexCoord3f does not work
+        glTexCoord4f(0, 0, 0, 1);
+
+        if(hasDT && hasSH)
+        {
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE);
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_GEQUAL);
+            glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_LUMINANCE);
+        }
+        else
+        {
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_SGIX, GL_TRUE);
+            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_OPERATOR_SGIX, GL_TEXTURE_GEQUAL_R_SGIX);
+        }
+
+        glColor3f(shadowmapintensity/100.0f, shadowmapintensity/100.0f, shadowmapintensity/100.0f);
+
+        if(ffsmscissor) calcscissorbox();
+        return;
+    }
+
     glActiveTexture_(GL_TEXTURE7_ARB);
-    glEnable(GL_TEXTURE_2D);
     glBindTexture(GL_TEXTURE_2D, shadowmaptex.rendertex);
 
     glActiveTexture_(GL_TEXTURE2_ARB);
-    glEnable(GL_TEXTURE_2D);
     glMatrixMode(GL_TEXTURE);
-    glLoadIdentity();
-    glTranslatef(0.5f, 0.5f, 1-shadowmapbias/float(shadowmapdist));
-    glScalef(0.5f, 0.5f, -1);
-    glMultMatrixd(shadowmapprojection);
-    glMultMatrixd(shadowmapmodelview);
+    glLoadMatrixf(shadowmapmatrix.v);
     glPushMatrix();
     glMatrixMode(GL_MODELVIEW);
+
     glActiveTexture_(GL_TEXTURE0_ARB);
     glClientActiveTexture_(GL_TEXTURE0_ARB);
 
     float r, g, b;
 	if(!shadowmapambient)
 	{
-		if(hdr.skylight[0] || hdr.skylight[1] || hdr.skylight[2])
+		if(skylightcolor[0] || skylightcolor[1] || skylightcolor[2])
 		{
-			r = max(25.0f, 0.4f*hdr.ambient + 0.6f*max(hdr.ambient, hdr.skylight[0]));
-			g = max(25.0f, 0.4f*hdr.ambient + 0.6f*max(hdr.ambient, hdr.skylight[1]));
-			b = max(25.0f, 0.4f*hdr.ambient + 0.6f*max(hdr.ambient, hdr.skylight[2]));
+			r = max(25.0f, 0.4f*ambientcolor[0] + 0.6f*max(ambientcolor[0], skylightcolor[0]));
+			g = max(25.0f, 0.4f*ambientcolor[1] + 0.6f*max(ambientcolor[1], skylightcolor[1]));
+			b = max(25.0f, 0.4f*ambientcolor[2] + 0.6f*max(ambientcolor[2], skylightcolor[2]));
 		}
-		else r = g = b = max(25.0f, 2.0f*hdr.ambient);
+		else 
+        {
+            r = max(25.0f, 2.0f*ambientcolor[0]);
+            g = max(25.0f, 2.0f*ambientcolor[1]);
+            b = max(25.0f, 2.0f*ambientcolor[2]);
+        }
 	}
-    else if(shadowmapambient<=255) r = g = b = shadowmapambient;
-    else { r = (shadowmapambient>>16)&0xFF; g = (shadowmapambient>>8)&0xFF; b = shadowmapambient&0xFF; }
+    else { r = shadowmapambientcolor[0]; g = shadowmapambientcolor[1]; b = shadowmapambientcolor[2]; }
     setenvparamf("shadowmapambient", SHPARAM_PIXEL, 7, r/255.0f, g/255.0f, b/255.0f);
 }
 
@@ -334,39 +476,63 @@ void adjustshadowmatrix(const ivec &o, float scale)
 {
     if(!shadowmap || !shadowmaptex.rendertex) return;
 
-    glActiveTexture_(GL_TEXTURE2_ARB);
-    glMatrixMode(GL_TEXTURE);
-    glPopMatrix();
-    glPushMatrix();
-    glTranslatef(o.x, o.y, o.z);
-    glScalef(scale, scale, scale);
-    glMatrixMode(GL_MODELVIEW);
-    glActiveTexture_(GL_TEXTURE0_ARB);
+    if(renderpath==R_FIXEDFUNCTION)
+    {
+        const GLfloat *v = shadowmapmatrix.v;
+        GLfloat texgenS[4] = { v[0]*scale, v[4]*scale, v[8]*scale, v[0]*o.x + v[4]*o.y + v[8]*o.z + v[12] },
+                texgenT[4] = { v[1]*scale, v[5]*scale, v[9]*scale, v[1]*o.x + v[5]*o.y + v[9]*o.z + v[13] },
+                texgenR[4] = { v[2]*scale, v[6]*scale, v[10]*scale, v[2]*o.x + v[6]*o.y + v[10]*o.z + v[14] };
+        glTexGenfv(GL_S, GL_OBJECT_PLANE, texgenS);
+        glTexGenfv(GL_T, GL_OBJECT_PLANE, texgenT);
+        glTexGenfv(GL_R, GL_OBJECT_PLANE, texgenR);
+    }
+    else
+    {
+        glActiveTexture_(GL_TEXTURE2_ARB);
+        glMatrixMode(GL_TEXTURE);
+        glPopMatrix();
+        glPushMatrix();
+        glTranslatef(o.x, o.y, o.z);
+        glScalef(scale, scale, scale);
+        glMatrixMode(GL_MODELVIEW);
+        glActiveTexture_(GL_TEXTURE0_ARB);
+    }
 }
 
 void popshadowmap()
 {
     if(!shadowmap || !shadowmaptex.rendertex) return;
 
-    glActiveTexture_(GL_TEXTURE7_ARB);
-    glDisable(GL_TEXTURE_2D);
+    if(renderpath!=R_FIXEDFUNCTION) 
+    {
+        glActiveTexture_(GL_TEXTURE2_ARB);
+        glMatrixMode(GL_TEXTURE);
+        glPopMatrix();
+        glMatrixMode(GL_MODELVIEW);
 
-    glActiveTexture_(GL_TEXTURE2_ARB);
-    glMatrixMode(GL_TEXTURE);
-    glPopMatrix();
-    glMatrixMode(GL_MODELVIEW);
+        glActiveTexture_(GL_TEXTURE0_ARB);
+    }
+    else
+    {
+        popscissor();
 
-    glActiveTexture_(GL_TEXTURE0_ARB);
+        glDisable(GL_TEXTURE_GEN_S);
+        glDisable(GL_TEXTURE_GEN_T);
+        glDisable(GL_TEXTURE_GEN_R);
+
+        if(hasDT && hasSH) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE);
+        else glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_SGIX, GL_FALSE);
+    }
 }
 
 void rendershadowmap()
 {
-    if(!shadowmap || renderpath==R_FIXEDFUNCTION) return;
+    if(!shadowmap || (renderpath==R_FIXEDFUNCTION && (!hasSGIDT || !hasSGISH))) return;
 
     // Apple/ATI bug - fixed-function fog state can force software fallback even when fragment program is enabled
-    glDisable(GL_FOG); 
-    shadowmaptex.render(1<<shadowmapsize, 1<<shadowmapsize, blurshadowmap, blursmsigma/100.0f);
-    glEnable(GL_FOG);
+    if(renderpath!=R_FIXEDFUNCTION || !fogging) glDisable(GL_FOG); 
+    shadowmaptex.render(1<<shadowmapsize, 1<<shadowmapsize, renderpath!=R_FIXEDFUNCTION ? blurshadowmap : 0, blursmsigma/100.0f);
+    if(renderpath!=R_FIXEDFUNCTION || !fogging) glEnable(GL_FOG);
 }
 
 VAR(debugsm, 0, 0, 1);
