@@ -12,7 +12,7 @@ struct editline
         set(init);
     }
 
-    bool empty() { return len <= 1; }
+    bool empty() { return len <= 0; }
 
     void clear()
     {
@@ -71,11 +71,11 @@ struct editline
         len += slen;
     }
 
-    bool read(FILE *f, int chop = -1)
+    bool read(stream *f, int chop = -1)
     {
         if(chop < 0) chop = INT_MAX; else chop++;
         set("");
-        while(len + 1 < chop && fgets(&text[len], min(maxlen, chop) - len, f))
+        while(len + 1 < chop && f->getline(&text[len], min(maxlen, chop) - len))
         {
             len += strlen(&text[len]);
             if(len > 0 && text[len-1] == '\n')
@@ -88,7 +88,7 @@ struct editline
         if(len + 1 >= chop)
         {
             char buf[CHUNKSIZE];
-            while(fgets(buf, sizeof(buf), f))
+            while(f->getline(buf, sizeof(buf)))
             {
                 int blen = strlen(buf);
                 if(blen > 0 && buf[blen-1] == '\n') return true;
@@ -190,12 +190,12 @@ struct editor
     {
         if(!filename) return;
         clear(NULL);
-        FILE *file = openfile(filename, "r");
+        stream *file = openfile(filename, "r");
         if(file) 
         {
             while(lines.add().read(file, maxx) && (maxy < 0 || lines.length() <= maxy));
             lines.pop().clear();
-            fclose(file);
+            delete file;
         }
         if(lines.empty()) lines.add().set("");
     }
@@ -203,10 +203,10 @@ struct editor
     void save()
     {
         if(!filename) return;
-        FILE *file = openfile(filename, "w");
+        stream *file = openfile(filename, "w");
         if(!file) return;
-        loopv(lines) fprintf(file, "%s\n", lines[i].text);
-        fclose(file);
+        loopv(lines) file->putline(lines[i].text);
+        delete file;
     }
    
     void mark(bool enable) 
@@ -332,7 +332,11 @@ struct editor
     void del() // removes the current selection (if any)
     {
         int sx, sy, ex, ey;
-        if(!region(sx, sy, ex, ey)) return;
+        if(!region(sx, sy, ex, ey)) 
+        { 
+            mark(false); 
+            return; 
+        }
         if(sy == ey) 
         {
             if(sx == 0 || ex == lines[ey].len) removelines(sy, 1);
@@ -374,6 +378,11 @@ struct editor
         }
     }
 
+    void insert(const char *s)
+    {
+        while(*s) insert(*s++);
+    }
+
     void insertallfrom(editor *b) 
     {   
         if(b==this) return;
@@ -487,6 +496,9 @@ struct editor
                 }
                 break;
             }
+            case SDLK_LSHIFT:
+            case SDLK_RSHIFT:
+                break;
             case SDLK_RETURN:    
                 cooked = '\n';
                 // fall through
@@ -627,7 +639,7 @@ struct editor
 // a 'stack' where the last is the current focused editor
 static vector <editor*> editors;
 
-static editor *currentfocus() { return (editors.length() > 0)?editors.last():NULL; }
+static editor *currentfocus() { return editors.length() ? editors.last() : NULL; }
 
 static void readyeditors() 
 {
@@ -668,8 +680,8 @@ ICOMMAND(textlist, "", (), // @DEBUG return list of all the editors
     string s = "";
     loopv(editors)
     {   
-        if(i > 0) s_strcat(s, ", ");
-        s_strcat(s, editors[i]->name);
+        if(i > 0) concatstring(s, ", ");
+        concatstring(s, editors[i]->name);
     }
     result(s);
 );
@@ -688,7 +700,7 @@ TEXTCOMMAND(textmode, "i", (int *m), // (1= keep while focused, 2= keep while us
     if(*m) top->mode = *m;
     else
     {
-        s_sprintfd(s)("%d", top->mode);
+        defformatstring(s)("%d", top->mode);
         result(s);
     } 
 );
