@@ -1,10 +1,11 @@
 // world.cpp: core map management stuff
 
-#include "pch.h"
 #include "engine.h"
 
-header hdr;
-int worldscale;
+VARR(mapversion, 1, MAPVERSION, 0);
+VARNR(mapscale, worldscale, 1, 0, 0);
+VARNR(mapsize, worldsize, 1, 0, 0);
+SVARR(maptitle, "Untitled Map by Unknown");
 
 VAR(octaentsize, 0, 128, 1024);
 VAR(entselradius, 0, 2, 10);
@@ -61,10 +62,10 @@ void modifyoctaentity(int flags, int id, cube *c, const ivec &cor, int size, con
         {
             if(!c[i].ext || !c[i].ext->ents) ext(c[i]).ents = new octaentities(o, size);
             octaentities &oe = *c[i].ext->ents;
-            switch(et->getents()[id]->type)
+            switch(entities::getents()[id]->type)
             {
                 case ET_MAPMODEL:
-                    if(loadmodel(NULL, et->getents()[id]->attr2))
+                    if(loadmodel(NULL, entities::getents()[id]->attr2))
                     {
                         if(va)
                         {
@@ -93,10 +94,10 @@ void modifyoctaentity(int flags, int id, cube *c, const ivec &cor, int size, con
         else if(c[i].ext && c[i].ext->ents)
         {
             octaentities &oe = *c[i].ext->ents;
-            switch(et->getents()[id]->type)
+            switch(entities::getents()[id]->type)
             {
                 case ET_MAPMODEL:
-                    if(loadmodel(NULL, et->getents()[id]->attr2))
+                    if(loadmodel(NULL, entities::getents()[id]->attr2))
                     {
                         oe.mapmodels.removeobj(id);
                         if(va)
@@ -112,7 +113,7 @@ void modifyoctaentity(int flags, int id, cube *c, const ivec &cor, int size, con
                         oe.bbmin.add(oe.size);
                         loopvj(oe.mapmodels)
                         {
-                            extentity &e = *et->getents()[oe.mapmodels[j]];
+                            extentity &e = *entities::getents()[oe.mapmodels[j]];
                             ivec eo, er;
                             if(getentboundingbox(e, eo, er)) loopk(3)
                             {
@@ -147,23 +148,38 @@ void modifyoctaentity(int flags, int id, cube *c, const ivec &cor, int size, con
     }
 }
 
-static void modifyoctaent(int flags, int id)
+vector<int> outsideents;
+
+static bool modifyoctaent(int flags, int id)
 {
-    vector<extentity *> &ents = et->getents();
-    if(!ents.inrange(id)) return;
+    vector<extentity *> &ents = entities::getents();
+    if(!ents.inrange(id)) return false;
     ivec o, r;
     extentity &e = *ents[id];
-    if((e.inoctanode!=0)==flags || !getentboundingbox(e, o, r)) return;
-
-    int leafsize = octaentsize, limit = max(r.x, max(r.y, r.z));
-    while(leafsize < limit) leafsize *= 2;
-    int diff = ~(leafsize-1) & ((o.x^(o.x+r.x))|(o.y^(o.y+r.y))|(o.z^(o.z+r.z)));
-    if(diff && (limit > octaentsize/2 || diff < leafsize*2)) leafsize *= 2;
+    if((e.inoctanode!=0)==flags || !getentboundingbox(e, o, r)) return false;
 
+    if(!insideworld(e.o)) 
+    {
+        int idx = outsideents.find(id);
+        if(flags&MODOE_ADD)
+        {
+            if(idx < 0) outsideents.add(id);
+        }
+        else if(idx >= 0) outsideents.removeunordered(idx);
+    }
+    else
+    {
+        int leafsize = octaentsize, limit = max(r.x, max(r.y, r.z));
+        while(leafsize < limit) leafsize *= 2;
+        int diff = ~(leafsize-1) & ((o.x^(o.x+r.x))|(o.y^(o.y+r.y))|(o.z^(o.z+r.z)));
+        if(diff && (limit > octaentsize/2 || diff < leafsize*2)) leafsize *= 2;
+        modifyoctaentity(flags, id, worldroot, ivec(0, 0, 0), worldsize>>1, o, r, leafsize);
+    }
     e.inoctanode = flags&MODOE_ADD ? 1 : 0;
-    modifyoctaentity(flags, id, worldroot, ivec(0, 0, 0), hdr.worldsize>>1, o, r, leafsize);
     if(e.type == ET_LIGHT) clearlightcache(id);
+    else if(e.type == ET_PARTICLES) clearparticleemitters();
     else if(flags&MODOE_ADD) lightent(e);
+    return true;
 }
 
 static inline void addentity(int id)    { modifyoctaent(MODOE_ADD|MODOE_UPDATEBB, id); }
@@ -172,7 +188,7 @@ static inline void removeentity(int id) { modifyoctaent(MODOE_UPDATEBB, id); }
 void freeoctaentities(cube &c)
 {
     if(!c.ext) return;
-    if(et->getents().length())
+    if(entities::getents().length())
     {
         while(c.ext->ents && !c.ext->ents->mapmodels.empty()) removeentity(c.ext->ents->mapmodels.pop());
         while(c.ext->ents && !c.ext->ents->other.empty())     removeentity(c.ext->ents->other.pop());
@@ -186,19 +202,68 @@ void freeoctaentities(cube &c)
 
 void entitiesinoctanodes()
 {
-    loopv(et->getents()) modifyoctaent(MODOE_ADD, i);
+    const vector<extentity *> &ents = entities::getents();
+    loopv(ents) modifyoctaent(MODOE_ADD, i);
+}
+
+static inline void findents(octaentities &oe, int low, int high, bool notspawned, const vec &pos, const vec &radius, vector<int> &found)
+{
+    vector<extentity *> &ents = entities::getents();
+    loopv(oe.other)
+    {
+        int id = oe.other[i];
+        extentity &e = *ents[id];
+        if(e.type >= low && e.type <= high && (e.spawned || notspawned) && vec(e.o).mul(radius).squaredlen() <= 1) found.add(id);
+    }
+}
+
+static inline void findents(cube *c, const ivec &o, int size, const ivec &bo, const ivec &br, int low, int high, bool notspawned, const vec &pos, const vec &radius, vector<int> &found)
+{
+    loopoctabox(o, size, bo, br)
+    {
+        if(c[i].ext && c[i].ext->ents) findents(*c[i].ext->ents, low, high, notspawned, pos, radius, found);
+        if(c[i].children && size > octaentsize) 
+        {
+            ivec co(i, o.x, o.y, o.z, size);
+            findents(c[i].children, co, size>>1, bo, br, low, high, notspawned, pos, radius, found);
+        }
+    }
+}
+
+void findents(int low, int high, bool notspawned, const vec &pos, const vec &radius, vector<int> &found)
+{
+    vec invradius(1/radius.x, 1/radius.y, 1/radius.z);
+    ivec bo = vec(pos).sub(radius).sub(1),
+         br = vec(radius).add(1).mul(2);
+    int diff = (bo.x^(bo.x+br.x)) | (bo.y^(bo.y+br.y)) | (bo.z^(bo.z+br.z)) | octaentsize,
+        scale = worldscale-1;
+    if(diff&~((1<<scale)-1) || uint(bo.x|bo.y|bo.z|(bo.x+br.x)|(bo.y+br.y)|(bo.z+br.z)) >= uint(worldsize))
+    {
+        findents(worldroot, ivec(0, 0, 0), 1<<scale, bo, br, low, high, notspawned, pos, invradius, found);
+        return;
+    }
+    cube *c = &worldroot[octastep(bo.x, bo.y, bo.z, scale)];
+    if(c->ext && c->ext->ents) findents(*c->ext->ents, low, high, notspawned, pos, invradius, found);
+    scale--;
+    while(c->children && !(diff&(1<<scale)))
+    {
+        c = &c->children[octastep(bo.x, bo.y, bo.z, scale)];
+        if(c->ext && c->ext->ents) findents(*c->ext->ents, low, high, notspawned, pos, invradius, found);
+        scale--;
+    }
+    if(c->children && 1<<scale >= octaentsize) findents(c->children, ivec(bo).mask(~((2<<scale)-1)), 1<<scale, bo, br, low, high, notspawned, pos, invradius, found);
 }
 
 char *entname(entity &e)
 {
     static string fullentname;
-    s_strcpy(fullentname, "@");
-    s_strcat(fullentname, et->entname(e.type));
-    const char *einfo = et->entnameinfo(e);
+    copystring(fullentname, "@");
+    concatstring(fullentname, entities::entname(e.type));
+    const char *einfo = entities::entnameinfo(e);
     if(*einfo)
     {
-        s_strcat(fullentname, ": ");
-        s_strcat(fullentname, einfo);
+        concatstring(fullentname, ": ");
+        concatstring(fullentname, einfo);
     }
     return fullentname;
 }
@@ -255,7 +320,7 @@ undoblock *newundoent()
     loopv(entgroup)
     {
         e->i = entgroup[i];
-        e->e = *et->getents()[entgroup[i]];
+        e->e = *entities::getents()[entgroup[i]];
         e++;
     }
     return u;
@@ -287,13 +352,13 @@ void attachentity(extentity &e)
             break;
 
         default:
-            if(e.type<ET_GAMESPECIFIC || !et->mayattach(e)) return;
+            if(e.type<ET_GAMESPECIFIC || !entities::mayattach(e)) return;
             break;
     }
 
     detachentity(e);
 
-    vector<extentity *> &ents = et->getents();
+    vector<extentity *> &ents = entities::getents();
     int closest = -1;
     float closedist = 1e10f;
     loopv(ents)
@@ -307,7 +372,7 @@ void attachentity(extentity &e)
                 break;
 
             default:
-                if(e.type<ET_GAMESPECIFIC || !et->attachent(e, *a)) continue;
+                if(e.type<ET_GAMESPECIFIC || !entities::attachent(e, *a)) continue;
                 break;
         }
         float dist = e.o.dist(a->o);
@@ -324,7 +389,7 @@ void attachentity(extentity &e)
 
 void attachentities()
 {
-    vector<extentity *> &ents = et->getents();
+    vector<extentity *> &ents = entities::getents();
     loopv(ents) attachentity(*ents[i]);
 }
 
@@ -332,7 +397,7 @@ void attachentities()
 // e         entity, currently edited ent
 // n         int,    index to currently edited ent
 #define addimplicit(f)  { if(entgroup.empty() && enthover>=0) { entadd(enthover); undonext = (enthover != oldhover); f; entgroup.drop(); } else f; }
-#define entfocus(i, f)  { int n = efocus = (i); if(n>=0) { extentity &e = *et->getents()[n]; f; } }
+#define entfocus(i, f)  { int n = efocus = (i); if(n>=0) { extentity &e = *entities::getents()[n]; f; } }
 #define entedit(i, f) \
 { \
     entfocus(i, \
@@ -341,15 +406,23 @@ void attachentities()
     f; \
     if(oldtype!=e.type) detachentity(e); \
     if(e.type!=ET_EMPTY) { addentity(n); if(oldtype!=e.type) attachentity(e); } \
-    et->editent(n)); \
+    entities::editent(n)); \
 }
-#define addgroup(exp)   { loopv(et->getents()) entfocus(i, if(exp) entadd(n)); }
+#define addgroup(exp)   { loopv(entities::getents()) entfocus(i, if(exp) entadd(n)); }
 #define setgroup(exp)   { entcancel(); addgroup(exp); }
 #define groupeditloop(f){ entlooplevel++; int _ = efocus; loopv(entgroup) entedit(entgroup[i], f); efocus = _; entlooplevel--; }
 #define groupeditpure(f){ if(entlooplevel>0) { entedit(efocus, f); } else groupeditloop(f); }
 #define groupeditundo(f){ makeundoent(); groupeditpure(f); }
 #define groupedit(f)    { addimplicit(groupeditundo(f)); }
 
+vec getselpos()
+{
+    vector<extentity *> &ents = entities::getents();
+    if(entgroup.length() && ents.inrange(entgroup[0])) return ents[entgroup[0]]->o;
+    if(ents.inrange(enthover)) return ents[enthover]->o;
+    return sel.o.tovec();
+}
+
 undoblock *copyundoents(undoblock *u)
 {
     entcancel();
@@ -395,14 +468,26 @@ void entrotate(int *cw)
 void entselectionbox(const entity &e, vec &eo, vec &es) 
 {
     model *m = NULL;
-    if(e.type == ET_MAPMODEL && (m = loadmodel(NULL, e.attr2)))
+    const char *mname = entities::entmodel(e);
+    if(mname && (m = loadmodel(mname)))
+    {   
+        m->collisionbox(0, eo, es);
+        if(es.x > es.y) es.y = es.x; else es.x = es.y; // square
+        es.z = (es.z + eo.z + 1 + entselradius)/2; // enclose ent radius box and model box
+        eo.x += e.o.x;
+        eo.y += e.o.y;
+        eo.z = e.o.z - entselradius + es.z;
+    } 
+    else if(e.type == ET_MAPMODEL && (m = loadmodel(NULL, e.attr2)))
     {
         m->collisionbox(0, eo, es);
         rotatebb(eo, es, e.attr1);
+#if 0
         if(m->collide)
             eo.z -= player->aboveeye; // wacky but true. see physics collide                    
         else
             es.div(2);  // cause the usual bb is too big...
+#endif
         eo.add(e.o);
     }   
     else
@@ -454,148 +539,141 @@ void entdrag(const vec &ray)
 
 VAR(showentradius, 0, 1, 1);
 
-void renderentradius(extentity &e)
+void renderentring(const extentity &e, float radius, int axis)
+{
+    if(radius <= 0) return;
+    glBegin(GL_LINE_LOOP);
+    loopi(16)
+    {
+        vec p(e.o);
+        p[axis>=2 ? 1 : 0] += radius*cosf(2*M_PI*i/16.0f);
+        p[axis>=1 ? 2 : 1] += radius*sinf(2*M_PI*i/16.0f);
+        glVertex3fv(p.v);
+    }
+    glEnd();
+}
+
+void renderentsphere(const extentity &e, float radius)
+{
+    if(radius <= 0) return;
+    loopk(3) renderentring(e, radius, k);
+}
+
+void renderentattachment(const extentity &e)
+{
+    if(!e.attached) return;
+    glBegin(GL_LINES);
+    glVertex3fv(e.o.v);
+    glVertex3fv(e.attached->o.v);
+    glEnd();
+}
+
+void renderentarrow(const extentity &e, const vec &dir, float radius)
+{
+    if(radius <= 0) return;
+    float arrowsize = min(radius/8, 0.5f);
+    vec target = vec(dir).mul(radius).add(e.o), arrowbase = vec(dir).mul(radius - arrowsize).add(e.o), spoke;
+    spoke.orthogonal(dir);
+    spoke.normalize();
+    spoke.mul(arrowsize);
+    glBegin(GL_LINES);
+    glVertex3fv(e.o.v);
+    glVertex3fv(target.v);
+    glEnd();
+    glBegin(GL_TRIANGLE_FAN);
+    glVertex3fv(target.v);
+    loopi(5)
+    {
+        vec p(spoke);
+        p.rotate(2*M_PI*i/4.0f, dir);
+        p.add(arrowbase);
+        glVertex3fv(p.v);
+    }
+    glEnd();
+}
+
+void renderentcone(const extentity &e, const vec &dir, float radius, float angle)
+{
+    if(radius <= 0) return;
+    vec spot = vec(dir).mul(radius*cosf(angle*RAD)).add(e.o), spoke;
+    spoke.orthogonal(dir);
+    spoke.normalize();
+    spoke.mul(radius*sinf(angle*RAD));
+    glBegin(GL_LINES);
+    loopi(8)
+    {
+        vec p(spoke);
+        p.rotate(2*M_PI*i/8.0f, dir);
+        p.add(spot);
+        glVertex3fv(e.o.v);
+        glVertex3fv(p.v);
+    }
+    glEnd();
+    glBegin(GL_LINE_LOOP);
+    loopi(8)
+    {
+        vec p(spoke);
+        p.rotate(2*M_PI*i/8.0f, dir);
+        p.add(spot);
+        glVertex3fv(p.v);
+    }
+    glEnd();
+}
+
+void renderentradius(extentity &e, bool color)
 {
-    if(!showentradius) return;
-    float radius = 0.0f, angle = 0.0f, ring = 0.0f;
-    vec dir(0, 0, 0);
-    float color[3] = {0, 1, 1};
     switch(e.type)
     {
         case ET_LIGHT:
-            radius = e.attr1;
-            color[0] = e.attr2/255.0f;
-            color[1] = e.attr3/255.0f;
-            color[2] = e.attr4/255.0f;
+            if(color) glColor3f(e.attr2/255.0f, e.attr3/255.0f, e.attr4/255.0f);
+            renderentsphere(e, e.attr1);
             break;
 
         case ET_SPOTLIGHT:
             if(e.attached)
             {
-                radius = e.attached->attr1;
+                if(color) glColor3f(0, 1, 1);
+                float radius = e.attached->attr1;
                 if(!radius) radius = 2*e.o.dist(e.attached->o);
-                dir = vec(e.o).sub(e.attached->o).normalize();
-                angle = max(1, min(90, int(e.attr1)));
+                vec dir = vec(e.o).sub(e.attached->o).normalize();
+                float angle = max(1, min(90, int(e.attr1)));
+                renderentattachment(e);
+                renderentcone(*e.attached, dir, radius, angle); 
             }
             break;
 
         case ET_SOUND:
-            radius = e.attr2;
+            if(color) glColor3f(0, 1, 1);
+            renderentsphere(e, e.attr2);
             break;
 
         case ET_ENVMAP:
         {
             extern int envmapradius;
-            radius = e.attr1 ? max(0, min(10000, int(e.attr1))) : envmapradius;
+            if(color) glColor3f(0, 1, 1);
+            renderentsphere(e, e.attr1 ? max(0, min(10000, int(e.attr1))) : envmapradius);
             break;
         }
 
         case ET_MAPMODEL:
         case ET_PLAYERSTART:
-            radius = 4;
-            if(e.type==ET_MAPMODEL && e.attr3) ring = checktriggertype(e.attr3, TRIG_COLLIDE) ? 20 : 12;
+        {
+            if(color) glColor3f(0, 1, 1);
+            vec dir;
             vecfromyawpitch(e.attr1, 0, 1, 0, dir);
+            if(e.type==ET_MAPMODEL && e.attr3) renderentring(e, checktriggertype(e.attr3, TRIG_COLLIDE) ? 20 : 12);
+            renderentarrow(e, dir, 4);
             break;
+        }
 
         default:
-            if(e.type>=ET_GAMESPECIFIC) et->entradius(e, radius, angle, dir);
+            if(e.type>=ET_GAMESPECIFIC) 
+            {
+                if(color) glColor3f(0, 1, 1);
+                entities::entradius(e, color);
+            }
             break;
     }
-    if(radius<=0) return;
-    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
-    loopj(2)
-    {
-        if(!j)
-        {
-            glDepthFunc(GL_GREATER);
-            glColor3f(0.25f, 0.25f, 0.25f);
-        }
-        else 
-        {
-            glDepthFunc(GL_LESS);
-            glColor3fv(color);
-        }
-        if(e.attached)
-        {
-            glBegin(GL_LINES);
-            glVertex3fv(e.o.v);
-            glVertex3fv(e.attached->o.v);
-            glEnd();
-        }
-        if(ring)
-        {
-            glBegin(GL_LINE_LOOP);
-            loopi(16)
-            {
-                vec p(e.o);
-                p.x += ring*cosf(2*M_PI*i/16.0f);
-                p.y += ring*sinf(2*M_PI*i/16.0f);
-                glVertex3fv(p.v);
-            }
-            glEnd();
-        }
-        if(dir.iszero()) loopk(3)
-        {
-            glBegin(GL_LINE_LOOP);
-            loopi(16)
-            {
-                vec p(e.o);
-                p[k>=2 ? 1 : 0] += radius*cosf(2*M_PI*i/16.0f);
-                p[k>=1 ? 2 : 1] += radius*sinf(2*M_PI*i/16.0f);
-                glVertex3fv(p.v);
-            }
-            glEnd();
-        }
-        else if(!angle)
-        {
-            float arrowsize = min(radius/8, 0.5f);
-            vec target(vec(dir).mul(radius).add(e.o)), arrowbase(vec(dir).mul(radius - arrowsize).add(e.o)), spoke;
-            spoke.orthogonal(dir);
-            spoke.normalize();
-            spoke.mul(arrowsize);
-            glBegin(GL_LINES);
-            glVertex3fv(e.o.v);
-            glVertex3fv(target.v);
-            glEnd();
-            glBegin(GL_TRIANGLE_FAN);
-            glVertex3fv(target.v);
-            loopi(5)
-            {
-                vec p(spoke);
-                p.rotate(2*M_PI*i/4.0f, dir);
-                p.add(arrowbase);
-                glVertex3fv(p.v);
-            }
-            glEnd();
-        }
-        else
-        {
-            vec spot(vec(dir).mul(radius*cosf(angle*RAD)).add(e.attached->o)), spoke;
-            spoke.orthogonal(dir);
-            spoke.normalize();
-            spoke.mul(radius*sinf(angle*RAD));
-            glBegin(GL_LINES);
-            loopi(8)
-            {
-                vec p(spoke);
-                p.rotate(2*M_PI*i/8.0f, dir);
-                p.add(spot);
-                glVertex3fv(e.attached->o.v);
-                glVertex3fv(p.v);
-            }
-            glEnd();
-            glBegin(GL_LINE_LOOP);
-            loopi(8)
-            {
-                vec p(spoke);
-                p.rotate(2*M_PI*i/8.0f, dir);
-                p.add(spot);
-                glVertex3fv(p.v);
-            }
-            glEnd();
-        }
-    }
-    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
 }
 
 void renderentselection(const vec &o, const vec &ray, bool entmoving)
@@ -617,9 +695,9 @@ void renderentselection(const vec &o, const vec &ray, bool entmoving)
         {
             vec a, b;
             glColor3ub(20, 20, 20);
-            (a=eo).x=0; (b=es).x=hdr.worldsize; boxs3D(a, b, 1);  
-            (a=eo).y=0; (b=es).y=hdr.worldsize; boxs3D(a, b, 1);  
-            (a=eo).z=0; (b=es).z=hdr.worldsize; boxs3D(a, b, 1);
+            (a = eo).x = eo.x - fmod(eo.x, worldsize); (b = es).x = a.x + worldsize; boxs3D(a, b, 1);  
+            (a = eo).y = eo.y - fmod(eo.y, worldsize); (b = es).y = a.x + worldsize; boxs3D(a, b, 1);  
+            (a = eo).z = eo.z - fmod(eo.z, worldsize); (b = es).z = a.x + worldsize; boxs3D(a, b, 1);
         }
         glColor3ub(150,0,0);
         glLineWidth(5);
@@ -627,8 +705,18 @@ void renderentselection(const vec &o, const vec &ray, bool entmoving)
         glLineWidth(1);
     }
 
-    loopv(entgroup) entfocus(entgroup[i], renderentradius(e));
-    if(enthover>=0) entfocus(enthover, renderentradius(e));
+    if(showentradius && (entgroup.length() || enthover >= 0))
+    {
+        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
+        glDepthFunc(GL_GREATER);
+        glColor3f(0.25f, 0.25f, 0.25f);
+        loopv(entgroup) entfocus(entgroup[i], renderentradius(e, false));
+        if(enthover>=0) entfocus(enthover, renderentradius(e, false));
+        glDepthFunc(GL_LESS);
+        loopv(entgroup) entfocus(entgroup[i], renderentradius(e, true));
+        if(enthover>=0) entfocus(enthover, renderentradius(e, true));
+        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
+    }
 }
 
 bool enttoggle(int id)
@@ -713,7 +801,7 @@ void delent()
 
 int findtype(char *what)
 {
-    for(int i = 0; *et->entname(i); i++) if(strcmp(what, et->entname(i))==0) return i;
+    for(int i = 0; *entities::entname(i); i++) if(strcmp(what, entities::entname(i))==0) return i;
     conoutf(CON_ERROR, "unknown entity type \"%s\"", what);
     return ET_EMPTY;
 }
@@ -781,15 +869,15 @@ void attachent()
 
 COMMAND(attachent, "");
 
-extentity *newentity(bool local, const vec &o, int type, int v1, int v2, int v3, int v4)
+extentity *newentity(bool local, const vec &o, int type, int v1, int v2, int v3, int v4, int v5)
 {
-    extentity &e = *et->newentity();
+    extentity &e = *entities::newentity();
     e.o = o;
     e.attr1 = v1;
     e.attr2 = v2;
     e.attr3 = v3;
     e.attr4 = v4;
-    e.attr5 = 0;
+    e.attr5 = v5;
     e.type = type;
     e.reserved = 0;
     e.spawned = false;
@@ -802,35 +890,37 @@ extentity *newentity(bool local, const vec &o, int type, int v1, int v2, int v3,
         {
                 case ET_MAPMODEL:
                 case ET_PLAYERSTART:
+                    e.attr5 = e.attr4;
                     e.attr4 = e.attr3;
                     e.attr3 = e.attr2;
                     e.attr2 = e.attr1;
                     e.attr1 = (int)camera1->yaw;
                     break;
         }
-        et->fixentity(e);
+        entities::fixentity(e);
     }
     return &e;
 }
 
-void newentity(int type, int a1, int a2, int a3, int a4)
+void newentity(int type, int a1, int a2, int a3, int a4, int a5)
 {
-    extentity *t = newentity(true, player->o, type, a1, a2, a3, a4);
+    if(entities::getents().length() >= MAXENTS) { conoutf("too many entities"); return; }
+    extentity *t = newentity(true, player->o, type, a1, a2, a3, a4, a5);
     dropentity(*t);
-    et->getents().add(t);
-    int i = et->getents().length()-1;
+    entities::getents().add(t);
+    int i = entities::getents().length()-1;
     t->type = ET_EMPTY;
     enttoggle(i);
     makeundoent();
     entedit(i, e.type = type);
 }
 
-void newent(char *what, int *a1, int *a2, int *a3, int *a4)
+void newent(char *what, int *a1, int *a2, int *a3, int *a4, int *a5)
 {
     if(noentedit()) return;
     int type = findtype(what);
     if(type != ET_EMPTY)
-        newentity(type, *a1, *a2, *a3, *a4);
+        newentity(type, *a1, *a2, *a3, *a4, *a5);
 }
 
 int entcopygrid;
@@ -850,28 +940,28 @@ void entpaste()
     if(noentedit()) return;
     if(entcopybuf.length()==0) return;
     entcancel();
-    int last = et->getents().length()-1;
+    int last = entities::getents().length()-1;
     float m = float(sel.grid)/float(entcopygrid);
     loopv(entcopybuf)
     {
         entity &c = entcopybuf[i];
         vec o(c.o);
         o.mul(m).add(sel.o.v);
-        extentity *e = newentity(true, o, ET_EMPTY, c.attr1, c.attr2, c.attr3, c.attr4);
-        et->getents().add(e);
+        extentity *e = newentity(true, o, ET_EMPTY, c.attr1, c.attr2, c.attr3, c.attr4, c.attr5);
+        entities::getents().add(e);
         entadd(++last);
     }
     int j = 0;
     groupeditundo(e.type = entcopybuf[j++].type;);
 }
 
-COMMAND(newent, "siiii");
+COMMAND(newent, "siiiii");
 COMMAND(delent, "");
 COMMAND(dropent, "");
 COMMAND(entcopy, "");
 COMMAND(entpaste, "");
 
-void entset(char *what, int *a1, int *a2, int *a3, int *a4)
+void entset(char *what, int *a1, int *a2, int *a3, int *a4, int *a5)
 {
     if(noentedit()) return;
     int type = findtype(what);
@@ -879,19 +969,57 @@ void entset(char *what, int *a1, int *a2, int *a3, int *a4)
               e.attr1=*a1;
               e.attr2=*a2;
               e.attr3=*a3;
-              e.attr4=*a4;);
+              e.attr4=*a4;
+              e.attr5=*a5);
 }
 
+void printent(extentity &e, char *buf)
+{
+    switch(e.type)
+    {
+        case ET_PARTICLES:
+            if(printparticles(e, buf)) return; 
+            break;
+ 
+        default:
+            if(e.type >= ET_GAMESPECIFIC && entities::printent(e, buf)) return;
+            break;
+    }
+    formatstring(buf)("%s %d %d %d %d %d", entities::entname(e.type), e.attr1, e.attr2, e.attr3, e.attr4, e.attr5);
+}
+
+void nearestent()
+{
+    if(noentedit()) return;
+    int closest = -1;
+    float closedist = 1e16f;
+    vector<extentity *> &ents = entities::getents();
+    loopv(ents)
+    {
+        extentity &e = *ents[i];
+        if(e.type == ET_EMPTY) continue;
+        float dist = e.o.dist(player->o);
+        if(dist < closedist)
+        {
+            closest = i;
+            closedist = dist;
+        }
+    }
+    if(closest >= 0) entadd(closest);
+}    
+            
 ICOMMAND(enthavesel,"",  (), addimplicit(intret(entgroup.length())));
 ICOMMAND(entselect, "s", (char *body), if(!noentedit()) addgroup(e.type != ET_EMPTY && entgroup.find(n)<0 && execute(body)>0));
 ICOMMAND(entloop,   "s", (char *body), if(!noentedit()) addimplicit(groupeditloop(((void)e, execute(body)))));
 ICOMMAND(insel,     "",  (), entfocus(efocus, intret(pointinsel(sel, e.o))));
-ICOMMAND(entget,    "",  (), entfocus(efocus, s_sprintfd(s)("%s %d %d %d %d", et->entname(e.type), e.attr1, e.attr2, e.attr3, e.attr4);  result(s)));
-COMMAND(entset, "siiii");
+ICOMMAND(entget,    "",  (), entfocus(efocus, string s; printent(e, s); result(s)));
+ICOMMAND(entindex,  "",  (), intret(efocus));
+COMMAND(entset, "siiiii");
+COMMAND(nearestent, "");
 
 int findentity(int type, int index, int attr1, int attr2)
 {
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
     for(int i = index; i<ents.length(); i++) 
     {
         extentity &e = *ents[i];
@@ -907,14 +1035,14 @@ int findentity(int type, int index, int attr1, int attr2)
     return -1;
 }
 
-int spawncycle = -1, fixspawn = 4;
+int spawncycle = -1;
 
 void findplayerspawn(dynent *d, int forceent, int tag)   // place at random spawn. also used by monsters!
 {
     int pick = forceent;
     if(pick<0)
     {
-        int r = fixspawn-->0 ? 7 : rnd(10)+1;
+        int r = rnd(10)+1;
         loopi(r) spawncycle = findentity(ET_PLAYERSTART, spawncycle+1, -1, tag);
         pick = spawncycle;
     }
@@ -924,14 +1052,14 @@ void findplayerspawn(dynent *d, int forceent, int tag)   // place at random spaw
         d->roll = 0;
         for(int attempt = pick;;)
         {
-            d->o = et->getents()[attempt]->o;
-            d->yaw = et->getents()[attempt]->attr1;
+            d->o = entities::getents()[attempt]->o;
+            d->yaw = entities::getents()[attempt]->attr1;
             if(entinmap(d, true)) break;
             attempt = findentity(ET_PLAYERSTART, attempt+1, -1, tag);
             if(attempt<0 || attempt==pick)
             {
-                d->o = et->getents()[attempt]->o;
-                d->yaw = et->getents()[attempt]->attr1;
+                d->o = entities::getents()[attempt]->o;
+                d->yaw = entities::getents()[attempt]->attr1;
                 entinmap(d);
                 break;
             }    
@@ -939,7 +1067,8 @@ void findplayerspawn(dynent *d, int forceent, int tag)   // place at random spaw
     }
     else
     {
-        d->o.x = d->o.y = d->o.z = 0.5f*getworldsize();
+        d->o.x = d->o.y = d->o.z = 0.5f*worldsize;
+        d->o.z += 1;
         entinmap(d);
     }
 }
@@ -959,6 +1088,7 @@ void resetmap()
     clearoverrides();
     clearmapsounds();
     cleanreflections();
+    resetblendmap();
     resetlightmaps();
     clearpvs();
     clearparticles();
@@ -967,18 +1097,19 @@ void resetmap()
     cancelsel();
     pruneundos();
 
-    setvar("gamespeed", 100);
-    setvar("paused", 0);
+    setvar("gamespeed", 100, false);
+    setvar("paused", 0, false);
 
-    et->getents().deletecontentsp();
+    entities::clearents();
+    outsideents.setsizenodelete(0);
 }
 
 void startmap(const char *name)
 {
-    cl->startmap(name);
+    game::startmap(name);
 }
 
-bool emptymap(int scale, bool force, const char *mname)    // main empty world creation routine
+bool emptymap(int scale, bool force, const char *mname, bool usecfg)    // main empty world creation routine
 {
     if(!force && !editmode) 
     {
@@ -988,37 +1119,29 @@ bool emptymap(int scale, bool force, const char *mname)    // main empty world c
 
     resetmap();
 
-    strncpy(hdr.head, "OCTA", 4);
-    hdr.version = MAPVERSION;
-    hdr.headersize = sizeof(header);
-    worldscale = scale<10 ? 10 : (scale>20 ? 20 : scale);
-    hdr.worldsize = 1<<worldscale;
+    setvar("mapscale", scale<10 ? 10 : (scale>16 ? 16 : scale), true, false);
+    setvar("mapsize", 1<<worldscale, true, false);
     
-    s_strncpy(hdr.maptitle, "Untitled Map by Unknown", 128);
-    memset(hdr.watercolour, 0, sizeof(hdr.watercolour));
-    hdr.maple = 8;
-    hdr.mapprec = 32;
-    hdr.mapllod = 0;
-    hdr.numpvs = 0;
-    hdr.lightmaps = 0;
-    memset(hdr.skylight, 0, sizeof(hdr.skylight));
-    memset(hdr.reserved, 0, sizeof(hdr.reserved));
     texmru.setsize(0);
     freeocta(worldroot);
     worldroot = newcubes(F_EMPTY);
     loopi(4) solidfaces(worldroot[i]);
 
-    if(hdr.worldsize > VVEC_INT_MASK+1) splitocta(worldroot, hdr.worldsize>>1);
+    if(worldsize > VVEC_INT_MASK+1) splitocta(worldroot, worldsize>>1);
+
+    clearmainmenu();
+
+    if(usecfg)
+    {
+        overrideidents = true;
+        execfile("data/default_map_settings.cfg", false);
+        overrideidents = false;
+    }
 
     clearlights();
     allchanged();
 
-    overrideidents = true;
-    execfile("data/default_map_settings.cfg");
-    overrideidents = false;
-
     startmap(mname);
-    player->o.z += player->eyeheight+1;
 
     return true;
 }
@@ -1030,69 +1153,65 @@ bool enlargemap(bool force)
         conoutf(CON_ERROR, "mapenlarge only allowed in edit mode");
         return false;
     }
-    if(hdr.worldsize >= 1<<20) return false;
+    if(worldsize >= 1<<16) return false;
+
+    while(outsideents.length()) removeentity(outsideents.pop());
 
     worldscale++;
-    hdr.worldsize *= 2;
+    worldsize *= 2;
     cube *c = newcubes(F_EMPTY);
     c[0].children = worldroot;
     loopi(3) solidfaces(c[i+1]);
     worldroot = c;
 
-    if(hdr.worldsize > VVEC_INT_MASK+1) splitocta(worldroot, hdr.worldsize>>1);
+    if(worldsize > VVEC_INT_MASK+1) splitocta(worldroot, worldsize>>1);
+
+    enlargeblendmap();
 
     allchanged();
 
     return true;
 }
 
-void newmap(int *i) { if(emptymap(*i, false)) cl->newmap(max(*i, 0)); }
-void mapenlarge() { if(enlargemap(false)) cl->newmap(-1); }
+void newmap(int *i) { bool force = !isconnected() && !haslocalclients(); if(force) game::forceedit(""); if(emptymap(*i, force, NULL)) game::newmap(max(*i, 0)); }
+void mapenlarge() { if(enlargemap(false)) game::newmap(-1); }
 COMMAND(newmap, "i");
 COMMAND(mapenlarge, "");
 
 void mapname()
 {
-    result(cl->getclientmap());
+    result(game::getclientmap());
 }
 
 COMMAND(mapname, "");
 
-void mapsize()
+void mpeditent(int i, const vec &o, int type, int attr1, int attr2, int attr3, int attr4, int attr5, bool local)
 {
-    int size = 0;
-    while(1<<size < hdr.worldsize) size++;
-    intret(size);
-}
-
-COMMAND(mapsize, "");
-
-void mpeditent(int i, const vec &o, int type, int attr1, int attr2, int attr3, int attr4, bool local)
-{
-    if(et->getents().length()<=i)
+    if(entities::getents().length()<=i)
     {
-        while(et->getents().length()<i) et->getents().add(et->newentity())->type = ET_EMPTY;
-        extentity *e = newentity(local, o, type, attr1, attr2, attr3, attr4);
-        et->getents().add(e);
+        if(i >= MAXENTS) return;
+        while(entities::getents().length()<i) entities::getents().add(entities::newentity())->type = ET_EMPTY;
+        extentity *e = newentity(local, o, type, attr1, attr2, attr3, attr4, attr5);
+        entities::getents().add(e);
         addentity(i);
         attachentity(*e);
     }
     else
     {
-        extentity &e = *et->getents()[i];
+        extentity &e = *entities::getents()[i];
         removeentity(i);
         int oldtype = e.type;
         if(oldtype!=type) detachentity(e);
         e.type = type;
         e.o = o;
-        e.attr1 = attr1; e.attr2 = attr2; e.attr3 = attr3; e.attr4 = attr4;
+        e.attr1 = attr1; e.attr2 = attr2; e.attr3 = attr3; e.attr4 = attr4; e.attr5 = attr5;
         addentity(i);
         if(oldtype!=type) attachentity(e);
     }
 }
 
-int getworldsize() { return hdr.worldsize; }
-int getmapversion() { return hdr.version; }
+int getworldsize() { return worldsize; }
+int getmapversion() { return mapversion; }
 
 int triggertypes[NUMTRIGGERTYPES] =
 {
@@ -1116,7 +1235,7 @@ int triggertypes[NUMTRIGGERTYPES] =
 
 void resettriggers()
 {
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
     loopv(ents)
     {
         extentity &e = *ents[i];
@@ -1128,7 +1247,7 @@ void resettriggers()
 
 void unlocktriggers(int tag, int oldstate = TRIGGER_RESET, int newstate = TRIGGERING)
 {
-    const vector<extentity *> &ents = et->getents();
+    const vector<extentity *> &ents = entities::getents();
     loopv(ents)
     {
         extentity &e = *ents[i];
@@ -1138,7 +1257,7 @@ void unlocktriggers(int tag, int oldstate = TRIGGER_RESET, int newstate = TRIGGE
             if(newstate == TRIGGER_RESETTING && checktriggertype(e.attr3, TRIG_COLLIDE) && overlapsdynent(e.o, 20)) continue;
             e.triggerstate = newstate;
             e.lasttrigger = lastmillis;
-            if(checktriggertype(e.attr3, TRIG_RUMBLE)) et->rumble(e);
+            if(checktriggertype(e.attr3, TRIG_RUMBLE)) entities::rumble(e);
         }
     }
 }
@@ -1155,7 +1274,7 @@ VAR(triggerstate, -1, 0, 1);
 
 void doleveltrigger(int trigger, int state)
 {
-    s_sprintfd(aliasname)("level_trigger_%d", trigger);
+    defformatstring(aliasname)("level_trigger_%d", trigger);
     if(identexists(aliasname))
     {
         triggerstate = state;
@@ -1166,9 +1285,8 @@ void doleveltrigger(int trigger, int state)
 void checktriggers()
 {
     if(player->state != CS_ALIVE) return;
-    vec o(player->o);
-    o.z -= player->eyeheight;
-    const vector<extentity *> &ents = et->getents();
+    vec o = player->feetpos();
+    const vector<extentity *> &ents = entities::getents();
     loopv(ents)
     {
         extentity &e = *ents[i];
@@ -1206,8 +1324,8 @@ void checktriggers()
                 }
                 e.triggerstate = TRIGGERING;
                 e.lasttrigger = lastmillis;
-                if(checktriggertype(e.attr3, TRIG_RUMBLE)) et->rumble(e);
-                et->trigger(e);
+                if(checktriggertype(e.attr3, TRIG_RUMBLE)) entities::rumble(e);
+                entities::trigger(e);
                 if(e.attr4) doleveltrigger(e.attr4, 1);
                 break;
             case TRIGGERED:
@@ -1228,8 +1346,8 @@ void checktriggers()
                 if(checktriggertype(e.attr3, TRIG_COLLIDE) && overlapsdynent(e.o, 20)) break;
                 e.triggerstate = TRIGGER_RESETTING;
                 e.lasttrigger = lastmillis;
-                if(checktriggertype(e.attr3, TRIG_RUMBLE)) et->rumble(e);
-                et->trigger(e);
+                if(checktriggertype(e.attr3, TRIG_RUMBLE)) entities::rumble(e);
+                entities::trigger(e);
                 if(e.attr4) doleveltrigger(e.attr4, 0);
                 break;
         }
